#ifndef LIB_CORE_VECMATH_H_
#define LIB_CORE_VECMATH_H_

#include <cmath>
#include <cassert>

#include "core.h"

// sum_i a[i] * b[i]
template <typename T>
inline T Dot(int ndim, const T* a, const T* b) {
  T acc = 0;
  for (int i = 0; i < ndim; ++i) {
    acc += a[i] * b[i];
  }
  return acc;
}

// sum_i a[i]
template <typename T>
inline T Sum(int ndim, const T* a) {
  T acc = 0;
  for (int i = 0; i < ndim; ++i) {
    acc += a[i];
  }
  return acc;
}

// sum_i a[i]
template <typename T>
inline T SumAbs(int ndim, const T* a) {
  T acc = 0;
  for (int i = 0; i < ndim; ++i) {
    acc += (a[i] < 0.0) ? -a[i]: a[i];
  }
  return acc;
}

// min_i a[i]
template <typename T>
inline T Min(int ndim, const T* a) {
  T val = a[0];
  for (int i = 1; i < ndim; ++i) {
    val = (a[i] < val) ? a[i]: val;
  }
  return val;
}

// max_i a[i]
template <typename T>
inline T Max(int ndim, const T* a) {
  T val = a[0];
  for (int i = 1; i < ndim; ++i) {
    val = (a[i] > val) ? a[i]: val;
  }
  return val;
}

// min_i a[i]
template <typename T>
inline int ArgMin(int ndim, const T* a) {
  T val = a[0];
  int j = 0;
  for (int i = 1; i < ndim; ++i) {
    if (a[i] < val) {
      val = a[i];
      j = i;
    }
  }
  return j;
}

// max_i a[i]
template <typename T>
inline int ArgMax(int ndim, const T* a) {
  T val = a[0];
  int j = 0;
  for (int i = 1; i < ndim; ++i) {
    if (a[i] > val) {
      val = a[i];
      j = i;
    }
  }
  return j;
}

// a[i] *= b
template <typename T>
inline void ATimesEqB(int ndim, T* a, const T b) {
  for (int i = 0; i < ndim; ++i) {
    a[i] *= b;
  }
}

// a[i] += b[i]
template <typename T>
inline void Acc(int ndim, T* a, const T* b) {
  for (int i = 0; i < ndim; ++i) {
    a[i] += b[i];
  }
}

// a[i] += b[i] * b[i]
template <typename T>
inline void AccSqr(int ndim, T* a, const T* b) {
  for (int i = 0; i < ndim; ++i) {
    a[i] += b[i] * b[i];
  }
}

// a[i] += w * b[i]
template <typename T>
inline void AccW(int ndim, T* a, const T* b, const T w) {
  for (int i = 0; i < ndim; ++i) {
    a[i] += w * b[i];
  }
}

// a[i] += (b[i] - c[i])
template <typename T>
inline void AccDiff(int ndim, T* a, const T* b, const T* c) {
  for (int i = 0; i < ndim; ++i) {
    a[i] += (b[i] - c[i]);
  }
}

// a[i] += w * (b[i] - c[i])
template <typename T>
inline void AccWDiff(int ndim, T* a, const T* b, const T* c, const T w) {
  for (int i = 0; i < ndim; ++i) {
    a[i] += w * (b[i] - c[i]);
  }
}

// a[i] += b
template <typename T>
inline void APlusEqB(int ndim, T* a, const T b) {
  for (int i = 0; i < ndim; ++i) {
    a[i] += b;
  }
}

// c[i] = a[i] - b[i]
template <typename T>
inline void AMinusB(int ndim, const T* a, const T* b, T* c) {
  for (int i = 0; i < ndim; ++i) {
    c[i] = a[i] - b[i];
  }
}

// c[i] = a[i] / b[i]
template <typename T>
inline void ADivB(int ndim, const T* a, const T* b, T* c) {
  for (int i = 0; i < ndim; ++i) {
    c[i] = a[i] / b[i];
  }
}

// c[i] = a[i] * b[i]
template <typename T>
inline void AMultB(int ndim, const T* a, const T* b, T* c) {
  for (int i = 0; i < ndim; ++i) {
    c[i] = a[i] * b[i];
  }
}

// a[i] *= b[i]
template <typename T>
inline void AMultEqB(int ndim, T* a, const T* b) {
  for (int i = 0; i < ndim; ++i) {
    a[i] *= b[i];
  }
}

// sum_i (b[i] - c[i])^2
template <typename T>
inline T SSD(int ndim, const T* b, const T* c) {
  T acc = T(0);
  for (int i = 0; i < ndim; ++i) {
    const T diff = b[i] - c[i];
    acc += diff * diff;
  }
  return acc;
}

// sum_i |b[i] - c[i]|
template <typename T>
inline T SAD(int ndim, const T* b, const T* c) {
  T acc = T(0);
  for (int i = 0; i < ndim; ++i) {
    acc += (b[i] > c[i]) ? b[i] - c[i] : c[i] - b[i];
  }
  return acc;
}

// print
template <typename T>
inline void Print(int ndim, const T* a);
template <> inline void Print(int ndim, const Real* a) {
  static char str[1024] = "";
  sprintf(str, "%f", a[0]);
  for (int i = 1; i < ndim; ++i) {
    sprintf(str, "%s %f", str, a[i]);
  }
  printf("%s\n", str);
}

// Lp-normalization (inplace)
template <typename T>
T NormalizeLp(int ndim, T* a, const int p) {
  assert(is_real(T()));
  T norm = 0.0;
  switch (p) {
    case 0: {
      for (int i = 0; i < ndim; ++i) {
        norm += (a[i] > T(0)) ? T(1) : T(0);
      }
      break;
    }
    case 1: {
      for (int i = 0; i < ndim; ++i) {
        norm += (a[i] > T(0)) ? a[i] : -a[i];
      }
      break;
    }
    case 2: {
      for (int i = 0; i < ndim; ++i) {
        norm += a[i] * a[i];
      }
      norm = std::sqrt(norm);
      break;
    }
    default:
      assert(false);
  }
  if (norm > T(0)) {
    ATimesEqB(ndim, a, T(1.0) / norm);
  }
  return norm;
}

// Alpha/Power-transformation and Lp-normalization (inplace)
template <typename T>
T TransformPowerAndNormalizeLp(int ndim, T* a, const Real alpha, const int p) {
  assert(is_real(T()));
  const Real eps = Real(1e-6);
  T norm = T(0);

  // special case
  if ((std::abs(alpha - Real(0.5)) < eps) && (p == 2)) {
    for (int i = 0; i < ndim; ++i) {
      norm += (a[i] > T(0)) ? a[i] : -a[i];
      a[i] = (a[i] > T(0)) ? std::sqrt(a[i]) : -std::sqrt(-a[i]);
    }
    if (norm > T(0)) {
      ATimesEqB(ndim, a, T(1.0) / std::sqrt(norm));
    }
    return norm;
  }

  if (!(std::abs(alpha - Real(1.0)) < eps)) {
    for (int i = 0; i < ndim; ++i) {
      a[i] = (a[i] > T(0)) ? std::pow(a[i], T(alpha)) : -std::pow(-a[i], T(alpha));
    }
  }

  if (p >= 0) {
    norm = NormalizeLp(ndim, a, p);
  }

  return norm;
}

#endif  // LIB_CORE_VECMATH_H_
