#include "dsift.h"

extern "C" {
#include <vl/dsift.h>
}

#include <iostream>

#include <cmath>
#include <cstring>

#include "../core/core.h"
#include "../core/memory.h"
#include "../core/vecmath.h"

DSIFT::DSIFT(int step, int size, int norm, bool root)
    :step_(step)
    , size_(size)
    , norm_(norm)
    , root_(root) {
  assert(step > 0);
  assert(size > 0);

  dsf_ = NULL;
}

DSIFT::~DSIFT() {
  if (dsf_ != NULL) {
    vl_dsift_delete((VlDsiftFilter *)dsf_);
  }
}

int
DSIFT::Process(float *img, int w, int h) {

  // reinstantiate filter if the image size has changed
  if (dsf_ != NULL ) {
    if (w != ((VlDsiftFilter *)dsf_)->imWidth or
        h != ((VlDsiftFilter *)dsf_)->imHeight) {
      vl_dsift_delete((VlDsiftFilter *)dsf_);
      dsf_ = NULL;
    }
  }

  if (dsf_ == NULL ) {
    dsf_ = (void *)vl_dsift_new_basic(w, h, step_, size_);
    vl_dsift_set_flat_window((VlDsiftFilter *)dsf_, true);
  }

  vl_dsift_process((VlDsiftFilter *)dsf_, img);

  return nfeat();
}

void
DSIFT::Features(float *features) {
  assert (dsf_ != NULL);

  const int nfeat_ = nfeat();
  if (nfeat_ == 0) {
    return;
  }

  const int ndim_ = this->ndim();

  const float *dsift_ =
      (float*)vl_dsift_get_descriptors((VlDsiftFilter *)dsf_);
  memcpy(features, dsift_, nfeat_ * ndim_ * sizeof(float));

  // RootSIFT
  if (root_) {
#pragma omp parallel for
    for (int i = 0; i < nfeat_; ++i) {
      float* feat_i = features + i * ndim_;  // no need to restore original norm
      TransformPowerAndNormalizeLp(ndim_, feat_i, float(0.5), 2);
    }
    return;
  }

  // VLfeat DSIFTs are L2-normalized by default
  if (norm_ == 2) {
    return;
  }

  VlDsiftKeypoint *kp_ =
      (VlDsiftKeypoint *)vl_dsift_get_keypoints((VlDsiftFilter *)dsf_);

#pragma omp parallel for
  for (int i = 0; i < nfeat_; ++i) {
    float* feat_i = features + i * ndim_;
    if (norm_ < 0) {
      ATimesEqB(ndim_, feat_i, float(kp_[i].norm));  // restore original norm
    } else {
      NormalizeLp(ndim_, feat_i, norm_);
    }
  }
}

void
DSIFT::Keypoints(float *keypoints) {
  assert (dsf_ != NULL);

  const int nfeat_ = this->nfeat();

  if (nfeat_ > 0) {
    VlDsiftKeypoint *kp =
        (VlDsiftKeypoint *)vl_dsift_get_keypoints((VlDsiftFilter *)dsf_);

#pragma omp parallel for
    for (int i = 0; i < nfeat_; ++i) {
      int k = 3 * i;
      keypoints[k    ] = kp[i].x;
      keypoints[k + 1] = kp[i].y;
      keypoints[k + 2] = kp[i].norm;
    }
  }
}

int
DSIFT::ndim(void) const {
  if (dsf_ == NULL ) {
    return 0;
  }
  return vl_dsift_get_descriptor_size((VlDsiftFilter *)dsf_);
}

int
DSIFT::nfeat(void) const {
  if (dsf_ == NULL ) {
    return 0;
  }
  return vl_dsift_get_keypoint_num((VlDsiftFilter *)dsf_);
}

#ifdef _PYTHON
DSIFT* DSIFT_new(int step, int size, int norm, bool root) {
  return new DSIFT(step, size, norm, root);
}

void DSIFT_del(DSIFT* obj) {
  delete obj;
}

int DSIFT_Process(DSIFT* obj, float *img, int w, int h) {
  return obj->Process(img, w, h);
}

int DSIFT_ndim(DSIFT* obj) {
  return obj->ndim();
}

int DSIFT_nfeat(DSIFT* obj) {
  return obj->nfeat();
}

void DSIFT_Features(DSIFT* obj, float *features) {
  obj->Features(features);
}

void DSIFT_Keypoints(DSIFT* obj, float *keypoints) {
  obj->Keypoints(keypoints);
}
#endif
