#ifndef LIB_CV_DSIFT_
#define LIB_CV_DSIFT_
/**
 * @class DSIFT
 *  dense SIFT features based on vlfeat's DSIFT
 *
 * @date  21-Jan-2014
 */
class DSIFT {
 public:
  /**
   * Constructor
   *
   * @param step     sampling step
   * @param size     cell size
   * @param norm     Lp-normalization (-1: None, 1: L1, 2: L2)
   * @param root     Compute RootSIFT features (signed square rooting + L2-normalization)
   *
   */
  DSIFT(int step=8, int size=8, int norm=2, bool root=false);
  /**
   * Destructor
   */
  ~DSIFT();
  /**
   * Compute descriptors on an image of given dimensions
   *
   * @param img  image data (row-major)
   * @param w    image width
   * @param h    image height
   *
   * @return number of SIFT features
   */
  int Process(float *img, int w, int h);
  /**
   * Store features in an array of appropriate dimensions
   *
   * @param features  array where the data will be stored
   */
  void Features(float *features);
  /**
   * Store keypoint information (x,y-coords, scale and descriptor norm) in an array of approprite dimensions
   *
   * @param features  array where the data will be stored
   */
  void Keypoints(float *keypoints);
  /**
   * Return the feature dimensionality (after Process has been called)
   *
   * @return  number of SIFT dimensions
   */
  int ndim(void) const;
  /**
   * Return the feature count after last call to process
   *
   * @return  number of descriptors
   */
  int nfeat(void) const;

 private:

  int step_;  /**< sampling step */
  int size_;  /**< cell size */
  int norm_;  /**< Lp-normalization (0, 1 or 2) */
  bool root_;  /**< if true, computes RootSIFT */

  void* dsf_;  /**< VLFeat DSIFT object (defined as void to avoid linking issues) */
};

// for the Python interface with ctypes

#ifdef _PYTHON
extern "C" {
  DSIFT* DSIFT_new(int step, int size, int norm, bool root);
  void DSIFT_del(DSIFT* obj);
  int DSIFT_Process(DSIFT* obj, float *img, int w, int h);
  int DSIFT_ndim(DSIFT* obj);
  int DSIFT_nfeat(DSIFT* obj);
  void DSIFT_Features(DSIFT* obj, float *features);
  void DSIFT_Keypoints(DSIFT* obj, float *keypoints);
}
#endif

#endif  /* LIB_CV_DSIFT_ */
