#include "bernoulli.h"

#include <iostream>
#include <cstring>
#include <limits>
#include <cmath>
#include <algorithm>

#include "../core/core.h"
#include "../core/memory.h"
#include "../core/vecmath.h"
#include "../core/random.h"

inline Real LOG(const Real p) {
  return std::log(std::max(p, std::numeric_limits<Real>::epsilon()));
}

Bernoulli::Bernoulli(int nvar)
    :ExponentialFamily<uint8_t>(nvar, nvar)
    , min_phit_(kDefBernoulliMinPHit) {
  phit_ = Allocate<Real>(npar_);

  std::fill(eta_, eta_ + npar_, Real(0));  // phit=0.5
  set_eta(eta_);
}

Bernoulli::~Bernoulli() {
  Deallocate<Real>(phit_);
}

Real
Bernoulli::min_phit(void) {
  return min_phit_;
}

void
Bernoulli::set_min_phit(const Real min_phit) {
  assert(min_phit > 0.0);
  min_phit_ = min_phit;
  set_eta(eta_);  // validate w/ new value
}

void
Bernoulli::validate_eta(void) {
  // p = 1 / (1 + exp(-eta))
  const Real min_eta = LOG(min_phit_) -  LOG(Real(1.0) - min_phit_);
  const Real max_eta = -min_eta;
  for (int i = 0; i < npar_; i++) {
    eta_[i] = std::max(min_eta, std::min(max_eta, eta_[i]));
  }
}

void
Bernoulli::set_source_param(void) {
  for (int i = 0; i < npar_; i++) {
    phit_[i] = Real(1.0) / (Real(1.0) + std::exp(-eta_[i]));
  }
}

void
Bernoulli::set_psi(void) {
  psi_ = 0.0;
  for (int i = 0; i < npar_; ++i) {
    psi_ += std::log1p(std::exp(eta_[i]));
  }
}

void
Bernoulli::set_grad_psi(void) {
  for (int i = 0; i < npar_; ++i) {
    grad_psi_[i] = 1.0 / (1.0 + std::exp(-eta_[i]));
  }
}

Real
Bernoulli::LogH(const uint8_t* x) {
  (void)x;
  return Real(0.0);
}

void
Bernoulli::SuffStat(const uint8_t* x, Real* t_out) {
  for (int i = 0; i < nvar_; ++i) {
    t_out[i] = (x[i] > 0) ? Real(1.0): Real(0.0);
  }
}

void
Bernoulli::FitFromSuffStat(const Real* sstat, const Real n) {
  assert(n > Real(0.0));

  for (int i = 0; i < npar_; ++i) {
    const Real p = sstat[i] / n;
    eta_[i] = LOG(p) - LOG(Real(1.0) - p);
  }

  return set_eta(eta_);
}

void
Bernoulli::Sample(uint8_t* x, const int n) const {
  for (int i = 0; i < n; ++i) {
    uint8_t* x_i = x + i * nvar_;
    for (int j = 0; j < nvar_; ++j) {
      std::bernoulli_distribution d(phit_[j]);
      x_i[j] = d(*((RandomNumberEngine*)gen_));
    }
  }
}

const Real*
Bernoulli::phit(void) {
  return this->phit_;
}

#ifdef _PYTHON
EXPONENTIAL_FAMILY_IMPLEMENT_CFN(Bernoulli)
Real Bernoulli_min_phit(Bernoulli* obj) {
  return obj->min_phit();
}
void Bernoulli_set_min_phit(Bernoulli* obj, const Real min_phit) {
  return obj->set_min_phit(min_phit);
}
void Bernoulli_phit(Bernoulli* obj, Real* phit) {
  memcpy(phit, obj->phit(), obj->nvar() * sizeof(Real));
}
#endif
