#include "dirichlet.h"

#include <iostream>
#include <cstring>
#include <limits>
#include <cmath>
#include <algorithm>

#include "../core/core.h"
#include "../core/memory.h"
#include "../core/vecmath.h"
#include "../core/random.h"

const double kEulerMascheroni = 0.57721566490153286060;


inline Real LOG(const Real p) {
  return std::log(std::max(p, std::numeric_limits<Real>::epsilon()));
}


// http://people.sc.fsu.edu/~jburkardt/cpp_src/asa103/asa103.html
double digamma(const double x) {
  double r;
  double value;
  double x2;

  assert(x > 0.0);

  // Initialize.
  x2 = x;
  value = 0.0;

  // Use approximation for small argument.
  if (x2 <= 0.00001) {
    value = - kEulerMascheroni - 1.0 / x2;
    return value;
  }

  // Reduce to DIGAMMA(X + N).
  while (x2 < 8.5) {
    value = value - 1.0 / x2;
    x2 = x2 + 1.0;
  }

  // Use Stirling's (actually de Moivre's) expansion.
  r = 1.0 / x2;
  value = value + std::log(x2) - 0.5 * r;
  r = r * r;
  value = value - r * (1.0 / 12.0 - r * (1.0 / 120.0 - r * 1.0 / 252.0));

  return value;
}

// http://people.sc.fsu.edu/~jburkardt/cpp_src/asa121/asa121.html
double trigamma(const double x) {
  const double a = 0.0001;
  const double b = 5.0;
  const double b2 =  0.1666666667;
  const double b4 = -0.03333333333;
  const double b6 =  0.02380952381;
  const double b8 = -0.03333333333;
  double value;
  double y;
  double z;

  assert(x > 0.0);

  z = x;

  // Use small value approximation if X <= A.
  if (x <= a) {
    value = 1.0 / x / x;
    return value;
  }

  // Increase argument to ( X + I ) >= B.
  value = 0.0;

  while (z < b) {
    value = value + 1.0 / z / z;
    z += 1.0;
  }

  // Apply asymptotic formula if argument is B or greater.
  y = 1.0 / z / z;

  value = value + 0.5 * y + (1.0 + y * (b2 + y * (b4 + y * (b6 + y * b8)))) / z;

  return value;
}


// Appendix C of Minka's paper
double inv_digamma(const double y) {
  double x0 = 0.0;
  double x1 = 0.0;

  const int max_iter = 10;
  const double tol = 1e-9;

  // init
  if (y >= -2.22) {
    x0 = std::exp(y) + 0.5;
  } else {
    x0 = -1.0 / (y + kEulerMascheroni);
  }

  // Newton iterations
  int iter = 0;
  for (iter = 0; iter < max_iter; ++iter) {
    x1 = x0 - (digamma(x0) - y) / trigamma(x0);
    if (std::abs(x1 - x0) < tol) {
      break;
    }
    x0 = x1;
  }

  if (iter == max_iter) {
    printf("WARNING (inv_digamma): unable to converge after %d iterations\n", iter);
  }

  return x1;
}


Dirichlet::Dirichlet(int nvar)
    :ExponentialFamily<Real>(nvar, nvar) {
  alpha_ = Allocate<Real>(npar_);

  const Real eta0 = std::numeric_limits<Real>::epsilon() - Real(1.0);
  std::fill(eta_, eta_ + npar_, eta0);
  set_eta(eta_);
}

Dirichlet::~Dirichlet() {
  Deallocate(alpha_);
}

void
Dirichlet::validate_eta(void) {
  // alpha = eta + 1 > 0
  const Real min_eta = Real(-1.0) + std::numeric_limits<Real>::epsilon();
  for (int i = 0; i < npar_; i++) {
    eta_[i] = std::max(min_eta, eta_[i]);
  }
}

void
Dirichlet::set_source_param(void) {
  for (int i = 0; i < npar_; i++) {
    alpha_[i] = eta_[i] + Real(1.0);
  }
}

void
Dirichlet::set_psi(void) {
  psi_ = 0.0;
  Real eta_i_pone_sum = 0.0;
  for (int i = 0; i < npar_; ++i) {
    const Real eta_i_pone = eta_[i] + Real(1.0);
    psi_ += std::lgamma(eta_i_pone);
    eta_i_pone_sum += eta_i_pone;
  }
  psi_ -= std::lgamma(eta_i_pone_sum);
}

void
Dirichlet::set_grad_psi(void) {
  Real eta_i_pone_sum = 0.0;
  for (int i = 0; i < npar_; ++i) {
    const Real eta_i_pone = eta_[i] + Real(1.0);
    grad_psi_[i] = (Real)digamma((double)eta_i_pone);
    eta_i_pone_sum += eta_i_pone;
  }

  const Real dg_sum = (Real)digamma((double)eta_i_pone_sum);
  for (int i = 0; i < npar_; ++i) {
    grad_psi_[i] -= dg_sum;
  }
}

Real
Dirichlet::LogH(const Real* x) {
  (void)x;
  return Real(0.0);
}

void
Dirichlet::SuffStat(const Real* x, Real* t_out) {
  for (int i = 0; i < nvar_; ++i) {
    t_out[i] = LOG(x[i]);
  }
}

void
Dirichlet::FitFromSuffStat(const Real* sstat, const Real n) {
  // Minka's method for the estimation of Dirichlet pdf
  assert(n > Real(0.0));

  const int max_iter = 500;
  const double tol = 1e-9;

  // avg. log-likelihood
  double log_likelihood_old = double(Dot(npar_, eta_, sstat) / n - psi_);

  // NOTE: it is assumed eta_ has been set to a reasonable initial guess,
  // eg. by the method of moments (see Minka's paper)

  // Newton iterations
  int iter = 0;
  for (iter = 0; iter < max_iter; ++iter) {

    // inv digamma
    const double digamma_sum_alphas = digamma(Sum(npar_, alpha_));
    for (int k = 0; k < npar_; ++k) {
      alpha_[k] = inv_digamma(digamma_sum_alphas + double(sstat[k] / n));
      eta_[k] = alpha_[k] - Real(1.0);
    }

    // update parameters enforcing constraints on the alphas
    set_eta(eta_);

    double log_likelihood_new = double(Dot(npar_, eta_, sstat) / n - psi_);

    // check for convergence
    if (std::abs(log_likelihood_new - log_likelihood_old) / std::abs(log_likelihood_old) < tol) {
      break;
    }

    log_likelihood_old = log_likelihood_new;
  }

  if (iter == max_iter) {
    printf("WARNING (Dirichlet::FitFromSuffStat): unable to converge after %d iterations\n", iter);
  }

  // // this has already been called in the last iteration
  // for (int k = 0; k < npar_; ++k) {
  //   eta_[k] = Real(alpha_[k] - 1.0);
  // }
  // set_eta(eta_);

  return;
}

void
Dirichlet::Sample(Real* x, const int n) const {
  for (int i = 0; i < n; ++i) {
    Real* x_i = x + i * nvar_;

    Real x_i_sum = Real(0.0);
    for (int j = 0; j < nvar_; ++j) {
      std::gamma_distribution<Real> d(alpha_[j], Real(1.0));
      x_i[j] = d(*((RandomNumberEngine*)gen_));
      x_i_sum += x_i[j];
    }
    for (int j = 0; j < nvar_; ++j) {
      x_i[j] /= x_i_sum;
    }
  }
}

const Real*
Dirichlet::alpha(void) {
  return this->alpha_;
}

#ifdef _PYTHON
EXPONENTIAL_FAMILY_IMPLEMENT_CFN(Dirichlet)
void Dirichlet_alpha(Dirichlet* obj, Real* alpha) {
  memcpy(alpha, obj->alpha(), obj->nvar() * sizeof(Real));
}
#endif
