#include "exponential_family.h"

#include <iostream>
#include <cstring>
#include <cstdint>

#include "../core/core.h"
#include "../core/memory.h"
#include "../core/vecmath.h"
#include "../core/random.h"

template <typename T>
ExponentialFamily<T>::ExponentialFamily(int nvar, int npar)
    :nvar_(nvar)
    , npar_(npar) {
  assert(nvar > 0);
  assert(npar > 0);

  eta_ = Allocate<Real>(npar_);
  grad_psi_ = Allocate<Real>(npar_);
  t_buff_ = Allocate<Real>(npar_);

  gen_ = (void*)new RandomNumberEngine;
}

template <typename T>
ExponentialFamily<T>::~ExponentialFamily() {
  Deallocate(eta_);
  Deallocate(grad_psi_);
  Deallocate(t_buff_);

  delete (RandomNumberEngine*)gen_;
}

template <typename T>
const Real*
ExponentialFamily<T>::eta(void) const {
  return eta_;
}

template <typename T>
void
ExponentialFamily<T>::set_eta(const Real *eta) {
  assert(eta != NULL);
  if (eta != eta_) {
    memcpy(eta_, eta, this->npar_ * sizeof(Real));
  }

  validate_eta();

  set_source_param();

  set_psi();

  set_grad_psi();
}

template <typename T>
Real
ExponentialFamily<T>::LogLikelihood(const T* x) {
  SuffStat(x, t_buff_);
  return LogH(x) + Dot(npar_, eta_, t_buff_) - psi_;
}

template <typename T>
void
ExponentialFamily<T>::Score(const T* x, int nsamples, Real* scr) {
  if (nsamples == 1) {
    SuffStat(x, t_buff_);
    AMinusB(npar_, t_buff_, grad_psi_, scr);
    return;
  }

  std::fill(scr, scr + npar_, Real(0));
  for (int i = 0; i < nsamples; ++i) {
    SuffStat(x, t_buff_);
    AccDiff(npar_, scr, t_buff_, grad_psi_);
  }
}

template <typename T>
int
ExponentialFamily<T>::npar(void) const {
  return npar_;
}

template <typename T>
int
ExponentialFamily<T>::nvar(void) const {
  return nvar_;
}

template <typename T>
Real
ExponentialFamily<T>::psi(void) const {
  return psi_;
}

template <typename T>
const Real*
ExponentialFamily<T>::grad_psi(void) const {
  return grad_psi_;
}

template <typename T>
void
ExponentialFamily<T>::Seed(const int s) {
  ((RandomNumberEngine*)gen_)->seed(s);
}

template class ExponentialFamily<uint8_t>;
template class ExponentialFamily<uint32_t>;
template class ExponentialFamily<Real>;
