#include "exponential_family_mixture.h"

#include <algorithm>
#include <cstring>
#include <iostream>
#include <limits>

#include "../core/memory.h"
#include "../core/vecmath.h"
#include "../core/random.h"

const Real kPstThr = Real(0.1);

template <class TFamily>
ExponentialFamilyMixture<TFamily>::ExponentialFamilyMixture(int ncomp)
    :ncomp_(ncomp)
    , min_pst_(kDefExponentialFamilyMixtureMinPst) {
  Init();

  gen_ = (void*)new RandomNumberEngine;
}

template <class TFamily>
void
ExponentialFamilyMixture<TFamily>::Init(void) {
  assert(ncomp_ > 0 && min_pst_  >= Real(0));

  pst_buff_ = Allocate<Real>(ncomp_);

  t_buff_ = NULL;

  const Real w0 = Real(1.0) / Real(ncomp_);
  weights_ = Allocate<Real>(ncomp_);
  const Real log_w0 = std::log(w0);
  log_weights_ = Allocate<Real>(ncomp_);
  for (int k = 0; k < ncomp_; ++k) {
    weights_[k] = w0;
    log_weights_[k] = log_w0;
  }

  components_ = Allocate<TFamily*>(ncomp_);
  for (int k = 0; k < ncomp_; ++k) {
    components_[k] = NULL;
  }

  nz_indx_ = Allocate<int>(ncomp_);  // indices to non-zero posteriors
}

template <class TFamily>
ExponentialFamilyMixture<TFamily>::~ExponentialFamilyMixture() {
  if (components_ != NULL) {
    for (int k = 0; k < ncomp_; ++k) {
      delete components_[k];
    }
    Deallocate(components_);
  }

  Deallocate(weights_);

  Deallocate(log_weights_);

  Deallocate(pst_buff_);

  if (t_buff_ != NULL) {
    Deallocate(t_buff_);
  }

  Deallocate(nz_indx_);

  delete (RandomNumberEngine*)gen_;
}

template <class TFamily>
const TFamily*
ExponentialFamilyMixture<TFamily>::operator[](int k) const {
  assert(k >= 0 and k < ncomp_);
  return components_[k];
}

template <class TFamily>
TFamily*
ExponentialFamilyMixture<TFamily>::operator[](int k) {
  assert(k >= 0 and k < ncomp_);
  return components_[k];
}

template <class TFamily>
const Real*
ExponentialFamilyMixture<TFamily>::weights(void) const {
  return weights_;
}

template <class TFamily>
void
ExponentialFamilyMixture<TFamily>::set_weights(const Real* weights) {
  Real sum_weights = Sum(ncomp_, weights);
  assert(sum_weights > 0);

  memcpy(weights_, weights, ncomp_ * sizeof(Real));
  ATimesEqB(ncomp_, weights_, Real(1.0 / sum_weights));

  for (int k = 0; k < ncomp_; ++k) {
    log_weights_[k] = std::log(weights_[k]);
  }
}

template <class TFamily>
const TFamily**
ExponentialFamilyMixture<TFamily>::components(void) const {
  return (const TFamily**)components_;
}


template <class TFamily>
Real
ExponentialFamilyMixture<TFamily>::min_pst(void) const {
  return min_pst_;
}

template <class TFamily>
void
ExponentialFamilyMixture<TFamily>::set_min_pst(const Real min_pst) {
  assert(min_pst >= Real(0));
  min_pst_ = min_pst;
}

template <class TFamily>
int
ExponentialFamilyMixture<TFamily>::ncomp(void) const {
  return ncomp_;
}

template <class TFamily>
int
ExponentialFamilyMixture<TFamily>::npar(void) const {
  if (components_[0] != NULL) {
    return ncomp_ * (1 + components_[0]->npar());
  }
  return 0;
}

template <class TFamily>
int
ExponentialFamilyMixture<TFamily>::nvar(void) const {
  if (components_[0] != NULL) {
    return components_[0]->nvar();
  }
  return 0;
}

// log-sum-exp formula
template <class TFamily>
Real
ExponentialFamilyMixture<TFamily>::LSE(const TSample *x, Real *log_pst) {
  const int comp_npar = components_[0]->npar();
  if (t_buff_ == NULL) {
    t_buff_ = Allocate<Real>(comp_npar);
  }

  // compute sufficient statistics only once
  components_[0]->SuffStat(x, t_buff_);

  Real max_log_pst = -std::numeric_limits<Real>::max();
#pragma omp parallel for reduction(max: max_log_pst)
  for (int k = 0; k < ncomp_; ++k) {
    // Log(H(X)) cancels out in the posteriors. It is only required for the
    // sample log-likelihood (see return line)
    log_pst[k] = log_weights_[k] +
        Dot(comp_npar, components_[k]->eta_, t_buff_) - components_[k]->psi_;
    max_log_pst = std::max(max_log_pst, log_pst[k]);
  }

  Real log_p_sum = Real(0);
  for (int k = 0; k < ncomp_; ++k) {
    log_p_sum += std::exp(log_pst[k] - max_log_pst);
  }
  log_p_sum = max_log_pst + std::log(log_p_sum);

  APlusEqB(ncomp_, log_pst, -log_p_sum);

  return log_p_sum + components_[0]->LogH(x);
}

template <class TFamily>
Real
ExponentialFamilyMixture<TFamily>::Posteriors(const TSample *x, Real *pst) {
  Real* log_pst = pst;   // just for naming consistency
  Real log_p = LSE(x, log_pst);

  Real pst_sum = Real(0);
  const Real log_min_pst_ = std::log(min_pst_);
  for (int k = 0; k < ncomp_; ++k) {
    pst[k] = (log_pst[k] < log_min_pst_) ? Real(0.0) : std::exp(log_pst[k]);
    pst_sum += pst[k];
  }
  if (pst_sum > Real(0)) {
    ATimesEqB(ncomp_, pst, Real(1.0) / pst_sum);
  }

  return log_p;
}

template <class TFamily>
void
ExponentialFamilyMixture<TFamily>::Sample(TSample* x, const int n) {
  std::discrete_distribution<int> d(weights_, weights_ + ncomp_);
  const int nvar = components_[0]->nvar();
  for (int i = 0; i < n; i++) {
    const int k = d(*((RandomNumberEngine*)gen_));
    components_[k]->Sample(x + i * nvar, 1);
  }
}

template <class TFamily>
void
ExponentialFamilyMixture<TFamily>::Score(const TSample* X, int nsamples, Real* s_out, bool hard) {
  const int nvar = components_[0]->nvar();
  const int comp_npar = components_[0]->npar();

  std::fill(s_out, s_out + npar(), Real(0));

  if (hard) {  // hard-assignment: only the component with the largest
               // posterior is active

    // we must check this since LSE is never called
    if (t_buff_ == NULL) {
      t_buff_ = Allocate<Real>(comp_npar);
    }

    Real* buff = Allocate<Real>(ncomp_);

    for (int i = 0; i < nsamples; ++i) {
      const TSample* x = X + i * nvar;
      components_[0]->SuffStat(x, t_buff_);

      // argmax pst w/o computing the posteriors explicitly
#pragma omp parallel for
      for (int k = 0; k < ncomp_; ++k) {
        buff[k] = log_weights_[k] +
            Dot(comp_npar, components_[k]->eta_, t_buff_) - components_[k]->psi_;
      }

      int kmax = 0;
      Real maxval = buff[0];
      for (int k = 1; k < ncomp_; ++k) {
        if (buff[k] > maxval) {
          maxval = buff[k];
          kmax = k;
        }
      }

      // grad wrt. alpha_kmax
      s_out[kmax] += Real(1.0) - weights_[kmax];

      // grad wrt. eta_k
      Real* s_out_kmax = s_out + ncomp_ + kmax * comp_npar;
      AccDiff(comp_npar, s_out_kmax, t_buff_, components_[kmax]->grad_psi_);
    }

    Deallocate<Real>(buff);

  } else {  // soft-assignment

    for (int i = 0; i < nsamples; ++i) {
      const TSample* x = X + i * nvar;

      Posteriors(x, pst_buff_);   // call to LSE updates t_buff_

#pragma omp parallel for schedule(dynamic)
      for (int k = 0; k < ncomp_; ++k) {
        if (pst_buff_[k] > Real(0)) {
          // grad wrt. alpha_k
          s_out[k] += pst_buff_[k] - weights_[k];

          // grad wrt. eta_k
          Real* s_out_k = s_out + ncomp_ + k * comp_npar;
          AccWDiff(comp_npar, s_out_k, t_buff_, components_[k]->grad_psi_, pst_buff_[k]);
        }
      }
    }

  }
}

template <class TFamily>
void
ExponentialFamilyMixture<TFamily>::Fit(const TSample *X, int nsamples, Real tol, int max_iter, Real alpha, Real beta) {
  assert(alpha >= Real(0));
  assert(beta >= Real(0));

  const int nvar = components_[0]->nvar();
  const int comp_npar = components_[0]->npar();

  Real ll_prev = 0.0;
  Real ll_curr = 0.0;

  // acumulators
  Real* pst_acc = Allocate<Real>(ncomp_);
  Real** t_acc = Allocate<Real*>(ncomp_);
  for (int k = 0; k < ncomp_; ++k) {
    t_acc[k] = Allocate<Real>(comp_npar);
  }

  int iter = 0;
  int splits = 0;
  const int max_splits = ncomp_;
  while ((iter < max_iter) && (splits < max_splits)) {
    // reset accumulators
    std::fill(pst_acc, pst_acc + ncomp_, Real(0));
    for (int k = 0; k < ncomp_; ++k) {
      std::fill(t_acc[k], t_acc[k] + comp_npar, Real(0));
    }

    // E-Step
    ll_curr = Real(0);

    for (int i = 0; i < nsamples; ++i) {
      const TSample *X_i = X + i * nvar;
      const Real log_p = Posteriors(X_i, pst_buff_);  // LSE updates t_buff_
      ll_curr += log_p;  // accumulate log-likelihood

      // accumulate sufficient statistics
#pragma omp parallel for
      for (int k = 0; k < ncomp_; ++k) {
        pst_acc[k] += pst_buff_[k];
        AccW(comp_npar, t_acc[k], t_buff_, pst_buff_[k]);
      }
    }

    // handle empty components
    if (this->HandleEmpty(pst_acc, beta)) {
      iter = 0;  // if some component was empty, restart EM with the new
                 // (hopefully fixed) parameters as initial guess.
      ++splits;
      continue;
    }

    ll_curr /= Real(nsamples);  // avg log-likelihood
    printf("Iteration %2d, average log-likhelihood=%.3e", iter, ll_curr);

    // check for convergence
    if (iter > 0) {
      Real ll_diff = (ll_curr - ll_prev) / std::abs(ll_prev);
      printf(" (%.3e)\n", ll_diff);
      if (ll_diff < tol) {
        break;
      }
    } else {
      printf("\n");
    }

    ll_prev = ll_curr;

    // M-Step
    const Real pst_acc_sum = Sum(ncomp_, pst_acc);
    for (int k = 0; k < ncomp_; ++k) {
      // case beta=0: do not update empty components
      if ((beta < std::numeric_limits<Real>::epsilon()) && \
          (pst_acc[k] / pst_acc_sum <= kPstThr * min_pst_)) {
        continue;
      }
      components_[k]->FitFromSuffStat(t_acc[k], pst_acc[k]);
    }

    // adds a Dirichlet prior
    APlusEqB(ncomp_, pst_acc, alpha);

    // set weights and normalize them properly
    this->set_weights(pst_acc);

    ++iter;
  }

  if (splits == max_splits) {
    printf("Too many splits. Try to retrain the model using more samples.\n");
  }

  Deallocate(pst_acc);
  for (int k = 0; k < ncomp_; ++k) {
    Deallocate(t_acc[k]);
  }
  Deallocate(t_acc);
}

template <class TFamily>
bool
ExponentialFamilyMixture<TFamily>::HandleEmpty(Real *pst_acc, Real beta) {
  // if beta is zero, do nothing
  if (beta < std::numeric_limits<Real>::epsilon()) {
    return false;
  }

  const Real pst_acc_sum = Sum(ncomp_, pst_acc);
  const int comp_npar = components_[0]->npar();

  // check if the minimum number of samples assigned effectively to the
  // components is above a given threshold.
  const Real min_pst_acc = Min(ncomp_, pst_acc);
  if (min_pst_acc / pst_acc_sum > kPstThr * min_pst_) {
    return false;
  }

  // if not, remove that component and split the largest one into two and retrain.
  const int argmin_pst_acc = ArgMin(ncomp_, pst_acc);
  const int argmax_pst_acc = ArgMax(ncomp_, pst_acc);

  printf("Component %d is empty. Splitting component %d\n", argmin_pst_acc, argmax_pst_acc);

  assert(argmax_pst_acc != argmin_pst_acc);

  Real* eta1 = Allocate<Real>(comp_npar);
  Real* eta2 = Allocate<Real>(comp_npar);

  const Real* eta0 = components_[argmax_pst_acc]->eta();
  for (int i = 0; i < comp_npar; ++i) {
    eta1[i] = eta0[i] * (1.0 + beta);
    eta2[i] = eta0[i] * (1.0 - beta);
  }
  // having a valid eta vector is ensured by the set_parms_safe() method of the
  // base distribution (ExponentialFamily class)
  components_[argmax_pst_acc]->set_eta(eta1);
  components_[argmin_pst_acc]->set_eta(eta2);
  weights_[argmax_pst_acc] /= Real(2.0);
  weights_[argmin_pst_acc] = weights_[argmax_pst_acc];
  set_weights(weights_);

  Deallocate(eta1);
  Deallocate(eta2);

  return true;
}

template <class TFamily>
const Real*
ExponentialFamilyMixture<TFamily>::eta(const int k) const {
  return components_[k]->eta();
}

template <class TFamily>
void
ExponentialFamilyMixture<TFamily>::set_eta(const int k, const Real* eta) {
  assert(eta != NULL);
  components_[k]->set_eta(eta);
}

template <typename TFamily>
void
ExponentialFamilyMixture<TFamily>::Seed(const int s) {
  ((RandomNumberEngine*)gen_)->seed(s);

  std::uniform_int_distribution<int> d(0, std::numeric_limits<int>::max());
  for (int k = 0; k < ncomp_; k++) {
    components_[k]->Seed(d(*((RandomNumberEngine*)gen_)));
  }
}

#include "gaussian.h"
template class ExponentialFamilyMixture<Gaussian>;

#include "bernoulli.h"
template class ExponentialFamilyMixture<Bernoulli>;

#include "laplace.h"
template class ExponentialFamilyMixture<Laplace>;

#include "wishart.h"
template class ExponentialFamilyMixture<Wishart>;

#include "poisson.h"
template class ExponentialFamilyMixture<Poisson>;

#include "dirichlet.h"
template class ExponentialFamilyMixture<Dirichlet>;

#include "multinomial.h"
template class ExponentialFamilyMixture<Multinomial>;
