#include "fisher_vector.h"

#include <algorithm>
#include <cstring>
#include <iostream>
#include <limits>
#include <algorithm>

#include <Eigen/Dense>
#include <Eigen/Eigenvalues>

#include "../core/memory.h"
#include "../core/vecmath.h"

using Eigen::Map;
using Eigen::Matrix;
using Eigen::RowMajor;
using Eigen::Dynamic;
using Eigen::SelfAdjointEigenSolver;

typedef Matrix <Real, Dynamic, Dynamic, RowMajor> Mat;
typedef Matrix <const Real, Dynamic, Dynamic, RowMajor> CMat;
typedef Map <Matrix<Real, Dynamic, Dynamic, RowMajor> > MapMat;
typedef Map <const Matrix<Real, Dynamic, Dynamic, RowMajor> > CMapMat;
typedef Map <Matrix<double, Dynamic, Dynamic, RowMajor> > DMapMat;
typedef Map <const Matrix<double, Dynamic, Dynamic, RowMajor> > CDMapMat;


template <class TMixture>
FisherVector<TMixture>::FisherVector(TMixture* model, int normalizer, Real theta, bool hard, Real alpha, int norm)
    : model_(model)
    , normalizer_(normalizer)
    , theta_(theta)
    , hard_(hard)
    , alpha_(alpha)
    , norm_(norm) {
  assert(model != NULL);
  assert(alpha_ > Real(0));
  L_ = Allocate<Real>(nL());
}

template <class TMixture>
FisherVector<TMixture>::~FisherVector() {
  Deallocate<Real>(L_);
}

template <class TMixture>
const Real*
FisherVector<TMixture>::L(void) const {
  return L_;
}

template <class TMixture>
int
FisherVector<TMixture>::nL(void) const {
  switch (normalizer_) {
    case EYE: {
      return 0;
    }
    case VAR:
    case DIAG: {
      return model_->npar();
    }
    case BLOCK: {
      const int comp_npar = (*model_)[0]->npar();
      return model_->ncomp() * (comp_npar * comp_npar + 1);
    }
    case FULL: {
      return model_->npar() * model_->npar();
    }
    default: {
      assert(false);
    }
  }
}

template <class TMixture>
void
FisherVector<TMixture>::set_L(const Real* L) {
  if (normalizer_ == EYE) {
    return;
  }
  memcpy(L_, L, nL() * sizeof(Real));
}

template <class TMixture>
Real
FisherVector<TMixture>::theta(void) {
  return theta_;
}

template <class TMixture>
void
FisherVector<TMixture>::set_theta(const Real val) {
  theta_ = val;
}

template <class TMixture>
void
FisherVector<TMixture>::Fit(const int nsamples) {
  switch (normalizer_) {
    case EYE: {
      return;
    }
    case VAR: {
      return this->fit_var(nsamples);
    }
    case DIAG: {
      return this->fit_diag(nsamples);
    }
    case BLOCK: {
      return this->fit_block(nsamples);
    }
    case FULL: {
      return this->fit_full(nsamples);
    }
    default: {
      assert(false);
    }
  }
}

template <class TMixture>
void
FisherVector<TMixture>::fit_var(const int nsamples) {
  const Real eps = std::numeric_limits<Real>::epsilon();
  const int nvar = model_->nvar();
  const int npar = model_->npar();

  TSample* x_buff = Allocate<TSample>(nvar);
  Real* s_buff = Allocate<Real>(npar);
  double* acc2 = Allocate<double>(npar);
  std::fill(acc2, acc2 + npar, double(0));

  for (int i = 0; i < nsamples; ++i) {
    model_->Sample(x_buff, 1);
    model_->Score(x_buff, 1, s_buff, hard_);

    for (int j = 0; j < npar; ++j) {
      acc2[j] += double(s_buff[j] * s_buff[j]);
    }
  }

  for (int j = 0; j < npar; ++j) {
    L_[j] = Real(1.0) / (Real(std::sqrt(acc2[j] / double(nsamples))) + eps);
  }

  Deallocate<TSample>(x_buff);
  Deallocate<Real>(s_buff);
  Deallocate<double>(acc2);
}

template <class TMixture>
void
FisherVector<TMixture>::fit_diag(const int nsamples) {
  const Real eps = std::numeric_limits<Real>::epsilon();
  const int nvar = model_->nvar();
  const int ncomp = model_->ncomp();
  const int comp_npar = (*model_)[0]->npar();

  const Real* weights = model_->weights();

  TSample* x_buff = Allocate<TSample>(nvar * ncomp);
  Real* s_buff = Allocate<Real>(comp_npar * ncomp);
  double* acc2 = Allocate<double>(comp_npar * ncomp);
  std::fill(acc2, acc2 + comp_npar * ncomp, double(0));

#pragma omp parallel for
  for (int k = 0; k < ncomp; k++) {
    // normalizer corresp. to the k-th mixing weight
    L_[k] = Real(1.0) / (std::sqrt(weights[k]) + eps);

    TSample* x_buff_k = x_buff + k * nvar;
    Real* s_buff_k = s_buff + k * comp_npar;
    Real* acc2_k = acc2 + k * comp_npar;
    Real* L_k = L_ + ncomp + k * comp_npar;

    for (int i = 0; i < nsamples; ++i) {
      (*model_)[k]->Sample(x_buff_k, 1);
      (*model_)[k]->Score(x_buff_k, 1, s_buff_k);

      for (int j = 0; j < comp_npar; ++j) {
        acc2_k[j] += double(s_buff_k[j] * s_buff_k[j]);
      }
    }

    for (int j = 0; j < comp_npar; ++j) {
      const Real std_j = Real(std::sqrt(acc2_k[j] / double(nsamples)));
      L_k[j] = Real(1.0) / (std::sqrt(weights[k]) * std_j + eps);
    }
  }

  Deallocate<TSample>(x_buff);
  Deallocate<Real>(s_buff);
  Deallocate<double>(acc2);
}

template <class TMixture>
void
FisherVector<TMixture>::fit_block(const int nsamples) {
  const int nvar = model_->nvar();
  const int ncomp = model_->ncomp();
  const int comp_npar = (*model_)[0]->npar();
  const int comp_npar2 = comp_npar * comp_npar;

  const Real* weights = model_->weights();

  TSample* x_buff = Allocate<TSample>(ncomp * nvar);
  Real* s_buff = Allocate<Real>(ncomp * comp_npar);
  double* acc2 = Allocate<double>(ncomp * comp_npar2);
  std::fill(acc2, acc2 + ncomp * comp_npar2, double(0));

  Mat eye = Mat::Identity(comp_npar, comp_npar);

#pragma omp parallel for
  for (int k = 0; k < ncomp; k++) {
    assert(weights[k] > Real(0));

    // normalizer corresp. to the k-th mixing weight
    L_[k] = Real(1.0) / std::sqrt(weights[k]);

    TSample* x_buff_k = x_buff + k * nvar;
    Real* s_buff_k = s_buff + k * comp_npar;
    Real* acc2_k = acc2 + k * comp_npar2;

    MapMat s_buff_k_map(s_buff_k, comp_npar, 1);
    DMapMat acc2_k_map(acc2_k, comp_npar, comp_npar);

    for (int i = 0; i < nsamples; ++i) {
      (*model_)[k]->Sample(x_buff_k, 1);
      (*model_)[k]->Score(x_buff_k, 1, s_buff_k);

      acc2_k_map += (s_buff_k_map * s_buff_k_map.transpose()).cast<double>();
    }

    Real* L_k = L_ + ncomp + k * comp_npar2;
    MapMat L_k_map(L_k, comp_npar, comp_npar);

    // sample covariance
    L_k_map = (acc2_k_map / double(nsamples)).cast<Real>();

    // covariance estimate with shrinkage
    L_k_map = (Real(1.0) - kShrinkage) * L_k_map + kShrinkage * (L_k_map.trace() / Real(comp_npar)) * eye;

    // normalizer
    L_k_map = L_k_map.inverse().llt().matrixL().transpose();
    L_k_map /= std::sqrt(weights[k]);
  }

  Deallocate<TSample>(x_buff);
  Deallocate<Real>(s_buff);
  Deallocate<double>(acc2);
}

template <class TMixture>
void
FisherVector<TMixture>::fit_full(const int nsamples) {
  const int nvar = model_->nvar();
  const int npar = model_->npar();

  TSample* x_buff = Allocate<TSample>(nvar);
  Real* s_buff = Allocate<Real>(npar);
  double* acc2 = Allocate<double>(npar * npar);
  std::fill(acc2, acc2 + npar * npar, double(0));

  Mat eye = Mat::Identity(npar, npar);

  MapMat s_buff_map(s_buff, npar, 1);
  DMapMat acc2_map(acc2, npar, npar);

  for (int i = 0; i < nsamples; ++i) {
    model_->Sample(x_buff, 1);
    model_->Score(x_buff, 1, s_buff, hard_);

    acc2_map += (s_buff_map * s_buff_map.transpose()).cast<double>();
  }

  MapMat L_map(L_, npar, npar);

  // sample covariance
  L_map = (acc2_map / double(nsamples)).cast<Real>();

  // covariance estimate with shrinkage
  L_map = (Real(1.0) - kShrinkage) * L_map + kShrinkage * (L_map.trace() / Real(npar)) * eye;

  // normalizer
  L_map = L_map.inverse().llt().matrixL().transpose();

  Deallocate<TSample>(x_buff);
  Deallocate<Real>(s_buff);
  Deallocate<double>(acc2);
}

template <class TMixture>
void
FisherVector<TMixture>::Process(const TSample* X, const int nsamples, Real *efv) {
  const int npar0 = model_->npar();
  const int npar1 = (theta_ > Real(0)) ? npar0 + 1 : npar0;

  // gradients are stored as (g_theta, g_alpha_1, ..., g_alpha_k, g_eta_1, ... g_eta_k)
  Real* efv0 = (theta_ > Real(0)) ? efv + 1 : efv;

  // gradient vector for the sample
  model_->Score(X, nsamples, efv0, hard_);

  // FIM normalization
  switch (normalizer_) {
    case EYE: {
      break;
    }
    case VAR:
    case DIAG: {
      AMultEqB(npar0, efv0, L_);
      break;
    }
    case BLOCK: {
      const int ncomp = model_->ncomp();
      const int comp_npar = (*model_)[0]->npar();

#pragma omp parallel for
      for (int k = 0; k < ncomp; k++) {
        efv0[k] *= L_[k];

        Real* L_k = L_ + ncomp + k * comp_npar * comp_npar;
        MapMat L_k_map(L_k, comp_npar, comp_npar);

        Real* efv0_k = efv0 + ncomp + k * comp_npar;
        MapMat efv0_k_map(efv0_k, comp_npar, 1);

        efv0_k_map = L_k_map.triangularView<Eigen::Upper>() * efv0_k_map;
      }
      break;
    }
    case FULL: {
      MapMat L_map(L_, npar0, npar0);
      MapMat efv0_map(efv0, npar0, 1);
      efv0_map = L_map.triangularView<Eigen::Upper>() * efv0_map;
      break;
    }
    default: {
      assert(false);
    }
  }

  // Poisson term
  if (theta_ > Real(0)) {
    if (normalizer_ == EYE) {  // no FIM normalization
      efv[0] = (Real(nsamples) - theta_) / theta_;
    } else {
      efv[0] = (Real(nsamples) - theta_) / std::sqrt(theta_);
    }
  }

  if (normalizer_ != EYE) {
    if (norm_ < 0) {
      // w/ FIM but no Lp-normalization we need to scale the gradients by
      // sqrt(theta) or sqrt(N) according to whether the model includes the
      // gradient wrt the Poisson term (theta) or not
      if (theta_ > Real(0)) {
        ATimesEqB(npar0, efv0, Real(1.0) / std::sqrt(theta_));
      } else {
        ATimesEqB(npar0, efv0, Real(1.0) / std::sqrt(nsamples));
      }
    } else if (theta_ > Real(0)) {
      // in case we use an Lp-normalization the above factor cancels out but we
      // need to include it also on the gradient wrt theta since the
      // Lp-normalization acts on the whole vector
      efv[0] *= std::sqrt(theta_);
    }
  }

//   // log-transform and Lp-normalization
// #pragma omp parallel for
//   for (int i = 0; i < npar1; i++) {
//     if (efv[i] > Real(0)) {
//       efv[i] = +std::log10(Real(1.0) + efv[i]);
//     } else {
//       efv[i] = -std::log10(Real(1.0) - efv[i]);
//     }
//   }
//   TransformPowerAndNormalizeLp(npar1, efv, Real(1.0), norm_);

  // alpha-transform and Lp-normalization
  TransformPowerAndNormalizeLp(npar1, efv, alpha_, norm_);
}

// template instantations
#include "gaussian_mixture.h"
template class FisherVector<GaussianMixture>;

#include "bernoulli_mixture.h"
template class FisherVector<BernoulliMixture>;

#include "laplace_mixture.h"
template class FisherVector<LaplaceMixture>;

#include "wishart_mixture.h"
template class FisherVector<WishartMixture>;

#include "poisson_mixture.h"
template class FisherVector<PoissonMixture>;

#include "dirichlet_mixture.h"
template class FisherVector<DirichletMixture>;

#include "multinomial_mixture.h"
template class FisherVector<MultinomialMixture>;

#ifdef _PYTHON
FISHER_VECTOR_IMPLEMENT_CFN(GaussianMixture)
FISHER_VECTOR_IMPLEMENT_CFN(BernoulliMixture)
FISHER_VECTOR_IMPLEMENT_CFN(LaplaceMixture)
FISHER_VECTOR_IMPLEMENT_CFN(WishartMixture)
FISHER_VECTOR_IMPLEMENT_CFN(PoissonMixture)
FISHER_VECTOR_IMPLEMENT_CFN(DirichletMixture)
FISHER_VECTOR_IMPLEMENT_CFN(MultinomialMixture)
#endif
