#include "gaussian.h"

#include <iostream>
#include <cstring>
#include <limits>
#include <cmath>

#include "../core/core.h"
#include "../core/memory.h"
#include "../core/vecmath.h"
#include "../core/random.h"

Gaussian::Gaussian(int nvar)
    :ExponentialFamily<Real>(nvar, 2 * nvar)
    , min_var_(kDefGaussianMinVar) {
  kLogH_ = -0.5 * Real(nvar_) * std::log(2.0 * M_PI);

  mu_ = Allocate<Real>(nvar_);
  sigma_ = Allocate<Real>(nvar_);

  for (int i = 0; i < nvar_; i++) {
    eta_[i] = 0.0;  // mean=0
    eta_[nvar_ + i] = -0.5;  // var=1
  }
  set_eta(eta_);

}

Gaussian::~Gaussian() {
  Deallocate<Real>(mu_);
  Deallocate<Real>(sigma_);
}

Real
Gaussian::min_var(void) {
  return min_var_;
}

void
Gaussian::set_min_var(const Real min_var) {
  assert(min_var > 0.0);
  min_var_ = min_var;
  set_eta(eta_);  // validate w/ new value
}

void
Gaussian::validate_eta() {
  // var = -0.5 / eta_2
  // the constraints are enforced in the source domain in order to take into
  // account the interplay between eta_1, eta_2 and the variance
  Real* eta_1 = eta_;
  Real* eta_2 = eta_ + nvar_;
  for (int i = 0; i < nvar_; i++) {
    const Real mu_i = Real(-0.5) * eta_1[i] / eta_2[i];
    const Real var_i = std::max(min_var_, Real(-0.5) / eta_2[i]);
    eta_1[i] = mu_i / var_i;
    eta_2[i] = Real(-0.5) / var_i;
  }
}

void
Gaussian::set_source_param(void) {
  const Real* eta_1 = eta_;
  const Real* eta_2 = eta_ + nvar_;
  for (int i = 0; i < nvar_; i++) {
    mu_[i] = Real(-0.5) * eta_1[i] / eta_2[i];
    sigma_[i] = std::sqrt(Real(-0.5) / eta_2[i]);
  }
}

void
Gaussian::set_psi(void) {
  const Real* eta_1 = eta_;
  const Real* eta_2 = eta_ + nvar_;

  psi_ = 0;
  for (int i = 0; i < nvar_; ++i) {
    psi_ += (-0.25 * (eta_1[i] * eta_1[i] / eta_2[i])
             -0.5 * std::log(-2.0 * eta_2[i]));
  }
}

void
Gaussian::set_grad_psi(void) {
  const Real* eta_1 = eta_;
  const Real* eta_2 = eta_ + nvar_;

  Real *grad_psi_1 = grad_psi_;
  Real *grad_psi_2 = grad_psi_ + nvar_;
  for (int i = 0; i < nvar_; ++i) {
    grad_psi_1[i] = -0.5 * eta_1[i] / eta_2[i];
    grad_psi_2[i] = grad_psi_1[i] * grad_psi_1[i] - 0.5 / eta_2[i];
  }
}

Real
Gaussian::LogH(const Real* x) {
  (void)x;
  return kLogH_;
}

void
Gaussian::SuffStat(const Real* x, Real* t_out) {
  Real *t_out_1 = t_out;
  Real* t_out_2 = t_out_1 + nvar_;
  for (int i = 0; i < nvar_; ++i) {
    t_out_1[i] = x[i];
    t_out_2[i] = x[i] * x[i];
  }
}

void
Gaussian::FitFromSuffStat(const Real* sstat, const Real n) {
  assert(n > Real(0.0));
  Real *eta_1 = eta_;
  Real *eta_2 = eta_ + nvar_;

  const Real *sstat_1 = sstat;
  const Real *sstat_2 = sstat + nvar_;

  for (int i = 0; i < nvar_; ++i) {
    const Real mean_i = sstat_1[i] / n;
    const Real cov_i = sstat_2[i] / n - mean_i * mean_i;
    eta_1[i] = mean_i / (cov_i + std::numeric_limits<Real>::epsilon());
    eta_2[i] = Real(-0.5) / (cov_i + std::numeric_limits<Real>::epsilon());
  }

  return set_eta(eta_);
}

void
Gaussian::Sample(Real* x, const int n) const {
  for (int i = 0; i < n; ++i) {
    Real* x_i = x + i * nvar_;
    for (int j = 0; j < nvar_; ++j) {
      std::normal_distribution<Real> d(mu_[j], sigma_[j]);
      x_i[j] = d(*((RandomNumberEngine*)gen_));
    }
  }
  // for (int j = 0; j < nvar_; j++, x++) {
  //   std::normal_distribution<Real> d(mu_[j], sigma_[j]);
  //   for (int i = 0; i < n; i++) {
  //    x[i * nvar_]  = d(*((RandomNumberEngine*)gen_));
  //   }
  // }
}

const Real*
Gaussian::mu(void) {
  return this->mu_;
}

const Real*
Gaussian::sigma(void) {
  return this->sigma_;
}

#ifdef _PYTHON
EXPONENTIAL_FAMILY_IMPLEMENT_CFN(Gaussian)
Real Gaussian_min_var(Gaussian* obj) {
  return obj->min_var();
}
void Gaussian_set_min_var(Gaussian* obj, const Real min_var) {
  return obj->set_min_var(min_var);
}
void Gaussian_mu(Gaussian* obj, Real* mu) {
  memcpy(mu, obj->mu(), obj->nvar() * sizeof(Real));
}
void Gaussian_sigma(Gaussian* obj, Real* sigma) {
  memcpy(sigma, obj->sigma(), obj->nvar() * sizeof(Real));
}
#endif
