#include "kmeans.h"

#include <algorithm>
#include <cstring>
#include <iostream>
#include <limits>
#include <algorithm>

#include "../core/memory.h"
#include "../core/vecmath.h"
#include "../core/random.h"

KMeans::KMeans(int K, int ndim, int max_iter, Real epsilon)
    : K_(K)
    , ndim_(ndim)
    , max_iter_(max_iter)
    , epsilon_(epsilon) {
  centroids_ = Allocate<Real>(K_ * ndim_);
  labels_ = NULL;
  n_labels_ = 0;
}

KMeans::~KMeans() {
  Deallocate<Real>(centroids_);
  if (labels_ != NULL) {
    Deallocate<int>(labels_);
  }
}

Real
KMeans::Init(const Real* X, int nsamples) {
  RandomNumberEngine gen;
  std::uniform_int_distribution<uint32_t> dui(0, nsamples);
  std::uniform_real_distribution<Real> dur(Real(0), Real(1));

  Real** pcentroids = Allocate<Real*>(K_);
  Real* dist = Allocate<Real>(nsamples);

  uint32_t idx = dui(gen);
  pcentroids[0] = const_cast<Real*>(X + idx * ndim_);

  double curr_pot = 0.0;
#pragma omp parallel for reduction(+: curr_pot)
  for (uint32_t i = 0; i < uint32_t(nsamples); i++) {
    dist[i] = SSD(ndim_, pcentroids[0], X + i * ndim_);
    curr_pot += double(dist[i]);
  }

  const int num_local_tries = 2 + int(std::log(Real(K_)));

  for (int k = 1; k < K_; k++) {

    // choose center
    int best_new_idx = -1;
    double best_new_pot = -1.0;
    for (int j = 0; j < num_local_tries; j++) {

      // choose center
      double rval = curr_pot * dur(gen);
      for (idx = 0; idx < uint32_t(nsamples) - 1; idx++) {  // nsamples-1 in case it never brakes
        if (dist[idx] > 0.0) {
          if (rval <= dist[idx]) {
            break;
          } else {
            rval -= dist[idx];
          }
        }
      }

      // compute new potential
      double new_pot = 0.0;
#pragma omp parallel for reduction(+:new_pot)
      for (uint32_t i = 0; i < uint32_t(nsamples); i++) {
        const Real ssd = SSD(ndim_, X + idx * ndim_, X + i * ndim_);
        new_pot += double(std::min(dist[i], ssd));
      }

      // store best result
      if ((best_new_pot < 0.0) || (new_pot < best_new_pot)) {
        best_new_pot = new_pot;
        best_new_idx = idx;
      }

    }  // for j

    pcentroids[k] = const_cast<Real*>(X + best_new_idx * ndim_);

    // compute new potential
    curr_pot = best_new_pot;
#pragma omp parallel for
    for (uint32_t i = 0; i < uint32_t(nsamples); i++) {
      const Real ssd = SSD(ndim_, pcentroids[k], X + i * ndim_);
      dist[i] = std::min(dist[i], ssd);
    }

  }  // for k

  Deallocate<Real*>(pcentroids);
  Deallocate<Real>(dist);

  return Real(curr_pot) / Real(nsamples);
}

Real
KMeans::RunOneIter(const Real* X, int nsamples) {
  // sample assignment
  Real pot = Encode(X, nsamples);

  // centroids computation
#pragma omp parallel for schedule(dynamic)
  for (int k = 0; k < K_; k++) {
    Real* centroid_k = centroids_ + k * ndim_;
    std::fill(centroid_k, centroid_k + ndim_, Real(0));

    uint32_t nk = 0;
    for (uint32_t i = 0; i < uint32_t(nsamples); i++) {
      if (labels_[i] == k) {
        Acc(ndim_, centroid_k, X + i * ndim_);
        nk += 1;
      }
    }

    if (nk > 0) {
      ATimesEqB(ndim_, centroid_k, Real(1.0) / Real(nk));
    }
  }

  return pot;
}

Real
KMeans::Fit(const Real* X, int nsamples) {
  Real pot = Init(X, nsamples);
  Encode(X, nsamples); // sample assignments

  int* prev_labels = Allocate<int>(nsamples);
  std::memcpy(prev_labels, labels_, nsamples * sizeof(int));

  std::cerr << "iter 0, avg. potential = " << pot << std::endl;

  for (int iter = 1; iter <= max_iter_; iter++) {
    pot = RunOneIter(X, nsamples);
    Encode(X, nsamples);

    int nchanged=0;
#pragma omp parallel for reduction(+:nchanged)
    for (uint32_t i = 0; i < uint32_t(nsamples); ++i) {
      nchanged += int(labels_[i] != prev_labels[i]);
    }

    std::cout << "iter " << iter << ", avg. potential = " << pot
              << ", reassignments = " << 100.0 * Real(nchanged) / Real(nsamples) << "%" << std::endl;

    if (Real(nchanged) < epsilon_ * Real(nsamples)) {
      break;
    }

    std::memcpy(prev_labels, labels_, nsamples * sizeof(int));
  }

  Deallocate<int>(prev_labels);

  return pot;
}

const Real*
KMeans::centroids(void) const {
  return centroids_;
}

void
KMeans::set_centroids(const Real* centroids) {
  std::memcpy(centroids_, centroids, K_ * ndim_ * sizeof(Real));
}

const int*
KMeans::labels(void) const {
  return labels_;
}

int
KMeans::K(void) {
  return K_;
}

int
KMeans::ndim(void) {
  return ndim_;
}

Real
KMeans::Quantize(Real* X, int nsamples) {
  Real pot = Encode(X, nsamples);
#pragma omp parallel for reduction(+: pot)
  for (int i = 0; i < nsamples; i++) {
    Real* x_i = X + i * ndim_;
    std::memcpy(x_i, centroids_ + labels_[i] * ndim_, ndim_ * sizeof(Real));
  }
  return pot;
}

Real
KMeans::Encode(const Real* X, int nsamples) {
  double pot = 0.0;

  if (n_labels_ < nsamples) {
    if (labels_ != NULL) {
      Deallocate<int>(labels_);
    }
    labels_ = Allocate<int>(nsamples);
    n_labels_ = nsamples;
  }

#pragma omp parallel for reduction(+: pot)
  for (int i = 0; i < nsamples; i++) {
    Real* x_i = const_cast<Real*>(X + i * ndim_);
    Real min_dist = SSD(ndim_, centroids_, x_i);
    labels_[i] = 0;
    for (int k = 1; k < K_; k++) {
      Real dist = SSD(ndim_, centroids_ + k * ndim_, x_i);
      if (dist < min_dist) {
        min_dist = dist;
        labels_[i] = k;
      }
    }
    pot += min_dist;
  }
  return Real(pot) / Real(nsamples);
}


#ifdef _PYTHON
Real kmeans_fit(int K, int ndim, int max_iter, Real epsilon, const Real* X, int nsamples, Real* centroids) {
  KMeans model(K, ndim, max_iter, epsilon);
  Real pot = model.Fit(X, nsamples);
  std::memcpy(centroids, model.centroids(), ndim * K * sizeof(Real));
  return pot;
}

Real kmeans_quantize(int K, int ndim, Real* centroids, Real* X, int nsamples) {
    double pot = 0.0;
#pragma omp parallel for reduction(+: pot)
  for (int i = 0; i < nsamples; i++) {
    Real* x_i = X + i * ndim;
    Real min_dist = SSD(ndim, centroids, x_i);
    int min_dist_idx = 0;
    for (int k = 1; k < K; k++) {
      const Real dist = SSD(ndim, centroids + k * ndim, x_i);
      if (dist < min_dist) {
        min_dist = dist;
        min_dist_idx = k;
      }
    }
    std::memcpy(x_i, centroids + min_dist_idx * ndim, ndim * sizeof(Real));
    pot += min_dist;
  }
  return Real(pot) / Real(nsamples);
}

Real kmeans_encode(int K, int ndim, Real* centroids, Real* X, int nsamples, int* q) {
  double pot = 0.0;
#pragma omp parallel for reduction(+: pot)
  for (int i = 0; i < nsamples; i++) {
    Real* x_i = X + i * ndim;
    Real min_dist = SSD(ndim, centroids, x_i);
    q[i] = 0;
    for (int k = 1; k < K; k++) {
      const Real dist = SSD(ndim, centroids + k * ndim, x_i);
      if (dist < min_dist) {
        min_dist = dist;
        q[i] = k;
      }
    }
    pot += min_dist;
  }
  return Real(pot) / Real(nsamples);
}

// KMeans* KMeans_new(int K, int ndim, int max_iter, Real epsilon) {
//   return new KMeans(K, ndim, max_iter, epsilon);
// }

// void KMeans_del(KMeans* model) {
//   delete model;
// }

// Real KMeans_Fit(KMeans* model, const Real* X, int nsamples) {
//   Real pot = model->Fit(X, nsamples);
//   return pot;
// }

// void KMeans_centroids(KMeans* model, Real* centroids) {
//   std::memcpy(centroids, model->centroids(), model->ndim() * model->K());
// }

// void KMeans_set_centroids(KMeans* model, const Real* centroids) {
//   model->set_centroids(centroids);
// }

// Real KMeans_Quantize(KMeans* model, Real* X, int nsamples) {
//   return model->Quantize(X, nsamples);
// }

// Real KMeans_Encode(KMeans* model, Real* X, int nsamples, int* q) {
//   return model->Encode(X, nsamples, q);
// }
#endif
