#include "laplace.h"

#include <iostream>
#include <algorithm>
#include <cstring>
#include <limits>
#include <cmath>

#include "../core/core.h"
#include "../core/memory.h"
#include "../core/vecmath.h"
#include "../core/random.h"

Laplace::Laplace(int nvar)
    :ExponentialFamily<Real>(nvar, nvar)
    , min_scale_(kDefLaplaceMinScale) {

  scale_ = Allocate<Real>(npar_);

  for (int i = 0; i < nvar_; i++) {
    eta_[i] = Real(-1.0);  // scale=1
  }
  set_eta(eta_);
}

Laplace::~Laplace() {
  Deallocate<Real>(scale_);
}

Real
Laplace::min_scale(void) {
  return min_scale_;
}

void
Laplace::set_min_scale(const Real min_scale) {
  assert(min_scale > 0.0);
  min_scale_ = min_scale;
  set_eta(eta_);  // validate w/ new value
}

void
Laplace::validate_eta(void) {
  // scale = -1/eta
  const Real min_eta = Real(-1.0) / min_scale_;
  for (int i = 0; i < npar_; i++) {
    eta_[i] = std::max(min_eta, eta_[i]);
  }
}

void
Laplace::set_source_param(void) {
  for (int i = 0; i < npar_; i++) {
    scale_[i] = Real(-1.0) / eta_[i];
  }
}

void
Laplace::set_psi(void) {
  psi_ = 0;
  for (int i = 0; i < npar_; ++i) {
    psi_ += std::log(Real(-2.0) / eta_[i]);
  }
}

void
Laplace::set_grad_psi(void) {
  for (int i = 0; i < npar_; ++i) {
    grad_psi_[i] = Real(-1.0) / eta_[i];
  }
}

Real
Laplace::LogH(const Real* x) {
  (void)x;
  return Real(0.0);
}

void
Laplace::SuffStat(const Real* x, Real* t_out) {
  for (int i = 0; i < npar_; ++i) {
    t_out[i] = std::abs(x[i]);
  }
}

void
Laplace::FitFromSuffStat(const Real* sstat, const Real n) {
  assert(n > Real(0.0));

  for (int i = 0; i < npar_; ++i) {
    const Real b_i = sstat[i] / n + std::numeric_limits<Real>::epsilon();
    eta_[i] = Real(-1.0) / b_i;
  }

  return set_eta(eta_);
}

void
Laplace::Sample(Real* x, const int n) const {
  std::bernoulli_distribution d1(0.5);
  for (int i = 0; i < n; ++i) {
    Real* x_i = x + i * nvar_;
    for (int j = 0; j < nvar_; ++j) {
      std::exponential_distribution<Real> d2(Real(1.0) / scale_[j]);
      x_i[j] = d2(*((RandomNumberEngine*)gen_)) * (Real(2.0) * Real(d1(*((RandomNumberEngine*)gen_))) - Real(1.0));
    }
  }
}

const Real*
Laplace::scale(void) {
  return this->scale_;
}

#ifdef _PYTHON
EXPONENTIAL_FAMILY_IMPLEMENT_CFN(Laplace)
Real Laplace_min_scale(Laplace* obj) {
  return obj->min_scale();
}
void Laplace_set_min_scale(Laplace* obj, const Real min_scale) {
  return obj->set_min_scale(min_scale);
}
void Laplace_scale(Laplace* obj, Real* scale) {
  memcpy(scale, obj->scale(), obj->nvar() * sizeof(Real));
}
#endif
