#include "lrpdd_opt.h"

#include <iostream>

#include <Eigen/Dense>

using Eigen::Map;
using Eigen::Matrix;
using Eigen::RowMajor;
using Eigen::Dynamic;

typedef Matrix<Real, Dynamic, Dynamic, RowMajor> Mat;
typedef Map<Matrix<Real, Dynamic, Dynamic, RowMajor> > MapMat;
typedef Map<const Matrix<Real, Dynamic, Dynamic, RowMajor> > CMapMat;

Real
lrpdd_opt(const int K, const int nsamples, const int ndim,
          const Real* X_, const Real* VarX_, const Real *p,
          Real *grad) {

  CMapMat X(X_, nsamples, ndim);
  CMapMat VarX(VarX_, ndim, 1);
  CMapMat D(p, ndim, 1);
  CMapMat a(p + ndim, ndim, K);

  Mat invD2 = (D.cwiseAbs2()).cwiseInverse();
  Mat invD2a = invD2.asDiagonal() * a;

  Mat Q = Mat::Identity(K, K) + a.transpose() * invD2a;
  Eigen::LLT<Mat> llt_invQ(Q.inverse());
  Mat L = llt_invQ.matrixL();

  Real log_det_D = 0.0;
  for (int i = 0; i < ndim; i++) {
    log_det_D += std::log(std::abs(D(i,0)));
  }

  Real log_det_Q = 0.0;
  for (int k = 0; k < K; k++) {
    log_det_Q -= std::log(std::abs(L(k,k)));
  }
  log_det_Q *= Real(2.0);

  Mat Xta = X * a;

  Mat Dt = D.transpose();
  Real err = - 2.0 * log_det_D - log_det_Q;
#pragma omp parallel for reduction(+:err)
  for (int i = 0; i < nsamples; i++) {
    err += (Xta.row(i).squaredNorm() + X.row(i).cwiseProduct(Dt).squaredNorm())
        / Real(nsamples);
  }

  if (grad == NULL) {
    return err;
  }

  MapMat grad_D(grad, ndim, 1);
  Mat invD2aL = invD2a * L;
  Mat diagP = invD2 - (invD2aL.cwiseAbs2().rowwise().sum());
  grad_D = 2.0 * (VarX - diagP).cwiseProduct(D);

  Mat Sigmaa(ndim, K);
  Sigmaa.fill(0);
#pragma omp parallel for
  for (int k = 0; k < K; k++) {
    for (int i = 0; i < nsamples; i++) {
      Sigmaa.col(k) += X.row(i).transpose() * Xta(i, k);
    }
  }
  Sigmaa /= Real(nsamples);

  MapMat grad_a(grad + ndim, ndim, K);
  grad_a = 2.0 * (Sigmaa - invD2a + invD2aL * invD2aL.transpose() * a);

  // static bool print = true;
  // if (print) {
  //   std::cout << grad_D.transpose() << std::endl;
  //   std::cout << grad_a.transpose() << std::endl;
  //   print=false;
  // }

  return err;
}
