#include "multinomial.h"

#include <iostream>
#include <cstring>
#include <limits>
#include <cmath>
#include <algorithm>

#include "../core/core.h"
#include "../core/memory.h"
#include "../core/vecmath.h"
#include "../core/random.h"

inline Real LOG(const Real p) {
  return std::log(std::max(p, std::numeric_limits<Real>::epsilon()));
}

Multinomial::Multinomial(int nvar)
    :ExponentialFamily<uint32_t>(nvar, nvar)
    , min_p_(kDefMultinomialMinP) {

  p_ = Allocate<Real>(npar_);

  Real init_eta = std::log(1.0 / Real(nvar_));
  for (int i = 0; i < nvar_; i++) {
    eta_[i] = init_eta;  // p=(1.0/nvar_)
  }
  set_eta(eta_);
}

Multinomial::~Multinomial() {
  Deallocate<Real>(p_);
}

uint32_t
Multinomial::ntrials(void) {
  return ntrials_;
}

void
Multinomial::set_ntrials(const uint32_t ntrials) {
  assert(ntrials > 0);
  ntrials_ = ntrials;

  log_ntrials_factorial_ = 0.0;
  for (int i = 2; i <= ntrials_; i++) {
    log_ntrials_factorial_ += std::log(i);
  }
}


Real
Multinomial::min_p(void) {
  return min_p_;
}

void
Multinomial::set_min_p(const Real min_p) {
  assert(min_p > 0.0);
  min_p_ = min_p;
  set_eta(eta_);  // validate w/ new value
}

void
Multinomial::validate_eta(void) {
  // p_k = exp(eta_k), sum_k exp(eta_k) = 1
  // it is easier to enforce the normalization constraint on the source domain
  Real acc1 = Real(0.0);
  Real acc2 = Real(0.0);
  for(int i = 0; i < npar_; i++) {
    p_[i] = std::exp(eta_[i]);  // compute p from eta
    if (p_[i] < min_p_) {
      p_[i] = min_p_;
      acc1 += p_[i];
    } else {
      p_[i] = -p_[i];  // the sign is used as a flag to indicate > 0
      acc2 -= p_[i];
    }
  }
  for(int i = 0; i < npar_; i++) {
    if (p_[i] < Real(0.0)) {
      p_[i] = (-p_[i] - acc1) / acc2;
    }
    eta_[i] = LOG(p_[i]);  // compute eta from p
  }
}

void
Multinomial::set_source_param(void) {
  for(int i = 0; i < npar_; i++) {
    p_[i] = std::exp(eta_[i]);
  }
}

void
Multinomial::set_psi(void) {
  psi_ = 0.0;
}

void
Multinomial::set_grad_psi(void) {
  for (int i = 0; i < npar_; ++i) {
    grad_psi_[i] = 0.0;
  }
}

Real
Multinomial::LogH(const uint32_t* x) {
  // ln[h(x)] = -sum[from i=1 to i=nvar_](ln(x[i]!))
  Real log_h = log_ntrials_factorial_;
  for(int i = 0; i < nvar_; ++i) {
    for(uint32_t j = 1; j <= x[i]; ++j) {
      log_h -= std::log(j);
    }
  }
  return log_h;
}

void
Multinomial::SuffStat(const uint32_t* x, Real* t_out) {
  for (int i = 0; i < nvar_; ++i) {
    t_out[i] = Real(x[i]);
  }
}

void
Multinomial::FitFromSuffStat(const Real* sstat, const Real n) {
  assert(n > Real(0.0));

  Real sum = 0.0;
  for (int i = 0; i < npar_; ++i) {
    p_[i] = sstat[i] / (n * Real(ntrials_));
    sum += p_[i];
  }

  for (int i = 0; i < npar_; ++i) {
    eta_[i] = LOG(p_[i] / sum);
  }

  return set_eta(eta_);
}

void
Multinomial::Sample(uint32_t* x, const int n) const {
  std::fill(x, x + n * nvar_, uint32_t(0));
  std::discrete_distribution<int> d(p_, p_ + nvar_);
  for (int i = 0; i < n; ++i) {
    uint32_t* x_i = x + i * nvar_;
    for (int j = 0; j < ntrials_; ++j) {   // TODO: look for a more efficient way
      x_i[d(*((RandomNumberEngine*)gen_))] += uint32_t(1);
    }
  }
}

const Real*
Multinomial::p(void) {
  return this->p_;
}

#ifdef _PYTHON
EXPONENTIAL_FAMILY_IMPLEMENT_CFN(Multinomial)
uint32_t Multinomial_ntrials(Multinomial* obj) {
  return obj->ntrials();
}
void Multinomial_set_ntrials(Multinomial* obj, const uint32_t ntrials) {
  return obj->set_ntrials(ntrials);
}
Real Multinomial_min_p(Multinomial* obj) {
  return obj->min_p();
}
void Multinomial_set_min_p(Multinomial* obj, const Real min_p) {
  return obj->set_min_p(min_p);
}
void Multinomial_p(Multinomial* obj, Real* p) {
  memcpy(p, obj->p(), obj->nvar() * sizeof(Real));
}
#endif
