#include "poisson.h"

#include <iostream>
#include <cstring>
#include <limits>
#include <cmath>
#include <algorithm>

#include "../core/core.h"
#include "../core/memory.h"
#include "../core/vecmath.h"
#include "../core/random.h"

inline Real LOG(const Real p) {
  return std::log(std::max(p, std::numeric_limits<Real>::epsilon()));
}

Poisson::Poisson(int nvar)
    :ExponentialFamily<uint32_t>(nvar, nvar)
    , min_lambda_(kDefPoissonMinLambda) {

  lambda_ = Allocate<Real>(npar_);

  for (int i = 0; i < nvar_; i++) {
    eta_[i] = std::log(1.0);  // lambda=1.0
  }
  set_eta(eta_);
}

Poisson::~Poisson() {
  Deallocate<Real>(lambda_);
}

Real
Poisson::min_lambda(void) {
  return min_lambda_;
}

void
Poisson::set_min_lambda(const Real min_lambda) {
  assert(min_lambda > 0.0);
  min_lambda_ = min_lambda;
  set_eta(eta_);  // validate w/ new value
}

void
Poisson::validate_eta(void) {
  // lambda = exp(eta)
  const Real min_eta = std::log(min_lambda_);
  for(int i = 0; i < npar_; i++) {
    eta_[i] = std::max(min_eta, eta_[i]);
  }
}

void
Poisson::set_source_param(void) {
  for(int i = 0; i < npar_; i++) {
    lambda_[i] = std::exp(eta_[i]);
  }
}

void
Poisson::set_psi(void) {
  psi_ = 0.0;
  for (int i = 0; i < npar_; ++i) {
    psi_ += std::exp(eta_[i]);
  }
}

void
Poisson::set_grad_psi(void) {
  for (int i = 0; i < npar_; ++i) {
    grad_psi_[i] = std::exp(eta_[i]);
  }
}

Real
Poisson::LogH(const uint32_t* x) {
  // ln[h(x)] = -sum[from i=1 to i=nvar_](ln(x[i]!))
  Real log_h = 0.0;
  for(int i = 0; i < nvar_; ++i) {
    for(uint32_t j = 1; j <= x[i]; ++j) {
      log_h += std::log(j);
    }
  }
  return -log_h;
}

void
Poisson::SuffStat(const uint32_t* x, Real* t_out) {
  for (int i = 0; i < nvar_; ++i) {
    t_out[i] = Real(x[i]);
  }
}

void
Poisson::FitFromSuffStat(const Real* sstat, const Real n) {
  assert(n > Real(0.0));

  const Real log_n = std::log(n);

  for (int i = 0; i < npar_; ++i) {
    eta_[i] = LOG(sstat[i]) - log_n;
  }

  return set_eta(eta_);
}

void
Poisson::Sample(uint32_t* x, const int n) const {
  for (int i = 0; i < n; ++i) {
    uint32_t* x_i = x + i * nvar_;
    for (int j = 0; j < nvar_; ++j) {
      std::poisson_distribution<uint32_t> d(lambda_[j]);
      x_i[j] = d(*((RandomNumberEngine*)gen_));
    }
  }
}

const Real*
Poisson::lambda(void) {
  return this->lambda_;
}

#ifdef _PYTHON
EXPONENTIAL_FAMILY_IMPLEMENT_CFN(Poisson)
Real Poisson_min_lambda(Poisson* obj) {
  return obj->min_lambda();
}
void Poisson_set_min_lambda(Poisson* obj, const Real min_lambda) {
  return obj->set_min_lambda(min_lambda);
}
void Poisson_lambda(Poisson* obj, Real* lambda) {
  memcpy(lambda, obj->lambda(), obj->nvar() * sizeof(Real));
}
#endif
