#include "wishart.h"

#include <iostream>
#include <cstring>
#include <limits>
#include <cmath>
#include <algorithm>

#include <Eigen/Dense>
#include <Eigen/Eigenvalues>

#include "../core/core.h"
#include "../core/memory.h"
#include "../core/vecmath.h"
#include "../core/random.h"

using Eigen::Map;
using Eigen::Matrix;
using Eigen::RowMajor;
using Eigen::Dynamic;
using Eigen::SelfAdjointEigenSolver;

typedef Matrix <Real, Dynamic, Dynamic, RowMajor> Mat;
typedef Matrix <const Real, Dynamic, Dynamic, RowMajor> CMat;
typedef Map <Matrix<Real, Dynamic, Dynamic, RowMajor> > MapMat;
typedef Map <const Matrix<Real, Dynamic, Dynamic, RowMajor> > CMapMat;

// ref http://en.wikipedia.org/wiki/Multivariate_gamma_function
Real log_multivariate_gamma(Real a, int d) {
  Real out = 0.0;
#pragma omp parallel for reduction(+:out)
  for (int j = 1; j <= d; j++) {
    out += std::lgamma(a - Real(0.5) * (Real(j - 1)));
  }
  return std::log(M_PI) * Real(d * (d - 1)) * Real(0.25) + out;
}

Wishart::Wishart(int nvar)
    :ExponentialFamily<Real>(nvar, nvar) {
  ndim_ = std::sqrt(nvar_);

  set_dof(Real(ndim_));

  V_ = Allocate<Real>(nvar_);

  L_ = Allocate<Real>(nvar_);

  /* init with diagonal matrix */
  std::fill(eta_, eta_ + npar_, Real(0));
  const Real eta0 = Real(-0.5);
  for (int i = 0; i < ndim_; i++) {
    eta_[i * ndim_ + i] = eta0;
  }
}

Wishart::~Wishart() {
  Deallocate<Real>(V_);
  Deallocate<Real>(L_);
}

Real
Wishart::dof(void) {
  return dof_;
}

void
Wishart::set_dof(const Real dof) {
  assert(dof > Real(ndim_ - 1));
  dof_ = dof;
  set_psi();
  set_grad_psi();
}

void
Wishart::validate_eta() {
  MapMat eta_map(eta_, ndim_, ndim_);
  eta_map = Real(0.5) * (eta_map + eta_map.transpose());  // force SPD
}

void
Wishart::set_source_param() {
  MapMat eta_map(eta_, ndim_, ndim_);

  MapMat V_map(V_, ndim_, ndim_);
  V_map = Real(-0.5) * eta_map.inverse();

  MapMat L_map(L_, ndim_, ndim_);
  L_map = V_map.llt().matrixL();
}

void
Wishart::set_psi(void) {
  // log(det(-eta))
  CMapMat eta(eta_, ndim_, ndim_);
  Mat L = (-eta).llt().matrixL();
  const Real log_det_m_eta = Real(2.0) * L.diagonal().array().log().sum();

  const Real dof_otwo = Real(0.5) * dof_;
  psi_ = log_multivariate_gamma(dof_otwo, ndim_) - dof_otwo * log_det_m_eta;
}

void
Wishart::set_grad_psi(void) {
  MapMat eta_map(eta_, ndim_, ndim_);
  MapMat grad_psi_map(grad_psi_, ndim_, ndim_);
  grad_psi_map = Real(-0.5) * dof_ * eta_map.inverse();
}

Real
Wishart::LogH(const Real* x) {
  // log(det(x))
  CMapMat X(x, ndim_, ndim_);
  Mat L = X.llt().matrixL();
  const Real log_det_x = Real(2.0) * L.diagonal().array().log().sum();

  return Real(0.5) * (dof_ - Real(ndim_ + 1)) * log_det_x;
}

void
Wishart::SuffStat(const Real* x, Real* t_out) {
  memcpy(t_out, x, nvar_ * sizeof(Real));
}

void
Wishart::FitFromSuffStat(const Real* sstat, const Real n) {
  assert(n > Real(0.0));

  CMapMat ss_map(sstat, ndim_, ndim_);
  MapMat eta_map(eta_, ndim_, ndim_);  // matrix natural param
  eta_map = Real(-0.5) * dof_ * (ss_map / n).inverse();

  return set_eta(eta_);
}

void
Wishart::Sample(Real* x, const int n) const {
  std::fill(x, x + n * nvar_, Real(0.0));

  MapMat L_map(L_, ndim_, ndim_);

  std::normal_distribution<Real> normal(Real(0.0), Real(1.0));

  for (int i = 0; i < n; ++i) {
    Real* x_i = x + i * nvar_;

    for (int j = 0; j < ndim_; ++j) {
      // fill diagonal
      std::gamma_distribution<Real> gamma(Real(0.5) * (dof_ - Real(j)), Real(2.0));
      x_i[j * ndim_ + j] = std::sqrt(gamma(*((RandomNumberEngine*)gen_)));

      // fill lower triangle
      for (int k = 0; k < j; ++k) {
        x_i[j * ndim_ + k] = normal(*((RandomNumberEngine*)gen_));
      }
    }

    MapMat x_i_map(x_i, ndim_, ndim_);
    x_i_map = L_map * x_i_map;
    x_i_map = x_i_map * x_i_map.transpose();
  }
}

const Real*
Wishart::V(void) {
  return this->V_;
}

const Real*
Wishart::L(void) {
  return this->L_;
}

#ifdef _PYTHON
EXPONENTIAL_FAMILY_IMPLEMENT_CFN(Wishart)
Real Wishart_dof(Wishart* obj) {
  return obj->dof();
}
void Wishart_set_dof(Wishart* obj, const Real dof) {
  return obj->set_dof(dof);
}
void Wishart_V(Wishart* obj, Real* V) {
  memcpy(V, obj->V(), obj->nvar() * sizeof(Real));
}
void Wishart_L(Wishart* obj, Real* L) {
  memcpy(L, obj->L(), obj->nvar() * sizeof(Real));
}
#endif
