#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
sys.path.append('.')

import unittest
import numpy as np

from skimage import io

from vrl.cv import LBP


class TestLBP(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        self.img = io.imread('tests/mandrill.jpg', as_grey=True)
        return

    def test_compute_img(self):
        croped_img = self.img[:10, :10]
        lbp = LBP(max_layer=0, downscale=1.414, P=8, R=1.0)
        data = lbp.transform(croped_img)
        loc = data['loc']
        feats = data['data']
        self.assertEqual(len(feats), len(loc))
        self.assertEqual(len(feats), croped_img.shape[0] * croped_img.shape[1])
        self.assertEqual(len(feats), len(loc))

        return

    def test_compute_downscale_max_layer(self):
        lbp = LBP(max_layer=0, downscale=1.414, P=8, R=1.0)
        nfeats = self.img.shape[0] * self.img.shape[1]
        data = lbp.transform(self.img)
        loc = data['loc']
        feats = data['data']
        self.assertEqual(len(feats), len(loc))
        self.assertEqual(len(feats), nfeats)
        lbp = LBP(max_layer=0, downscale=2.0, P=8, R=1.0)
        data = lbp.transform(self.img)
        loc = data['loc']
        feats = data['data']
        self.assertEqual(len(feats), len(loc))
        self.assertEqual(len(feats), nfeats)

    def test_compute_max_layer(self):
        lbp = LBP(max_layer=0, downscale=1.414, P=8, R=1.0)
        mat = np.zeros((50, 50))
        data = lbp.transform(mat)
        loc = data['loc']
        feats = data['data']
        self.assertEqual(len(feats), len(loc))
        self.assertEqual(len(feats), 50 * 50)
        lbp = LBP(max_layer=1, downscale=2.0, P=8, R=1.0)
        data = lbp.transform(mat)
        loc = data['loc']
        feats = data['data']
        self.assertEqual(len(feats), len(loc))
        self.assertEqual(len(feats), 50 * 50 + 25 * 25)
        return

    def test_compute_simple(self):
        lbp = LBP(max_layer=0, downscale=1.414, P=8, R=1.0)
        mat = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])
        data = lbp.transform(mat)
        loc = data['loc']
        feats = data['data']
        self.assertEqual(len(feats), len(loc))
        self.assertEqual(len(feats), 9)
        for val in feats[4]:
            self.assertEqual(val, 0)
        mat = 1 - mat
        data = lbp.transform(mat)
        loc = data['loc']
        feats = data['data']
        self.assertEqual(len(feats), len(loc))
        self.assertEqual(len(feats), 9)
        for val in feats[4]:
            self.assertEqual(val, 1)
        return

    def test_compute_simple_loc_one_scale(self):
        lbp = LBP(max_layer=0, downscale=1.414, P=8, R=1.0)
        mat = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])
        data = lbp.transform(mat)
        loc = data['loc']
        feats = data['data']
        self.assertEqual(len(feats), len(loc))
        self.assertEqual(len(feats), 9)
        for val in feats[4]:
            self.assertEqual(val, 0)

        self.assertEqual(loc[0][0], 0.0)
        self.assertEqual(loc[0][1], 0.0)
        self.assertEqual(loc[0][2], 1.0)

        self.assertEqual(loc[8][0], 2.0)
        self.assertEqual(loc[8][1], 2.0)
        self.assertEqual(loc[8][2], 1.0)

        return

    def test_compute_simple_loc_two_scales(self):
        lbp = LBP(max_layer=1, downscale=1.0 / 0.7, P=8, R=1.0)
        mat = np.array([[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]])
        data = lbp.transform(mat)
        loc = data['loc']
        feats = data['data']
        self.assertEqual(len(feats), len(loc))
        self.assertEqual(len(feats), 4 * 4 + 3 * 3)
        for val in feats[5]:
            self.assertEqual(val, 0)

        self.assertEqual(loc[0][0], 0.0)
        self.assertEqual(loc[0][1], 0.0)
        self.assertEqual(loc[0][2], 1.0)

        self.assertEqual(loc[16][0], 0.0)
        self.assertEqual(loc[16][1], 0.0)
        self.assertEqual(loc[16][2], 1.0 / 0.7)

        return


if __name__ == '__main__':
    unittest.main()
