#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
sys.path.append('.')
import unittest
import numpy as np
from numpy.linalg import norm
from numpy.testing import assert_array_almost_equal

from vrl.ml import Bernoulli


class TestBernoulli(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        return

    def test_min_phit_max_pmiss(self):
        nvar = 2
        min_phit = 1e-5
        phit = 1e-6
        p = np.array([phit, 1.0 - phit])
        eta = np.log(np.divide(p, 1 - p))
        bernoulli = Bernoulli(nvar, min_phit=min_phit)
        self.assertEqual(bernoulli.npar, nvar)
        self.assertEqual(bernoulli.min_phit, min_phit)
        bernoulli.eta_ = eta
        min_p = np.array([min_phit, 1.0 - min_phit])
        min_eta = np.log(np.divide(min_p, 1 - min_p))
        assert_array_almost_equal(bernoulli.eta_, min_eta)
        assert_array_almost_equal(bernoulli.phit, min_p)

        min_phit = 1e-4
        min_p = np.array([min_phit, 1.0 - min_phit])
        min_eta = np.log(np.divide(min_p, 1 - min_p))
        bernoulli.min_phit = min_phit
        self.assertEqual(bernoulli.min_phit, min_phit)
        assert_array_almost_equal(bernoulli.eta_, min_eta)
        assert_array_almost_equal(bernoulli.phit, min_p)

    def test_log_likelihood(self):
        nvar = 4
        p = np.array([0.5, 0.2, 0.3, 0.7])
        eta = np.log(np.divide(p, 1 - p))
        bernoulli = Bernoulli(nvar)
        bernoulli.eta_ = eta
        assert_array_almost_equal(bernoulli.eta_, eta)
        x = np.array([1, 0, 0, 1], np.uint8)
        #octave:4> log(0.5 * 0.8 * 0.7 * 0.7)
        #ans = -1.62964061975162
        self.assertAlmostEqual(bernoulli.log_likelihood(x), -1.62964061975162)
        ##change x
        x = np.array([1, 1, 1, 0], np.uint8)
        #octave:5> log(0.5 * 0.2 * 0.3 * 0.3)
        #ans = -4.71053070164592
        self.assertAlmostEqual(bernoulli.log_likelihood(x), -4.71053070164592)
        ##change eta
        p = np.array([0.3, 0.1, 0.1, 0.2])
        eta = np.log(np.divide(p, 1 - p))
        bernoulli.eta_ = eta
        assert_array_almost_equal(bernoulli.eta_, eta)
        x = np.array([1, 0, 0, 1], np.uint8)
        #octave:10> log(0.3 * 0.9 * 0.9 * 0.2)
        #ans = -3.02413174807569
        self.assertAlmostEqual(bernoulli.log_likelihood(x), -3.02413174807569)

        # test with nvar = 3
        nvar = 3
        p = np.array([0.5, 0.2, 0.9])
        eta = np.log(np.divide(p, 1 - p))
        bernoulli = Bernoulli(nvar)
        bernoulli.eta_ = eta
        assert_array_almost_equal(bernoulli.eta_, eta)
        x = np.array([1, 1, 1], dtype=np.uint8)
        #octave:11> log(0.5 * 0.2 * 0.9)
        #ans = -2.40794560865187
        self.assertAlmostEqual(bernoulli.log_likelihood(x), -2.40794560865187)

    def test_log_likelihood_many_dimensions(self):
        nvar = 64
        p = 0.5 * np.ones((nvar))
        eta = np.log(np.divide(p, 1 - p))
        bernoulli = Bernoulli(nvar)
        bernoulli.eta_ = eta
        assert_array_almost_equal(bernoulli.eta_, eta)
        x = np.ones((nvar), dtype=np.uint8)
        #octave:12> 64 * log(0.5)
        #ans = -44.3614195558365
        self.assertAlmostEqual(bernoulli.log_likelihood(x), -44.3614195558365)

    def test_sample_check_dimensions(self):
        nvar = 4
        p = np.array([0.99, 0.75, 0.75, 0.3])
        eta = np.log(np.divide(p, 1 - p))
        bernoulli = Bernoulli(nvar)
        bernoulli.eta_ = eta
        assert_array_almost_equal(bernoulli.eta_, eta)
        # nsamples = 1
        nsamples = 1
        samples = bernoulli.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 1)
        self.assertEqual(len(samples), nvar)
        # nsamples > 1
        nsamples = 10000
        samples = bernoulli.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 2)
        self.assertEqual(len(samples), nsamples)
        self.assertEqual(len(samples[0]), nvar)

    def test_sample(self):
        nvar = 4
        nsamples = int(10000)
        p = np.array([0.5, 0.75, 0.35, 0.1])
        eta = np.log(np.divide(p, 1 - p))
        bernoulli = Bernoulli(nvar)
        bernoulli.eta_ = eta
        assert_array_almost_equal(bernoulli.eta_, eta)
        samples = bernoulli.sample(nsamples)
        sample_mean = np.mean(samples, axis=0)
        expected_mean = p
        self.assertTrue(norm(sample_mean - expected_mean) < 0.01 * nvar)

    def test_fit_with_one_sample(self):
        nvar = 4
        sample = np.array([[1, 0, 0, 1]], dtype=np.uint8)
        min_phit = 1e-5
        p = np.array([1 - min_phit, min_phit, min_phit, 1 - min_phit])
        eta = np.log(np.divide(p, 1 - p))

        estimator = Bernoulli(nvar=nvar, min_phit=min_phit)
        estimator.fit(sample)

        assert_array_almost_equal(estimator.eta_, eta)

    def test_fit(self):
        nvar = 4
        nsamples = int(10000)
        gt_eta = np.array([0.3, 0.75, 0.6, 0.9])
        bernoulli = Bernoulli(nvar)
        bernoulli._random_state.seed(0)
        bernoulli.eta_ = gt_eta
        samples = bernoulli.sample(nsamples)
        self.assertEqual(nsamples, len(samples))

        estimator = Bernoulli(nvar=nvar)
        init_eta = np.array([0.5, 0.5, 0.5, 0.5])
        estimator.eta_ = init_eta
        estimator._random_state.seed(0)
        estimator.fit(samples)
        est_eta = estimator.eta_
        self.assertTrue(norm(est_eta - gt_eta) < 0.05 * norm(gt_eta))


if __name__ == '__main__':
    unittest.main()
