#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
sys.path.append('.')
import unittest
import numpy as np
from numpy.linalg import norm
from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal

from vrl.ml import Bernoulli
from vrl.ml import BernoulliMixture


class TestBernoulliMixture(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        return

    def test_init(self):
        nvar = 2
        ncomp = 2
        mixture = BernoulliMixture(ncomp=ncomp, nvar=nvar)
        assert_array_almost_equal(mixture.etas_, np.array([[0.0, 0.0], [0.0, 0.0]]))
        assert_array_equal(mixture.weights_, np.array([0.5, 0.5]))
        weights = np.array([0.8, 0.2])
        etas = np.array([[1.0, 2.0], [3.0, 4.0]])
        mixture = BernoulliMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        self.assertAlmostEqual(mixture.min_pst, 1e-4)
        self.assertAlmostEqual(mixture.min_phit, 1e-6)
        assert_array_almost_equal(mixture.etas_, etas)
        assert_array_equal(mixture.weights_, weights)
        weights = np.array([1.6, 0.4])
        mixture = BernoulliMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        weights = np.array([0.8, 0.2])
        assert_array_equal(mixture.weights_, weights)

    def test_min_phit(self):
        nvar = 2
        ncomp = 2
        min_phit = 1e-3
        mixture = BernoulliMixture(ncomp=ncomp, nvar=nvar, min_phit=min_phit)
        self.assertEqual(mixture.min_phit, min_phit)
        for i in range(ncomp):
            self.assertEqual(mixture._component(i).min_phit, min_phit)
        min_phit = 1e-5
        mixture.min_phit = min_phit
        self.assertEqual(mixture.min_phit, min_phit)
        for i in range(ncomp):
            self.assertEqual(mixture._component(i).min_phit, min_phit)

    def test_component(self):
        nvar = 2
        ncomp = 2
        weights = np.array([0.8, 0.2])
        etas = np.array([[1.0, 2.0], [3.0, 4.0]])
        mixture = BernoulliMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        assert_array_almost_equal(mixture.etas_, etas)
        assert_array_equal(mixture.weights_, weights)
        for i in range(ncomp):
            component = mixture._component(i)
            assert_array_almost_equal(component.eta_, etas[i])

    def test_fit_one_component(self):
        np.random.seed(0)
        ncomp = 1
        nvar = 4
        nsamples = 10000
        # sample from a laplacian
        eta = np.array([-3.0, 1.0, 3.0, -1.0])
        bernoulli = Bernoulli(nvar)
        bernoulli.eta_ = eta
        bernoulli._random_state.seed(0)
        samples = bernoulli.sample(nsamples)

        model = BernoulliMixture(ncomp=ncomp, nvar=nvar)
        model._random_state.seed(0)
        model.compute_initial_estimate(samples)
        model.fit(samples)
        est_eta = model.etas_[0]
        self.assertTrue(norm(est_eta - eta) < 0.05 * norm(eta))

    def test_fit_two_components(self):
        np.random.seed(0)
        ncomp = 2
        nvar = 4
        nsamples = 10000
        weights = np.array([0.1, 0.9])
        nsamples_1 = int(nsamples * weights[0])
        nsamples_2 = int(nsamples * weights[1])
        # sample from a bernoulli
        etas = np.array([[5.0, -1.0, -5.0, 1.0], [-0.75, 4.0, 0.125, -5.0]])
        bernoulli = Bernoulli(nvar)
        bernoulli.eta_ = etas[0]
        bernoulli._random_state.seed(0)
        samples1 = bernoulli.sample(nsamples_1)
        # sample from a bernoulli
        bernoulli = Bernoulli(nvar)
        bernoulli.eta_ = etas[1]
        bernoulli._random_state.seed(0)
        samples2 = bernoulli.sample(nsamples_2)

        samples = np.concatenate((samples1, samples2))
        np.random.shuffle(samples)

        model = BernoulliMixture(ncomp=ncomp, nvar=nvar)
        model._random_state.seed(0)
        model.compute_initial_estimate(samples)
        model.fit(samples)
        eta_0 = model.etas_[0]
        eta_1 = model.etas_[1]
        print('nsamples = ', samples.shape[0])
        print('real weights = ', weights)
        print('model.weights_ = ', model.weights_)
        print('real eta_0 = ', etas[0])
        print('real eta_1 = ', etas[1])
        print('eta_0 = ', eta_0)
        print('eta_1 = ', eta_1)

    def test_get_params(self):
        nvar = 2
        ncomp = 2
        weights = np.array([0.8, 0.2])
        etas = np.array([[1.0, 2.0], [3.0, 4.0]])
        mixture = BernoulliMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        params = {'ncomp': 2, 'nvar': 2, 'min_pst': 1e-4, 'min_phit': 1e-6}
        mixture_params = mixture.get_params()
        assert_array_almost_equal(mixture_params.pop('etas_'), etas)
        assert_array_equal(mixture_params.pop('weights_'), weights)
        self.assertEqual(mixture_params, params)

if __name__ == '__main__':
    unittest.main()
