#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
sys.path.append('.')
import unittest
import numpy as np
from numpy.linalg import norm
from numpy.testing import assert_array_equal

from vrl.ml import Dirichlet


class TestDirichlet(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        return

    def test_log_likelihood(self):
        nvar = 2
        alpha = np.array([0.5, 1.0])
        eta = alpha - 1.0
        dirichlet = Dirichlet(nvar)
        dirichlet.eta_ = eta
        self.assertEqual(dirichlet.npar, nvar)
        assert_array_equal(dirichlet.eta_, eta)
        x = np.array([0.8, 0.2])
        self.assertAlmostEqual(dirichlet.log_likelihood(x),
                               -0.58157540490284054)
        #change x
        x = np.array([0.4, 0.6])
        self.assertAlmostEqual(dirichlet.log_likelihood(x),
                               -0.23500181462286801)
        #change eta
        alpha = np.array([1.0, 3.0])
        eta = alpha - 1.0
        dirichlet.eta_ = eta
        assert_array_equal(dirichlet.eta_, eta)
        x = np.array([0.9, 0.1])
        self.assertAlmostEqual(dirichlet.log_likelihood(x),
                               -3.5065578973199814)

    def test_sample_check_dimensions(self):
        nvar = 2
        eta = np.array([1.5, 0.75])
        dirichlet = Dirichlet(nvar)
        dirichlet.eta_ = eta
        # nsamples = 1
        nsamples = 1
        samples = dirichlet.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 1)
        self.assertEqual(len(samples), nvar)
        self.assertAlmostEquals(np.sum(samples), 1.0)
        # nsamples > 1
        nsamples = 10000
        samples = dirichlet.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 2)
        self.assertEqual(len(samples), nsamples)
        self.assertEqual(len(samples[0]), nvar)
        self.assertAlmostEquals(np.sum(np.sum(samples)), nsamples)

    def test_sample(self):
        nvar = 2
        nsamples = int(10000)
        alpha = np.array([1.5, 0.75])
        eta = alpha - 1.0
        dirichlet = Dirichlet(nvar)
        dirichlet.eta_ = eta
        assert_array_equal(dirichlet.eta_, eta)
        samples = dirichlet.sample(nsamples)
        sample_mean = np.mean(samples, axis=0)
        expected_mean = alpha / sum(alpha)
        self.assertTrue(norm(sample_mean - expected_mean) < (0.01 * nvar *
                        norm(expected_mean)))

    def test_fit(self):
        nvar = 2
        nsamples = int(5000)
        gt_eta = np.array([1.5, 0.75])
        dirichlet = Dirichlet(nvar)
        dirichlet._random_state.seed(0)
        dirichlet.eta_ = gt_eta

        assert_array_equal(dirichlet.eta_, gt_eta)
        samples = dirichlet.sample(nsamples)

        estimator = Dirichlet(nvar=nvar)
        estimator.fit(samples)
        est_eta = estimator.eta_
        self.assertTrue(norm(est_eta - gt_eta) < (0.05 * norm(gt_eta)))


if __name__ == '__main__':
    unittest.main()
