#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
sys.path.append('.')
import unittest
import numpy as np
from numpy.linalg import norm
from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal

from vrl.ml import Dirichlet
from vrl.ml import DirichletMixture


class TestDirichletMixture(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        return

    def test_init(self):
        nvar = 2
        ncomp = 2
        mixture = DirichletMixture(ncomp=ncomp, nvar=nvar)
        assert_array_almost_equal(mixture.etas_, np.array([[-1.0, -1.0], [-1.0, -1.0]]))
        assert_array_equal(mixture.weights_, np.array([0.5, 0.5]))
        weights = np.array([0.8, 0.2])
        etas = np.array([[1.0, 2.0], [3.0, 4.0]])
        mixture = DirichletMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        self.assertAlmostEqual(mixture.min_pst, 1e-4)
        assert_array_equal(mixture.etas_, etas)
        assert_array_equal(mixture.weights_, weights)
        weights = np.array([1.6, 0.4])
        mixture = DirichletMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        weights = np.array([0.8, 0.2])
        assert_array_equal(mixture.weights_, weights)

    def test_component(self):
        nvar = 2
        ncomp = 2
        weights = np.array([0.8, 0.2])
        etas = np.array([[1.0, 2.0], [3.0, 4.0]])
        mixture = DirichletMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        assert_array_equal(mixture.etas_, etas)
        assert_array_equal(mixture.weights_, weights)
        for i in range(ncomp):
            component = mixture._component(i)
            assert_array_equal(component.eta_, etas[i])

    def test_fit_one_component(self):
        ncomp = 1
        nvar = 2
        nsamples = 10000
        # sample from a dirichlet
        eta = np.array([0.5, 1.0])
        dirichlet = Dirichlet(nvar)
        dirichlet._random_state.seed(0)
        dirichlet.eta_ = eta
        samples = dirichlet.sample(nsamples)

        model = DirichletMixture(ncomp=ncomp, nvar=nvar)
        model._random_state.seed(0)
        model.compute_initial_estimate(samples)
        model.fit(samples)
        est_eta = model.etas_[0]
        self.assertTrue(norm(est_eta - eta) < 0.1 * norm(eta))

    def test_fit_two_components(self):
        ncomp = 2
        nvar = 2
        nsamples = 10000
        nsamples_1 = int(nsamples * 0.1)
        nsamples_2 = int(nsamples * 0.9)
        # sample from a dirichlet
        etas = np.array([[5.0, 1.0], [0.75, 4.0]])
        dirichlet = Dirichlet(nvar)
        dirichlet.eta_ = etas[0]
        dirichlet._random_state.seed(0)
        samples1 = dirichlet.sample(nsamples_1)
        # sample from a dirichlet
        dirichlet = Dirichlet(nvar)
        dirichlet.eta_ = etas[1]
        dirichlet._random_state.seed(0)
        samples2 = dirichlet.sample(nsamples_2)

        samples = np.concatenate((samples1, samples2))
        np.random.shuffle(samples)

        weights = np.array([0.5, 0.5])
        model = DirichletMixture(ncomp=ncomp, nvar=nvar)
        model.weights_ = weights
        model._random_state.seed(0)
        model.compute_initial_estimate(samples)
        model.fit(samples)
        #eta_0 = model.etas_[0]
        #eta_1 = model.etas_[1]
        #print('nsamples = ', samples.shape[0])
        #print('real weights = ', weights)
        #print('model.weights_ = ', model.weights_)
        #print('real eta_0 = ', etas[0])
        #print('real eta_1 = ', etas[1])
        #print('eta_0 = ', eta_0)
        #print('eta_1 = ', eta_1)

    def test_get_params(self):
        nvar = 2
        ncomp = 2
        weights = np.array([0.8, 0.2])
        etas = np.array([[1.0, 2.0], [3.0, 4.0]])
        mixture = DirichletMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        mdict = {'ncomp': 2, 'nvar': 2, 'min_pst': 0.0001}
        params = mixture.get_params()
        assert_array_equal(params.pop('etas_'), etas)
        assert_array_equal(params.pop('weights_'), weights)
        self.assertEqual(params, mdict)


if __name__ == '__main__':
    unittest.main()
