#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
sys.path.append('.')
import unittest
import numpy as np
from numpy.linalg import norm
from numpy.testing import assert_array_equal

from vrl.ml import Gaussian


class TestGaussian(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        return

    def test_min_var(self):
        nvar = 2
        var = 1e-8
        mean = np.array([1.0, 1.0])
        cov = np.array([var, var])
        eta = np.concatenate((np.divide(mean, cov), -0.5 * np.reciprocal(cov)))
        min_var = 1e-7
        min_cov = np.array([min_var, min_var])
        min_eta = np.concatenate((np.divide(mean, min_cov), -0.5 * np.reciprocal(min_cov)))
        gaussian = Gaussian(nvar, min_var=min_var)
        self.assertEqual(gaussian.npar, 2 * nvar)
        self.assertEqual(gaussian.min_var, min_var)
        gaussian.eta_ = eta
        assert_array_equal(gaussian.eta_, min_eta)

        min_var = 1e-6
        min_cov = np.array([min_var, min_var])
        min_eta = np.concatenate((np.divide(mean, min_cov), -0.5 * np.reciprocal(min_cov)))
        gaussian.min_var = min_var
        self.assertEqual(gaussian.min_var, min_var)
        assert_array_equal(gaussian.eta_, min_eta)

    def test_source_params(self):
        nvar = 2
        eta = np.array([1, 3, -0.25, -0.5])
        gaussian = Gaussian(nvar)
        gaussian.eta_ = eta
        mean = np.array([2.0, 3.0])
        cov = np.array([2.0, 1.0])
        for i in range(len(gaussian.mu)):
            self.assertAlmostEqual(mean[i], gaussian.mu[i])
            self.assertAlmostEqual(cov[i], gaussian.sigma[i]**2.)
            self.assertAlmostEqual(np.sqrt(cov[i]), gaussian.sigma[i])

    def test_log_likelihood(self):
        nvar = 2
        eta = np.array([-5.0, 1.0, -5.0, -1.0])
        gaussian = Gaussian(nvar)
        gaussian.eta_ = eta
        x = np.array([1.0, -0.5])
        #octave:16> mvnpdf([1, -0.5], [-0.5, 0.5], [[0.1, 0];[0, 0.5]])
        #ans =  3.40586732557151e-06
        self.assertAlmostEqual(np.exp(gaussian.log_likelihood(x)),
                               3.40586732557e-06)
        #change x
        x = np.array([-1.0, 0.0])
        #octave:19> mvnpdf([-1, 0], [-0.5, 0.5], [[0.1, 0];[0, 0.5]])
        #ans =  0.158815690300328
        self.assertAlmostEqual(np.exp(gaussian.log_likelihood(x)),
                               0.158815690300328)
        #change eta
        eta = np.array([1, 3, -0.25, -0.5])
        gaussian.eta_ = eta
        x = np.array([1.0, 0.0])
        #octave:25> mvnpdf([1, 0], [2, 3], [[2, 0];[0, 1]])
        #ans =  9.73657794223459e-04
        self.assertAlmostEqual(np.exp(gaussian.log_likelihood(x)),
                               9.73657794223459e-04)

    def test_sample_check_dimensions(self):
        nvar = 2
        eta = np.array([-1.5, -0.75, -0.75, -10.0])
        gaussian = Gaussian(nvar)
        gaussian.eta_ = eta
        # nsamples = 1
        nsamples = 1
        samples = gaussian.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 1)
        self.assertEqual(len(samples), nvar)
        # nsamples > 1
        nsamples = 10000
        samples = gaussian.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 2)
        self.assertEqual(len(samples), nsamples)
        self.assertEqual(len(samples[0]), nvar)

    def test_sample(self):
        nvar = 2
        nsamples = int(10000)
        eta = np.array([-1.5, -0.75, -0.75, -10.0])
        gaussian = Gaussian(nvar)
        gaussian.eta_ = eta
        samples = gaussian.sample(nsamples)
        sample_mean = (sum(samples) / nsamples).flatten()
        expected_mean = gaussian.mu
        self.assertTrue(norm(sample_mean - expected_mean) < 0.1)

    def test_fit(self):
        nvar = 2
        nsamples = int(5000)
        gt_eta = np.array([-1.5, -0.75, -0.75, -1.0])
        gaussian = Gaussian(nvar)
        gaussian._random_state.seed(0)
        gaussian.eta_ = gt_eta
        samples = gaussian.sample(nsamples)

        estimator = Gaussian(nvar=nvar)
        estimator._random_state.seed(0)
        init_eta = np.array([-1.0, -0.5, -0.5, -1.0])
        estimator.eta_ = init_eta

        estimator.fit(samples)
        est_eta = estimator.eta_
        self.assertTrue(norm(est_eta - gt_eta) < 0.05 * norm(gt_eta))


if __name__ == '__main__':
    unittest.main()
