#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
sys.path.append('.')
import unittest
import numpy as np
from numpy.linalg import norm
from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal

from vrl.ml import Laplace


class TestLaplace(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        return

    def test_min_scale(self):
        nvar = 2
        scale = 1e-8
        eta = -1.0 / np.array([scale, scale])
        min_scale = 1e-7
        min_eta = -1.0 / np.array([min_scale, min_scale])
        laplace = Laplace(nvar, min_scale=min_scale)
        self.assertEqual(laplace.npar, nvar)
        self.assertEqual(laplace.min_scale, min_scale)
        laplace.eta_ = eta
        assert_array_equal(laplace.eta_, min_eta)
        min_scale = 1e-6
        min_eta = -1.0 / np.array([min_scale, min_scale])
        laplace.min_scale = min_scale
        self.assertEqual(laplace.min_scale, min_scale)
        assert_array_equal(laplace.eta_, min_eta)

    def test_set_source_params(self):
        nvar = 2
        scale = np.array([5.0, 0.5])
        eta = -1.0 / scale
        laplace = Laplace(nvar)
        laplace.eta_ = eta
        assert_array_almost_equal(laplace.scale, scale)

    def test_log_likelihood(self):
        nvar = 2
        scale = np.array([5.0, 0.5])
        eta = -1.0 / scale
        laplace = Laplace(nvar)
        laplace.eta_ = eta
        assert_array_equal(laplace.eta_, eta)
        x = np.array([1.0, -0.5])
        #octave:12> sum(log(laplace_pdf([-1.0, 0.4]./[5.0, 0.5])./[5.0, 0.5]))
        #ans = -3.50258509299405
        self.assertAlmostEqual(laplace.log_likelihood(x), -3.50258509299405)
        #change x
        x = np.array([-1.0, 0.4])
        #octave:14> sum(log(laplace_pdf([-1.0, 0.4]./[5.0, 0.5])./[5.0, 0.5]))
        #ans = -3.30258509299405
        self.assertAlmostEqual(laplace.log_likelihood(x), -3.30258509299405)
        #change eta
        scale = np.array([1.0, 3.0])
        eta = -1.0 / scale
        laplace.eta_ = eta
        assert_array_equal(laplace.eta_, eta)
        x = np.array([1.0, 0.0])
        #octave:15> sum(log(laplace_pdf([1.0, 0.0]./[1.0, 3.0])./[1.0, 3.0]))
        #ans = -3.48490664978800
        self.assertAlmostEqual(laplace.log_likelihood(x), -3.48490664978800)

    def test_sample_check_dimensions(self):
        nvar = 2
        eta = np.array([-1.5, -0.75])
        laplace = Laplace(nvar)
        laplace.eta_ = eta
        # nsamples = 1
        nsamples = 1
        samples = laplace.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 1)
        self.assertEqual(len(samples), nvar)
        # nsamples > 1
        nsamples = 10000
        samples = laplace.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 2)
        self.assertEqual(len(samples), nsamples)
        self.assertEqual(len(samples[0]), nvar)

    def test_sample(self):
        nvar = 2
        nsamples = int(10000)
        scale = np.array([5.0, 0.75])
        eta = -1.0 / scale
        laplace = Laplace(nvar)
        laplace._random_state.seed(0)
        laplace.eta_ = eta
        samples = laplace.sample(nsamples)
        sample_mean = np.mean(samples, axis=0)
        self.assertTrue(norm(sample_mean) < 0.1 * nvar)
        sample_var = np.var(samples, axis=0)
        exp_var = 2 * np.multiply(scale, scale)
        self.assertTrue(norm(sample_var - exp_var) < 0.1 * norm(exp_var))

    def test_fit(self):
        nvar = 2
        nsamples = int(10000)
        gt_eta = np.array([-1.5, -0.75])
        laplace = Laplace(nvar)
        laplace._random_state.seed(0)
        laplace.eta_ = gt_eta
        samples = laplace.sample(nsamples)

        estimator = Laplace(nvar=nvar)
        estimator._random_state.seed(0)
        init_eta = np.array([-1.0, -0.5])
        estimator.eta_ = init_eta

        estimator.fit(samples)
        est_eta = estimator.eta_
        self.assertTrue(norm(est_eta - gt_eta) < 0.05 * norm(gt_eta))


if __name__ == '__main__':
    unittest.main()
