#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
sys.path.append('.')
import unittest
import numpy as np
from numpy.linalg import norm
from numpy.testing import assert_array_equal

from vrl.ml import Laplace
from vrl.ml import LaplaceMixture


class TestLaplaceMixture(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        return

    def test_init(self):
        nvar = 2
        ncomp = 2
        mixture = LaplaceMixture(ncomp=ncomp, nvar=nvar)
        assert_array_equal(mixture.etas_, np.array([[-1.0, -1.0],
                                                   [-1.0, -1.0]]))
        assert_array_equal(mixture.weights_, np.array([0.5, 0.5]))
        weights = np.array([0.8, 0.2])
        etas = np.array([[-1.0, -2.0], [-3.0, -4.0]])
        mixture = LaplaceMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        assert_array_equal(mixture.etas_, etas)
        assert_array_equal(mixture.weights_, weights)
        weights = np.array([1.6, 0.4])
        mixture = LaplaceMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        weights = np.array([0.8, 0.2])
        assert_array_equal(mixture.weights_, weights)

    def test_min_scale(self):
        nvar = 2
        ncomp = 2
        mixture = LaplaceMixture(ncomp=ncomp, nvar=nvar)
        self.assertEqual(mixture.min_scale, 1e-7)
        min_scale = 1e-3
        mixture = LaplaceMixture(ncomp=ncomp, nvar=nvar, min_scale=min_scale)
        self.assertEqual(mixture.min_scale, min_scale)
        for i in range(ncomp):
            self.assertEqual(mixture._component(i).min_scale, min_scale)
        min_scale = 1e-5
        mixture.min_scale = min_scale
        self.assertEqual(mixture.min_scale, min_scale)
        for i in range(ncomp):
            self.assertEqual(mixture._component(i).min_scale, min_scale)

    def test_fit_one_component(self):
        ncomp = 1
        nvar = 4
        nsamples = 100000
        # sample from a laplacian
        eta = np.array([-3.0, -1.0, -3.0, -1.0])
        laplace = Laplace(nvar)
        laplace.eta_ = eta
        laplace._random_state.seed(0)
        samples = laplace.sample(nsamples)

        model = LaplaceMixture(ncomp=ncomp, nvar=nvar)
        model.compute_initial_estimate(samples)
        model.fit(samples)
        est_eta = model.etas_[0]
        self.assertTrue(norm(est_eta - eta) < 0.1 * norm(eta))

    def test_fit_two_components(self):
        ncomp = 2
        nvar = 4
        nsamples = 10000
        weights = np.array([0.1, 0.9])
        nsamples_1 = int(nsamples * weights[0])
        nsamples_2 = int(nsamples * weights[1])
        # sample from a laplace
        etas = np.array([[-5.0, -1.0, -5.0, -1.0],
                         [-0.75, -40.0, -0.125, -5.0]])
        laplace = Laplace(nvar)
        laplace.eta_ = etas[0]
        laplace._random_state.seed(0)
        samples1 = laplace.sample(nsamples_1)
        # sample from a laplace
        laplace = Laplace(nvar)
        laplace.eta_ = etas[1]
        laplace._random_state.seed(0)
        samples2 = laplace.sample(nsamples_2)

        samples = np.concatenate((samples1, samples2))
        np.random.seed(0)
        np.random.shuffle(samples)

        model = LaplaceMixture(ncomp=ncomp, nvar=nvar)
        model.compute_initial_estimate(samples)
        model.fit(samples)
        #eta_0 = model.etas_[0]
        #eta_1 = model.etas_[1]
        #print('nsamples = ', samples.shape[0])
        #print('real weights = ', weights)
        #print('model.weights = ', model.weights)
        #print('real eta_0 = ', etas[0])
        #print('real eta_1 = ', etas[1])
        #print('eta_0 = ', eta_0)
        #print('eta_1 = ', eta_1)

    def test_get_params(self):
        nvar = 2
        ncomp = 2
        weights = np.array([0.8, 0.2])
        etas = np.array([[-1.0, -2.0], [-3.0, -4.0]])
        mixture = LaplaceMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        params = {'ncomp': 2, 'nvar': 2, 'min_pst': 0.0001, 'min_scale': 1e-7}
        mixture_params = mixture.get_params()
        assert_array_equal(mixture_params.pop('etas_'), etas)
        assert_array_equal(mixture_params.pop('weights_'), weights)
        self.assertEqual(mixture_params, params)

if __name__ == '__main__':
    unittest.main()
