#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
sys.path.append('.')
import unittest
import numpy as np
from numpy.linalg import norm
from numpy.testing import assert_array_equal, assert_array_almost_equal

from vrl.ml import Multinomial


class TestMultinomial(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        return

    def test_min_p(self):
        nvar = 2
        ntrials = 4
        min_p = 1e-7
        multinomial = Multinomial(nvar=nvar, ntrials=ntrials, min_p=min_p)
        self.assertEqual(multinomial.npar, nvar)
        self.assertEqual(multinomial.min_p, min_p)

        p = 1e-8

        min_eta = np.log([min_p, 1.0 - min_p])
        eta = np.log([p, 1.0 - p])
        multinomial.eta_ = eta
        assert_array_almost_equal(multinomial.eta_, min_eta)

        nvar = 3
        multinomial = Multinomial(nvar=nvar, ntrials=ntrials, min_p=min_p)
        min_eta = np.log([min_p, min_p, 1.0 - 2 * min_p])
        eta = np.log([p, p, 1.0 - 2 * p])
        multinomial.eta_ = eta
        assert_array_almost_equal(multinomial.eta_, min_eta)

    def test_set_min_p(self):
        nvar = 2
        ntrials = 4
        min_p = 1e-8
        multinomial = Multinomial(nvar=nvar, ntrials=ntrials, min_p=min_p)
        self.assertEqual(multinomial.npar, nvar)
        self.assertEqual(multinomial.min_p, min_p)

        min_eta = np.log([min_p, 1.0 - min_p])
        multinomial.eta_ = min_eta
        assert_array_almost_equal(multinomial.eta_, min_eta)

        new_min_p = 1e-7
        multinomial.min_p = new_min_p
        new_min_eta = np.log([new_min_p, 1.0 - new_min_p])
        assert_array_almost_equal(multinomial.eta_, new_min_eta)

    def test_source_params(self):
        nvar = 2
        ntrials = 4
        p = np.array([0.25, 0.75])
        eta = np.log(p)
        multinomial = Multinomial(nvar, ntrials)
        multinomial.eta_ = eta
        assert_array_equal(multinomial.p_, p)

    def test_log_likelihood(self):
        #n = ntrials
        #x = (x_1, ..., x_nvar)
        #pmf = (n!/(x_1!*...*x_nvar!)) * (p_1^x_1) * ... * (p_nvar^x_nvar)
        nvar = 2
        ntrials = 4
        eta = np.log([0.25, 0.75])
        multinomial = Multinomial(nvar, ntrials)
        multinomial.eta_ = eta
        x = np.array([2, 2], dtype=np.uint32)
        #octave-cli:6> log(mnpdf([2, 2], [0.25, 0.75]))
        #ans = -1.55619339791529
        self.assertAlmostEqual(multinomial.log_likelihood(x),
                               -1.55619339791529)
        #change x
        x = np.array([1, 3], dtype=np.uint32)
        #octave-cli:3> log(mnpdf([1, 3], [0.25, 0.75]))
        #ans = -0.863046217355342
        self.assertAlmostEqual(multinomial.log_likelihood(x),
                               -0.863046217355342)
        #change eta
        eta = np.log([0.8, 0.2])
        multinomial.eta_ = eta
        x = np.array([0, 4], dtype=np.uint32)
        #octave-cli:4> log(mnpdf([0, 4], [0.8, 0.2]))
        #ans = -6.43775164973640
        self.assertAlmostEqual(multinomial.log_likelihood(x),
                               -6.43775164973640)

    def test_sample_check_dimensions(self):
        nvar = 2
        ntrials = 4
        eta = np.log([0.4, 0.6])
        multinomial = Multinomial(nvar, ntrials)
        multinomial.eta_ = eta
        # nsamples = 1
        nsamples = 1
        samples = multinomial.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 1)
        self.assertEqual(len(samples), nvar)
        self.assertEqual(np.sum(samples), ntrials)
        # nsamples > 1
        nsamples = 10000
        samples = multinomial.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 2)
        self.assertEqual(len(samples), nsamples)
        self.assertEqual(len(samples[0]), nvar)

    def test_sample(self):
        nvar = 2
        ntrials = 4
        nsamples = int(1000)
        eta = np.log([0.2, 0.8])
        multinomial = Multinomial(nvar, ntrials)
        multinomial.eta_ = eta
        multinomial._random_state.seed(0)
        samples = multinomial.sample(nsamples)
        samples = np.array(samples, dtype=np.float32)
        sample_mean = (np.sum(samples, axis=0) / nsamples).flatten()
        expected_mean = multinomial.p_ * ntrials
        self.assertTrue(norm(sample_mean - expected_mean) < 0.05)

    def test_fit(self):
        nvar = 2
        ntrials = 4
        nsamples = int(10000)
        gt_eta = np.log([0.4, 0.6])
        multinomial = Multinomial(nvar, ntrials)
        multinomial._random_state.seed(0)
        multinomial.eta_ = gt_eta
        samples = multinomial.sample(nsamples)

        estimator = Multinomial(nvar, ntrials)
        estimator._random_state.seed(0)
        estimator.fit(samples)
        est_eta = estimator.eta_
        self.assertTrue(norm(est_eta - gt_eta) < 0.05)


if __name__ == '__main__':
    unittest.main()
