#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
sys.path.append('.')
import unittest
import numpy as np
from numpy.linalg import norm
from numpy.testing import assert_array_equal

from vrl.ml import Poisson


class TestPoisson(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        return

    def test_min_lambda(self):
        nvar = 2
        lam = 1e-8
        eta = np.log(np.array([lam, lam]))
        min_lam = 1e-7
        min_eta = np.log(np.array([min_lam, min_lam]))
        poisson = Poisson(nvar, min_lambda=min_lam)
        self.assertEqual(poisson.npar, nvar)
        self.assertAlmostEqual(poisson.min_lambda, min_lam)
        poisson.eta_ = eta
        assert_array_equal(poisson.eta_, min_eta)

        min_lam = 1e-6
        min_eta = np.log(np.array([min_lam, min_lam]))
        poisson.min_lambda = min_lam
        assert_array_equal(poisson.eta_, min_eta)

    def test_log_likelihood(self):
        nvar = 2
        lam = np.array([2, 1])
        eta = np.log(lam)
        poisson = Poisson(nvar)
        poisson.eta_ = eta
        assert_array_equal(eta, poisson.eta_)

        x = np.array([1, 2], dtype=poisson._dtype)
        #octave:7> log(prod(poisspdf([1,2], [2,1])))
        #ans = -3
        self.assertAlmostEqual(poisson.log_likelihood(x), -3)

        x = np.array([3, 5], dtype=poisson._dtype)
        #octave:10> log(prod(poisspdf([3,5], [2,1])))
        #ans = -7.49980967033027
        self.assertAlmostEqual(poisson.log_likelihood(x), -7.49980967033027)

        lam = np.array([0.1, 7])
        eta = np.log(lam)
        poisson.eta_ = eta
        assert_array_equal(eta, poisson.eta_)
        #octave:12> log(prod(poisspdf([3,5], [0.1,7])))
        #ans = -10.8574557457157
        self.assertAlmostEqual(poisson.log_likelihood(x), -10.8574557457157)

        nvar = 3
        lam = np.array([2.3, 1.9, 3.5])
        eta = np.log(lam)
        poisson = Poisson(nvar)
        poisson.eta_ = eta
        assert_array_equal(eta, poisson.eta_)

        x = np.array([5, 2, 10], dtype=poisson._dtype)
        #octave:18> log(prod(poisspdf([5, 2, 10], [2.3, 1.9, 3.5])))
        #ans = -10.3091684244435
        self.assertAlmostEqual(poisson.log_likelihood(x), -10.3091684244435)

        nvar = 20
        lam = np.array([2.3, 1.9, 3.5, 10.9, 2.3, 5.6, 19.4, 32.0, 12.2, 2.4,
                        2.3, 1.9, 3.5, 10.9, 2.3, 5.6, 19.4, 32.0, 12.2, 2.4])
        eta = np.log(lam)
        poisson = Poisson(nvar)
        poisson.eta_ = eta
        assert_array_equal(eta, poisson.eta_)

        x = np.array([5, 2, 10, 3, 6, 7, 9, 11, 20, 10,
                      5, 2, 10, 3, 6, 7, 9, 11, 20, 10], dtype=poisson._dtype)
        #octave:21> log(prod(poisspdf([5, 2, 10, 3, 6, 7, 9, 11, 20, 10,
                                      #5, 2, 10, 3, 6, 7, 9, 11, 20, 10],
                                     #[2.3, 1.9, 3.5, 10.9, 2.3, 5.6, 19.4,
                                      #32.0, 12.2, 2.4, 2.3, 1.9, 3.5, 10.9,
                                      #2.3, 5.6, 19.4, 32.0, 12.2, 2.4])))
        #ans = -103.864873206568
        self.assertAlmostEqual(poisson.log_likelihood(x), -103.864873206568)

    def test_sample_check(self):
        nvar = 4
        eta = np.array([1.5, 0.75, 0.75, 10.0])
        poisson = Poisson(nvar)
        poisson.eta_ = eta
        # nsamples = 1
        nsamples = 1
        samples = poisson.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 1)
        self.assertEqual(0, np.sum(samples < 0))
        # nsamples > 1
        nsamples = 10000
        samples = poisson.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 2)
        self.assertEqual(0, np.sum(samples < 0))

    def test_sample(self):
        nvar = 4
        nsamples = int(10000)
        eta = np.array([3.4, 2.75, 3.75, 5.0])
        poisson = Poisson(nvar)
        poisson.eta_ = eta
        samples = poisson.sample(nsamples)
        sample_mean = sum(samples) / nsamples
        diff_norm = norm(sample_mean - poisson.lambda_) / norm(poisson.lambda_)
        self.assertTrue(diff_norm < 0.01)

    def test_fit_with_one_sample(self):
        nvar = 4
        sample = np.array([[1, 5, 4, 7]], dtype=np.uint32)
        eta = np.log(sample[0])

        estimator = Poisson(nvar=nvar)
        estimator.fit(sample)

        assert_array_equal(estimator.eta_, eta)

    def test_fit(self):
        nvar = 4
        nsamples = int(1000)
        gt_eta = np.array([5, 7.5, 2, 1.0])
        poisson = Poisson(nvar)
        poisson._random_state.seed(0)
        poisson.eta_ = gt_eta
        samples = poisson.sample(nsamples)

        estimator = Poisson(nvar=nvar)
        estimator._random_state.seed(0)

        estimator.fit(samples)
        est_eta = estimator.eta_
        self.assertTrue(norm(est_eta - gt_eta) < 0.05 * norm(gt_eta))

if __name__ == '__main__':
    unittest.main()
