#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import print_function

import sys
sys.path.append('.')
import unittest
import numpy as np
from numpy.linalg import norm
from numpy.testing import assert_array_equal

from vrl.ml import Poisson
from vrl.ml import PoissonMixture


class TestPoissonMixture(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        return

    def test_init(self):
        nvar = 2
        ncomp = 2
        mixture = PoissonMixture(ncomp=ncomp, nvar=nvar)
        assert_array_equal(mixture.etas_, np.log(np.array([[1.0, 1.0],
                                                          [1.0, 1.0]])))
        assert_array_equal(mixture.weights_, np.array([0.5, 0.5]))
        weights = np.array([0.8, 0.2])
        etas = np.array([[1.0, 2.0], [3.0, 4.0]])
        mixture = PoissonMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        self.assertAlmostEqual(mixture.min_pst, 1e-4)
        self.assertAlmostEqual(mixture.min_lambda, 1e-7)
        assert_array_equal(mixture.etas_, etas)
        assert_array_equal(mixture.weights_, weights)
        weights = np.array([1.6, 0.4])
        mixture = PoissonMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        weights = np.array([0.8, 0.2])
        assert_array_equal(mixture.weights_, weights)

    def test_min_lambda(self):
        nvar = 2
        ncomp = 2
        min_lambda = 1e-3
        mixture = PoissonMixture(ncomp=ncomp, nvar=nvar, min_lambda=min_lambda)
        self.assertEqual(mixture.min_lambda, min_lambda)
        for i in range(ncomp):
            self.assertEqual(mixture._component(i).min_lambda, min_lambda)
        min_lambda = 1e-5
        mixture.min_lambda = min_lambda
        self.assertEqual(mixture.min_lambda, min_lambda)
        for i in range(ncomp):
            self.assertEqual(mixture._component(i).min_lambda, min_lambda)

    def test_component(self):
        nvar = 2
        ncomp = 2
        weights = np.array([0.8, 0.2])
        etas = np.array([[1.0, 2.0], [3.0, 4.0]])
        mixture = PoissonMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        assert_array_equal(mixture.etas_, etas)
        assert_array_equal(mixture.weights_, weights)
        for i in range(ncomp):
            component = mixture._component(i)
            assert_array_equal(component.eta_, etas[i])

    def test_fit_one_component(self):
        ncomp = 1
        nvar = 4
        nsamples = 1000
        # sample from a poisson
        eta = np.array([5, 2.5, 1.0, 3.0])
        poisson = Poisson(nvar)
        poisson.eta_ = eta
        samples = poisson.sample(nsamples)

        model = PoissonMixture(ncomp=ncomp, nvar=nvar)
        model.compute_initial_estimate(data=samples)
        model.fit(samples)
        est_eta = model.etas_[0]
        self.assertTrue(norm(est_eta - eta) < 0.05 * norm(eta))

    def test_fit_two_components(self):
        ncomp = 2
        nvar = 4
        nsamples = 10000
        weights = np.array([0.1, 0.9])
        nsamples_1 = int(nsamples * weights[0])
        nsamples_2 = int(nsamples * weights[1])
        poisson = Poisson(nvar)
        etas = np.array([[2.3, 1.5, 4.5, 1.0], [2.5, 1.25, 4.25, 2.0]])
        # sample from a poisson
        poisson = Poisson(nvar)
        poisson.eta_ = etas[0]
        samples1 = poisson.sample(nsamples_1)
        # sample from a poisson
        poisson = Poisson(nvar)
        poisson.eta_ = etas[1]
        samples2 = poisson.sample(nsamples_2)

        samples = np.concatenate((samples1, samples2))
        np.random.shuffle(samples)

        model = PoissonMixture(ncomp=ncomp, nvar=nvar)
        model.compute_initial_estimate(data=samples)
        model.fit(samples)
        #eta_0 = model.etas_[0]
        #eta_1 = model.etas_[1]
        #print('nsamples = ', samples.shape[0])
        #print('real weights = ', weights)
        #print('model.weights = ', model.weights_)
        #print('real eta_0 = ', etas[0])
        #print('real eta_1 = ', etas[1])
        #print('model.eta_0 = ', eta_0)
        #print('model.eta_1 = ', eta_1)

    def test_get_params(self):
        nvar = 2
        ncomp = 2
        weights = np.array([0.8, 0.2])
        etas = np.array([[1, 2], [3, 4]], dtype=np.float64)
        mixture = PoissonMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        params = {'ncomp': 2, 'nvar': 2, 'min_pst': 0.0001, 'min_lambda': 1e-7}
        mixture_params = mixture.get_params()
        assert_array_equal(mixture_params.pop('etas_'), etas)
        assert_array_equal(mixture_params.pop('weights_'), weights)
        self.assertEqual(mixture_params, params)


if __name__ == '__main__':
    unittest.main()
