#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
sys.path.append('.')
import unittest
import numpy as np
from numpy.linalg import norm

from vrl.ml import Wishart
from vrl.ml.utils import matrix_is_positive_definite, matrix_is_symmetric
from numpy.testing import assert_array_equal, assert_array_almost_equal


def positive_definite_matrix(ndim):
    mat = np.cov(np.random.random((ndim, 10 * ndim)))
    assert(matrix_is_positive_definite(mat))
    assert(matrix_is_symmetric(mat))
    return mat


class TestWishart(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        return

    def test_log_likelihood(self):
        ndim = 2
        nvar = ndim * ndim
        eta_2 = 0.5
        eta_1 = np.array([-0.5, 0.0, 0.0, -0.5])
        wishart = Wishart(nvar)
        self.assertEqual(wishart.npar, nvar)
        wishart.eta_ = eta_1
        dof = wishart._dtype(2) * eta_2 + wishart._dtype(ndim + 1)
        wishart.dof = dof
        x = np.array([1.0, 0.0, 0.0, 1.0])
        #octave:6> log(wishpdf([1, 0; 0, 1], [1, 0; 0, 1], 4))
        #ans = -4.22417142752924
        self.assertAlmostEqual(wishart.log_likelihood(x), -4.22417142752924)
        #change x
        x = np.array([1.0, 0.5, 0.5, 1.0])
        #octave:8> log(wishpdf([1, 0.5; 0.5, 1], [1, 0; 0, 1], 4))
        #ans = -4.36801246375513
        self.assertAlmostEqual(wishart.log_likelihood(x), -4.36801246375513)
        #change eta_1 and eta_2
        eta_2 = 1.0
        eta_1 = np.array([-0.3, 0.1, 0.1, -0.2])
        wishart.eta_ = eta_1
        dof = wishart._dtype(2) * eta_2 + wishart._dtype(ndim + 1)
        wishart.dof = dof
        x = np.array([1.0, 0.0, 0.0, 1.0])
        #octave:9> log(wishpdf([1, 0; 0, 1], [2, 1; 1, 3], 5))
        #ans = -8.84637849728260
        self.assertAlmostEqual(wishart.log_likelihood(x), -8.84637849728260)

        eta_1 = np.array([-2.6603366666012978, 0.1045192404963062,
                          0.10451924049630622, -3.7683571627978321])
        dof = 4
        wishart.dof = dof
        wishart.eta_ = eta_1
        x = np.array([0.056638211993907318, 0.16580994383294745,
                      0.16580994383294745, 1.9570721313347723])
        """
        octave:16> log(wishpdf([0.056638211993907318, 0.16580994383294745;
                                   0.16580994383294745, 1.9570721313347723],
                                  [0.18815116399861764, 0.0052185649900103404;
                                   0.0052185649900103404, 0.13282855600598231],
                                   4))
        ans = -4.57688303669188
        """
        self.assertAlmostEqual(wishart.log_likelihood(x), -4.57688303669188)

        # test with ndim = 3
        ndim = 3
        nvar = ndim * ndim
        eta_1 = np.array([-0.5, 0, 0, 0, -0.5, 0, 0, 0, -0.5],
                         dtype=wishart._dtype)
        dof = 5
        wishart = Wishart(nvar)
        wishart.dof = dof
        wishart.eta_ = eta_1
        self.assertAlmostEqual(wishart.dof, dof)
        x = np.array([4, 2, 3, 2, 5, 1, 3, 1, 9], dtype=wishart._dtype)
        #octave:> log(wishpdf([4,2,3; 2,5,1; 3,1,9],
                             #[1, 0, 0; 0, 1, 0; 0, 0, 1], 5))
        #ans = -13.7431848985804
        self.assertAlmostEqual(wishart.log_likelihood(x), -13.7431848985804)

    def test_log_likelihood_many_dimensions(self):
        ndim = 64
        nvar = ndim * ndim
        eta_1 = -0.5 * np.eye(ndim).flatten()
        eta_2 = 70.0
        wishart = Wishart(nvar)
        wishart.eta_ = eta_1
        dof = wishart._dtype(2) * eta_2 + wishart._dtype(ndim + 1)
        wishart.dof = dof
        x = np.eye(ndim).flatten()
        #octave:> wishpdf(eye(64), eye(64), 205, true)
        #ans = -24907.4964457670
        self.assertAlmostEqual(wishart.log_likelihood(x), -24907.4964457670)

    def test_sample_check_dimensions(self):
        ndim = 2
        nvar = ndim * ndim
        eta_1 = np.array([-1.5, -0.75, -0.75, -10.0])
        eta_2 = 50.0
        wishart = Wishart(nvar)
        wishart.eta_ = eta_1
        dof = wishart._dtype(2) * eta_2 + wishart._dtype(ndim + 1)
        wishart.dof = dof
        # nsamples = 1
        nsamples = 1
        samples = wishart.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 1)
        self.assertEqual(len(samples), nvar)
        # nsamples > 1
        nsamples = 10000
        samples = wishart.sample(nsamples)
        self.assertTrue(isinstance(samples, np.ndarray))
        self.assertEqual(samples.ndim, 2)
        self.assertEqual(len(samples), nsamples)
        self.assertEqual(len(samples[0]), nvar)

    def test_sample(self):
        ndim = 2
        nvar = ndim * ndim
        nsamples = int(10000)
        eta_1 = np.array([-1.5, -0.75, -0.75, -10.0])
        eta_2 = 50.0
        wishart = Wishart(nvar)
        wishart.eta_ = eta_1
        dof = wishart._dtype(2) * eta_2 + wishart._dtype(ndim + 1)
        wishart.dof = dof
        wishart._random_state.seed(0)
        samples = wishart.sample(nsamples)
        sample_mean = np.mean(samples, axis=0)
        expected_mean = (wishart.dof * wishart.V).flatten()
        self.assertTrue(norm(sample_mean - expected_mean) < 0.1)

    def test_sample_many_dimensions(self):
        np.random.seed(0)
        ndim = 50
        nvar = ndim * ndim
        nsamples = int(5000)
        eta_1 = -0.5 * positive_definite_matrix(ndim).flatten()
        eta_2 = 50.0
        wishart = Wishart(nvar)
        wishart.eta_ = eta_1
        dof = wishart._dtype(2) * eta_2 + wishart._dtype(ndim + 1)
        wishart.dof = dof
        samples = wishart.sample(nsamples)
        sample_mean = np.mean(samples, axis=0)
        expected_mean = (wishart.dof * wishart.V).flatten()
        self.assertTrue(norm(sample_mean - expected_mean) < 0.05 * norm(expected_mean))

    def test_set_source_params(self):
        ndim = 2
        nvar = ndim * ndim
        eta_1 = np.array([-1.5, -0.5, -0.5, -1.0])
        eta_2 = 50.0
        wishart = Wishart(nvar)
        wishart.eta_ = eta_1
        dof = wishart._dtype(2) * eta_2 + wishart._dtype(ndim + 1)
        wishart.dof = dof
        dof = 103
        self.assertEqual(wishart.dof, dof)
        V = np.array([0.4, -0.2, -0.2, 0.6])
        L = np.linalg.cholesky(V.reshape((ndim, ndim)))
        wishart_V = wishart.V.flatten()
        for i in range(len(wishart_V)):
            self.assertAlmostEqual(V[i], wishart_V[i])
        assert_array_almost_equal(wishart.L, L)

    def test_fit_with_one_sample(self):
        ndim = 2
        nvar = ndim * ndim
        dof = 5.0
        sample = np.array([[1.0, 0.5, 0.5, 1.0]])
        V = sample[0].reshape((2, 2)) / dof
        eta = -0.5 * np.linalg.inv(V).flatten()

        estimator = Wishart(nvar=nvar)
        estimator.dof = dof
        estimator.fit(sample)

        assert_array_equal(estimator.eta_, eta)
        assert_array_equal(estimator.V, V)

    def test_fit(self):
        ndim = 2
        nvar = ndim * ndim
        nsamples = int(1000)
        gt_eta_1 = np.array([-1.5, -0.75, -0.75, -1.0])
        eta_2 = np.float64(4.0)
        wishart = Wishart(nvar)
        wishart._random_state.seed(0)
        wishart.eta_ = gt_eta_1
        dof = wishart._dtype(2) * eta_2 + wishart._dtype(ndim + 1)
        wishart.dof = dof
        samples = wishart.sample(nsamples)

        estimator = Wishart(nvar=nvar)
        estimator._random_state.seed(0)
        init_eta_1 = np.array([-1.0, -0.5, -0.5, -1.0])
        estimator.eta_ = init_eta_1
        estimator.dof = dof

        estimator.fit(samples)
        est_eta_1 = estimator.eta_
        self.assertTrue(norm(est_eta_1 - gt_eta_1) < 0.05 * norm(gt_eta_1))

if __name__ == '__main__':
    unittest.main()
