#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import print_function

import sys
sys.path.append('.')
import unittest
import numpy as np
from numpy.linalg import norm
from numpy.random import multivariate_normal as mnormal
from numpy.testing import assert_array_equal

from vrl.ml import Wishart
from vrl.ml import WishartMixture


class TestWishartMixture(unittest.TestCase):
    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)

    def setUp(self):
        return

    def test_init(self):
        nvar = 4
        ncomp = 2
        mixture = WishartMixture(ncomp=ncomp, nvar=nvar)
        self.assertEqual(mixture.dof, 2.0)
        assert_array_equal(mixture.etas_, np.array([[-0.5, 0.0, 0.0, -0.5],
                                                   [-0.5, 0.0, 0.0, -0.5]]))
        assert_array_equal(mixture.weights_, np.array([0.5, 0.5]))
        weights = np.array([0.8, 0.2])
        etas = np.array([[-1.0, 0.3, 0.3, -4.5], [-1.5, 1.2, 1.2, -4.5]])
        dof = 5.0
        mixture = WishartMixture(ncomp=ncomp, nvar=nvar, dof=dof)
        mixture.etas_ = etas
        mixture.weights_ = weights
        self.assertEqual(mixture.min_pst, 1e-4)
        self.assertEqual(mixture.dof, dof)
        assert_array_equal(mixture.etas_, etas)
        assert_array_equal(mixture.weights_, weights)
        weights = np.array([1.6, 0.4])
        mixture = WishartMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        weights = np.array([0.8, 0.2])
        assert_array_equal(mixture.weights_, weights)

    def test_component(self):
        nvar = 4
        ncomp = 2
        weights = np.array([0.8, 0.2])
        etas = np.array([[-1.0, 0.2, 0.2, -4.5], [-1.5, 1.2, 1.2, -4.5]])
        dof = 5.0
        mixture = WishartMixture(ncomp=ncomp, nvar=nvar, dof=dof)
        mixture.etas_ = etas
        mixture.weights_ = weights
        for i in range(ncomp):
            component = mixture._component(i)
            assert_array_equal(component.eta_, etas[i])
            assert_array_equal(component.dof, dof)

    def test_fit_one_component(self):
        np.random.seed(0)
        ncomp = 1
        ndim = 2
        nvar = ndim * ndim
        nsamples = 5000
        # sample from a wishart
        eta_1 = np.array([-0.5, 0.25, 0.25, -1.0])
        wishart = Wishart(nvar)
        wishart.eta_ = eta_1
        wishart.dof = wishart._dtype(5)
        samples = wishart.sample(nsamples)

        weights = np.array([1.0])
        etas = np.array([[-1.5, -0.25, -0.25, -1.5]])
        dof = wishart._dtype(5)
        model = WishartMixture(ncomp=ncomp, nvar=nvar)
        model.weights__ = weights
        model.etas_ = etas
        model.dof = dof
        model.fit(samples)
        est_eta_1 = model.etas_[0]
        self.assertTrue(norm(est_eta_1 - eta_1) < 0.05 * norm(eta_1))

    def test_fit_one_component_3x3(self):
        np.random.seed(3)
        ncomp = 1
        ndim = 3
        nvar = ndim * ndim
        mean = np.zeros(ndim)
        cov0 = np.array([[1.1776, -0.2290, -1.1240],
                         [-0.2290, 5.7367, -3.0743],
                         [-1.1240, -3.0743, 3.4997]])
        dof = 20.0
        ns = 500
        samples = np.zeros((ns, nvar))
        for i in range(ns):
            samples[i] = np.cov(mnormal(mean, cov0, int(dof)).T).flatten()
        ################## init ###########
        model = WishartMixture(ncomp=ncomp, nvar=nvar, dof=dof)
        model.dof = dof
        model.compute_initial_estimate(data=samples)
        #############################
        model.fit(samples)
        #V0 = model._component(0).V
        #dof0 = model._component(0).dof
        #print("real_cov = ", cov0)
        #print("est_cov = ", dof0 * V0)

    def test_fit_two_components(self):
        ncomp = 2
        ndim = 2
        nvar = ndim * ndim
        nsamples = 5000
        weights = np.array([0.1, 0.9])
        nsamples_1 = int(nsamples * weights[0])
        nsamples_2 = int(nsamples * weights[1])
        wishart = Wishart(nvar)
        dof = wishart._dtype(5)
        etas = np.array([[-0.5, 0.25, 0.25, -1.0], [-2.5, -1.25, -1.25, -1.0]])
        # sample from a wishart
        wishart = Wishart(nvar)
        wishart.eta_ = etas[0]
        wishart.dof = dof
        samples1 = wishart.sample(nsamples_1)
        # sample from a wishart
        wishart = Wishart(nvar)
        wishart.eta_ = etas[1]
        wishart.dof = dof
        samples2 = wishart.sample(nsamples_2)

        samples = np.concatenate((samples1, samples2))
        np.random.seed(0)
        np.random.shuffle(samples)

        model = WishartMixture(ncomp=ncomp, nvar=nvar, dof=dof)
        model.dof = dof
        model.compute_initial_estimate(data=samples)
        model.fit(samples)
        #eta_0 = model.etas_[0]
        #eta_1 = model.etas_[1]
        #print('nsamples = ', samples.shape[0])
        #print('real weights = ', weights)
        #print('model.weights = ', model.weights)
        #print('real eta_0 = ', etas[0])
        #print('real eta_1 = ', etas[1])
        #print('model.eta_0 = ', eta_0)
        #print('model.eta_1 = ', eta_1)

    def test_fit_hidot_paper(self):
        np.random.seed(3)
        ncomp = 4
        ndim = 3
        nvar = ndim * ndim
        mean = np.zeros(ndim)
        cov0 = np.array([[1.1776, -0.2290, -1.1240],
                         [-0.2290, 5.7367, -3.0743],
                         [-1.1240, -3.0743, 3.4997]])
        cov1 = np.array([[2.8644, 1.7371, -2.9438],
                         [1.7371, 2.6639, -2.3116],
                         [-2.9438, -2.3116, 3.7669]])
        cov2 = np.array([[0.7886, 0.0112, -0.2788],
                         [0.0112, 4.9316, 3.5732],
                         [-0.2788, 3.5732, 2.7872]])
        cov3 = np.array([[2.9770, 1.9344, -0.4107],
                         [1.9344, 2.8154, 0.7750],
                         [-0.4107, 0.775, 1.5226]])

        dof = 20.0
        ns = 500
        samples = np.zeros((ns * 4, nvar))
        for i in range(ns):
            samples[i] = np.cov(mnormal(mean, cov0, int(dof)).T).flatten()
        for i in range(ns):
            samples[ns + i] = np.cov(mnormal(mean, cov1,
                                             int(dof)).T).flatten()
        for i in range(ns):
            samples[2 * ns + i] = np.cov(mnormal(mean, cov2,
                                                 int(dof)).T).flatten()
        for i in range(ns):
            samples[3 * ns + i] = np.cov(mnormal(mean, cov3,
                                                 int(dof)).T).flatten()

        np.random.shuffle(samples)

        ################## init ###########
        model = WishartMixture(ncomp=ncomp, nvar=nvar, dof=dof)
        model.dof = dof
        model.compute_initial_estimate(data=samples)
        #print("############ kmeans")
        #print('model.weights = ', model.weights)
        #V0 = model._component(0).V
        #dof0 = model._component(0).dof
        #V1 = model._component(1).V
        #dof1 = model._component(1).dof
        #V2 = model._component(2).V
        #dof2 = model._component(2).dof
        #V3 = model._component(3).V
        #dof3 = model._component(3).dof
        #print('est cov0 = ', dof0 * V0)
        #print('est cov1 = ', dof1 * V1)
        #print('est cov2 = ', dof2 * V2)
        #print('est cov3 = ', dof3 * V3)
        #############################
        model.fit(samples)
        #print("############ em")
        #print('model.weights = ', model.weights)
        #V0 = model._component(0).V
        #dof0 = model._component(0).dof
        #V1 = model._component(1).V
        #dof1 = model._component(1).dof
        #V2 = model._component(2).V
        #dof2 = model._component(2).dof
        #V3 = model._component(3).V
        #dof3 = model._component(3).dof
        #print('est cov0 = ', dof0 * V0)
        #print('est cov1 = ', dof1 * V1)
        #print('est cov2 = ', dof2 * V2)
        #print('est cov3 = ', dof3 * V3)

        #print('real cov0 = ', cov0)
        #print('real cov1 = ', cov1)
        #print('real cov2 = ', cov2)
        #print('real cov3 = ', cov3)

    def test_get_params(self):
        nvar = 4
        ncomp = 2
        dof = 10.0
        weights = np.array([0.8, 0.2])
        etas = np.array([[-1.0, 0.3, 0.3, -4.5], [-1.5, 1.2, 1.2, -4.5]])
        mixture = WishartMixture(ncomp=ncomp, nvar=nvar)
        mixture.etas_ = etas
        mixture.weights_ = weights
        mixture.dof = dof
        params = {'ncomp': 2, 'nvar': 4, 'min_pst': 0.0001, 'dof': dof}
        mixture_params = mixture.get_params()
        assert_array_equal(mixture_params.pop('etas_'), etas)
        assert_array_equal(mixture_params.pop('weights_'), weights)
        self.assertEqual(mixture_params, params)


if __name__ == '__main__':
    unittest.main()
