# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

from os import makedirs, environ

from os.path import curdir, abspath, join, dirname, exists, splitext, pardir, \
    getmtime, sep, extsep

from ctypes import cdll

import numpy as np

import scipy.sparse as sps

from .io import load, save, file_exists, VRL_EXT

from .utils import get_logger

from .random import RNG


def load_libcv():
    return load_library('libcv.so')


def load_libml():
    return load_library('libml.so')


def load_library(libname):
    if 'VIRTUAL_ENV' in environ:
        return cdll.LoadLibrary(join(join(environ['VIRTUAL_ENV'], 'lib'), libname))
    else:
        try:
            return cdll.LoadLibrary(libname)
        except:
            raise OSError(libname + ' not found. Try updating the LD_LIBRARY_PATH environment variable (e.g., export LD_LIBRARY_PATH=/usr/local/lib:$LD_LIBRARY_PATH)')


class VRLBaseEstimator(object):
    '''Base class (in the spirit of sklearn.baseEstimator) allowing models and
    data to be saved to and loaded from files in a pipelined manner.

    Note:
      using a VRLBaseEstimator instance with sklearn.Pipeline is possible but
      not recommended since the paths are not updated when the parameters for
      the different steps are changed.

    Attributes
    ----------
    input_path : str
      Input files base path
    output_base_path : str
      Root path for the model and output files
    output_path : str
      Output files base path
    model_path : str
      Model file base path
    force_overwrite : bool
      If True, files will be overwritten if they already exist
    image_reader : {vrl.cv.ImageReader,  None}
      Relevant for estimators that take images as input. It defines the way
      images are to be read, eg. for color histograms images should be read as
      RGB but for computing gradients only gray scale images are needed.
    random_state : {vrl.random.RNG, None}
      Random generator

    Examples
    --------

    let's say we have the following estimators:

    class A(VRLBaseEstimator):
      def __init__(self):
        VRLBaseEstimator.__init__(self)
      def get_param(self):
        return {\'x\': 1.0}

    class B(VRLBaseEstimator):
      def __init__(self):
        VRLBaseEstimator.__init__(self)
      def get_param(self):
        return {\'y\': 2.0}

    setting the file pipeline as follows:

    >>> a = A()
    >>> a.input_path = \'/input/path/\'
    >>> a.output_base_path = \'/output/base/path/\'
    >>> b = B()
    >>> b.input_path = a.output_path
    >>> b.output_base_path = a.output_path

    should give the following output:

    >>> input_file = \'data.dat\'
    >>> output_file = a.pass_through(input_file)
    >>> print(output_file)
    \'/output/base/path/A__x_1.0/data.dat\'
    >>> output_file = b.pass_through(input_file)
    >>> print(output_file)
    \'/output/base/path/A__x_1.0/B__y_2.0/data.dat\'
    '''
    def __init__(self, **kwargs):
        self.input_path = curdir
        self.output_base_path = curdir
        self.force_overwrite = False

        self._image_reader = None

        self._params_runtime = ()

        self._suffix = None

        # # implement transform method only if it is present at derived class
        # if hasattr(self, '_transform'):
        #     self.transform = self.__transform

        # if hasattr(self, '_fit'):
        #     self.fit = self.__fit

        self.logger = get_logger(self.__module__ + '.' + self.__class__.__name__)

        rng = RNG()
        if 'random_state' in kwargs:
            rng = RNG() if kwargs['random_state'] is None else kwargs['random_state']
        self._random_state = rng

    @property
    def input_path(self):
        '''get input_path'''
        return self._input_path

    @input_path.setter
    def input_path(self, path):
        '''set input_path'''
        self._input_path = abspath(curdir if path is None else path)

    @property
    def output_base_path(self):
        '''get output_base_path'''
        return self._output_base_path

    @output_base_path.setter
    def output_base_path(self, path):
        '''set output_base_path'''
        self._output_base_path = abspath(curdir if path is None else path)

    @property
    def output_path(self):
        '''returns `output_path` based on the current parameters'''
        # must be recomputed every time since parameters might have changed
        # since last call

        path_0 = join(self._output_base_path, self.__class__.__name__)

        # add image_reader params
        if self._image_reader is not None:
            params = self._image_reader.get_params()
            if params['color_space'] in ('gray', 'grey'):
                del params['color_space']
            path = self._params_to_path(params)
            path_0 += ('__' if len(path) > 0 else '') + path

        # split params into base and runtime
        params = self.get_params()
        params_runtime = {}
        for key in self._params_runtime:
            params_runtime[key] = params[key]
            del params[key]

        # setup path accordingly
        path_1 = self._params_to_path(params)
        path_0 += '__' if len(path_1) > 0 else ''
        output_path_ = abspath(path_0 + path_1)
        if len(self._params_runtime) > 0:
            path_2 = self._params_to_path(params_runtime)
            output_path_ = join(output_path_, path_2.lstrip('_'))

        # this is just for debugging/experimentation porpouses: if (internal)
        # variable "_suffix" is set, appends it at the end of the path
        if self._suffix is None:
            return output_path_
        return output_path_ + '__' + self._suffix

    @property
    def model_path(self):
        '''get model_path (output_path without runtime params)'''
        if len(self._params_runtime) == 0:
            return self.output_path
        return abspath(join(self.output_path, pardir))

    @property
    def model_file(self):
        '''get model_file fullpath'''
        return join(self.model_path, 'model.{}'.format(VRL_EXT))

    def _load_model(self):
        params = load(self.model_file).get_params()
        for key in [k for k in params if k in self._params_runtime]:
            del params[key]
        self.set_params(**params)

    def _save_model(self):
        save(self.model_file, self)

    @staticmethod
    def _params_to_path(params):
        '''transforms a dict with params to a readable path'''
        excluded = (np.ndarray, sps.spmatrix, list, tuple, dict, bool,
                    type(None), VRLBaseEstimator)
        keys = sorted(params.keys())

        keys_ = [k for k in keys if not isinstance(params[k], excluded)]
        path = '_'.join(['{}_{}'.format(k, params[k]) for k in keys_])

        # append True bool params at the end of the path
        for key in [k for k in keys if isinstance(params[k], bool)]:
            path += '_{}'.format(key) if params[key] else ''

        return path

    def __str__(self):
        '''makes the instance print readable'''
        unreadable = (np.ndarray, list, tuple, dict)
        params = []
        for key, val in self.get_params().items():
            if not isinstance(val, unreadable):
                params.append('{}={}'.format(key, val))
            else:
                params.append('{}=<{}>'.format(key, val.__class__.__name__))
        return '{}({})'.format(self.__class__.__name__, ', '.join(params))

    def _relpath(self, fullpath):
        '''replaces output_base_path by a shorter string just for visualization
        pourposes (avoid clutter when logging)
        '''
        return fullpath.replace(self.output_base_path, '(...)')

    @property
    def force_overwrite(self):
        ''' get force_overwrite '''
        return self._force_overwrite

    @force_overwrite.setter
    def force_overwrite(self, val):
        ''' set force_overwrite '''
        self._force_overwrite = bool(val)

    @property
    def image_reader(self):
        ''' get image_reader '''
        return self._image_reader

    @property
    def params_runtime(self):
        ''' get params_runtime '''
        return self._params_runtime

    @property
    def random_state(self):
        ''' get random_state '''
        return self._random_state

    def get_params(self, deep=True):
        ''' get dict with params. TBD at derived class '''
        return {}
        # return BaseEstimator.get_params(self, deep=deep)

    def set_params(self, **params):
        ''' set instance parameters '''
        # BaseEstimator.set_params(self, **params)
        for key in [k for k in self.get_params() if k in params]:
            try:
                setattr(self, key, params[key])
            except AttributeError:  # the attribute don't exists or can't be set at runtime
                pass
        return self

    def pass_through(self, input_file):
        ''' returns the path to the file that would have resulted from calling
        to `transform`
        '''
        if isinstance(input_file, (list, tuple)):
            return [self.pass_through(f) for f in input_file]

        if input_file.startswith(self._input_path):
            output_file = input_file.replace(self._input_path,
                                             self.output_path)
        else:
            output_file = join(self.output_path, input_file)

        return abspath(splitext(output_file)[0] + extsep + VRL_EXT)

    def fit(self, X, y=None):
        ''' fits the model using X as input
        '''
        if isinstance(X, (list, tuple)) and isinstance(X[0], np.ndarray):
            self._fit_from_list(X, y)
            return self

        elif isinstance(X, np.ndarray):
            self._fit(X, y)
            return self

        if file_exists(self.model_file) and not self.force_overwrite:
            self.logger.info('{} already exists'.format(self._relpath(self.model_file)))
            self._load_model()
            return self

        self.logger.info('fitting model ...')

        if X is None:  # some models are learned with no external data, e.g. by simulation
            assert(y is None)
            self._fit(None)

        elif isinstance(X, str):
            self._fit_from_file(X, y)

        elif isinstance(X, (list, tuple)) and isinstance(X[0], str):
            self._fit_from_filelist(X, y)

        else:
            raise TypeError('wrong data type')

        self.logger.info('done')

        if not exists(dirname(self.model_file)):
            makedirs(dirname(self.model_file))
        self._save_model()

        self.logger.info('saved as {}'.format(self._relpath(self.model_file)))

        return self

    def _fit(self, X, y=None):
        ''' fits a model from a numpy.ndarray. TBD @ derived class
        '''
        raise NotImplementedError

    def _fit_from_list(self, X, y=None):
        ''' fits a model from a list of numpy.ndarrays
        '''
        return self._fit(np.vstack(X), y)

    def _fit_from_file(self, X, y=None):
        ''' fits a model from a (vrl readable) file
        '''
        ifile = join(self._input_path, X)
        if not file_exists(ifile):
            raise IOError('{} not found'.format(ifile))

        # read data from disk and fit model
        data = load(ifile, stack=True)

        # get a valid field if data is a dictionary
        if isinstance(data, dict):
            if 'data' not in data:
                raise TypeError('wrong data type')
            data = data['data']

        # fit model from data
        return self._fit(data, y)

    def _fit_from_filelist(self, X, y=None):
        ''' fits a model from a list of (vrl readable) files
        '''
        flist = [join(self._input_path, fname) for fname in X]
        fexist = [file_exists(fname) for fname in flist]
        if not all(fexist):
            not_found_list = ''
            for i in np.where(np.bitwise_not(fexist))[0]:
                not_found_list += '{} not found\n'.format(flist[i])
            raise IOError('{}'.format(not_found_list))
        data = load(flist, stack=True)
        return self._fit(data, y)

    def transform(self, X):
        ''' transforms the input
        '''
        if isinstance(X, np.ndarray):
            return self._transform(X)

        elif isinstance(X, (tuple, list)):
            self.logger.info('transformimg {} elements ...'.format(len(X)))
            if isinstance(X[0], basestring):
                self.logger.info('saving files to {}'.format(self._relpath(self.output_path)) + sep)
            retval = [self.transform(f) for f in X]
            self.logger.info('done')
            return retval

        elif isinstance(X, str):
            # input file
            ifile = join(self._input_path, X)
            if not file_exists(ifile):
                raise IOError('{} not found'.format(ifile))

            # output file
            ofile = self.pass_through(ifile)
            if file_exists(ofile) and not self.force_overwrite and \
               not getmtime(ifile) > getmtime(ofile):
                self.logger.debug('{} already exists'.format(self._relpath(ofile)))
                return ofile.replace(self.output_path, '')[len(sep):]

            # read input data
            try:
                if self._image_reader is not None:
                    # the input is an image
                    idata = self._image_reader(ifile)
                else:
                    # the input is vrl data
                    idata = load(ifile)
            except:
                self.logger.error('error while reading {}'.format(ifile))
                raise

            # apply transformation
            try:
                odata = self._transform(idata)
            except:
                self.logger.error('error while computing {}'.format(ofile))
                raise

            if odata is idata:  # if dummy transform, do not store anything
                return ifile

            # mkdir output dir and subdirs
            ofile_path = dirname(ofile)
            if not exists(ofile_path):
                makedirs(ofile_path)

            # save data and return filename
            save(ofile, odata)
            self.logger.debug('{}'.format(self._relpath(ofile)))
            return ofile.replace(self.output_path, '')[len(sep):]

        else:
            raise TypeError('wrong data type')

    def _transform(self, X):
        ''' transforms the input. TBD @ derived class
        '''
        raise NotImplementedError

    def fit_transform(self, X, y=None):
        '''fits the model from the data provided as input and then transforms
        it according to the learned model
        '''
        self.fit(X, y)
        return self.transform(X)


import joblib


class SKLEstimator(VRLBaseEstimator):
    def __init__(self, estimator=None, to_show=None, to_hide=None, random_state=None):
        VRLBaseEstimator.__init__(self, random_state=random_state)
        self.estimator = estimator

        self.to_show = to_show
        self.to_hide = to_hide

        self.logger = get_logger(self.__module__ + '.(sklearn)' + \
                                 self.estimator.__class__.__name__)

    @property
    def to_show(self):
        return self._to_show

    @to_show.setter
    def to_show(self, val):
        self._to_show = [] if val is None else tuple(val)

    @property
    def to_hide(self):
        return self._to_hide

    @to_hide.setter
    def to_hide(self, val):
        self._to_hide = [] if val is None else list(val)
        self._to_hide += ['random_state', 'verbose', 'n_jobs',]
        self._to_hide = tuple(self._to_hide)

    def get_params(self, deep=True):
        return self.estimator.get_params(deep=deep)

    @VRLBaseEstimator.params_runtime.setter
    def params_runtime(self, val):
        ''' get params_runtime '''
        self._params_runtime = () if val is None else tuple(val)

    def _params_to_path(self, params):
        # sklearn objects to class names
        obj_params = []
        keys = params.keys()
        for key in keys:
            if any([k.startswith(key + '__') for k in keys]):
                params[key] = params[key].__class__.__name__
                obj_params.append(key)

        # show/hide specific parameters
        if len(self._to_show) > 0:
            params_ = {}
            for key in self._to_show:
                params_[key] = params[key]
            params = params_
        elif len(self._to_hide) > 0:
            for key in [k for k in params.keys() if k in self._to_hide]:
                del params[key]

        # remove the prefix for the parameters that correspond to other sklearn
        # sub-objets
        path_ = VRLBaseEstimator._params_to_path(params)
        for key in obj_params:
            path_ = path_.replace(key + '__', '')

        if len(path_) > 0:
            return self.estimator.__class__.__name__ + '_' + path_
        return self.estimator.__class__.__name__

    @property
    def model_file(self):
        return VRLBaseEstimator.model_file.fget(self) + '_'

    def _load_model(self):
        self.estimator = joblib.load(self.model_file)

    def _save_model(self):
        joblib.dump(value=self.estimator, filename=self.model_file, compress=1)

    def _fit(self, X, y=None):
        return self.estimator.fit(X, y)

    def _transform(self, X):
        return self.estimator.transform(X)
