# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from skimage import feature

from .image_reader import ImageReader

from .multiscale_feature import MultiscaleFeature

from .utils import sample_uniform_patches


class BRIEF(MultiscaleFeature):
    def __init__(self, max_layer=4, downscale=1.414, step=6, patch_size=24,
                 descriptor_size=256, random_state=None):
        ''' Multi-scale BRIEF descriptors (skimage.feature.BRIEF)

        Parameters:
        ----------
        max_layer: int, optional
          largest index for a resolution pyramid of (max_layer+1) levels
        downscale: float, optional
          down-scale factor between layer (i+1) and (i)
        step: int, optional
          sampling step along each axis
        patch_size: int, optional
          sampled patches have patch_size x patch_size pixels
        descriptor_size: int, optional
          lenght (in bits) of the binary descriptor
        random_state: vrl.base.RNG, optional
          external RNG. If None, creates its own
        '''
        MultiscaleFeature.__init__(self, max_layer=max_layer,
                                   downscale=downscale,
                                   random_state=random_state)
        self._step = int(step)

        self._patch_size = int(patch_size)

        self._descriptor_size = int(descriptor_size)

        self._descriptor_size = descriptor_size

        self._image_reader = ImageReader(color_space='gray')

        self._mode = 'normal'
        self._sigma = 1.0
        self._sample_seed = int(hash(self.__class__.__name__))

    @property
    def step(self):
        return self._step

    @property
    def patch_size(self):
        return self._patch_size

    @property
    def descriptor_size(self):
        return self._descriptor_size

    def get_params(self, deep=True):
        params = MultiscaleFeature.get_params(self, deep=deep)
        params.update({'step': self.step, 'patch_size': self.patch_size,
                       'descriptor_size': self.descriptor_size})
        return params

    def _transform_layer(self, X, layer=0):
        if X.shape[0] < self.patch_size or X.shape[1] < self.patch_size:
            return None

        # instantiate BRIEF object
        brief = feature.BRIEF(descriptor_size=self.descriptor_size,
                              patch_size=self.patch_size, mode=self._mode,
                              sigma=self._sigma, sample_seed=self._sample_seed)

        # get points sampled uniformly
        loc = sample_uniform_patches(X.shape, self.patch_size, self.step)
        if loc is None:
            return None

        # extract features (using keypoint center coordinates)
        brief.extract(X, loc + self.patch_size // 2)

        loc = np.hstack([loc, np.ones((len(loc), 1))])
        loc *= self.downscale ** float(layer)

        return {'data': brief.descriptors.astype(np.uint8), 'loc': loc}
