# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ..multiscale_feature import MultiscaleFeature


class Cellgrid(MultiscaleFeature):

    def __init__(self, max_layer=5, downscale=1.414, cell_size=(8, 8),
                 block_size=(4, 4), cell_norm=None, norm=None,
                 random_state=None):
        '''base class for the multiscale cellgrid features

        The image is first split into cells of cell_size[0]xcell_size[1]
        pixels. Each cell give rise to a cell-level descriptor, e.g. an
        histogram. Then, groups of block_size[0]xblock_size[1] cells are
        formed. A cellgrid descriptor is the concatenation of the cell-features
        from each block. Different normalization can be applied at cell- and/or
        block-level. Cellgrid features are computed on a regular grid with a
        step size of one cell.

        '''
        MultiscaleFeature.__init__(self, max_layer=max_layer,
                                   downscale=downscale,
                                   random_state=random_state)

        self._cell_size = tuple(cell_size)
        if len(self._cell_size) != 2:
            raise ValueError('cell_size must be in the form (cell_size_x, cell_size_y), where the sizes are given in pixels')

        self._block_size = tuple(block_size)
        if len(self._cell_size) != 2:
            raise ValueError('block_size must be in the form (block_size_x, block_size_y), where the sizes are given in cell units')

        self._cell_norm = None if cell_norm is None else int(cell_norm)

        self._norm = None if norm is None else int(norm)

        self._image_reader = None

        self._cells_per_block = self._block_size[0] * self._block_size[1]

        self._eps = 2 ** -23

    @property
    def cell_size(self):
        return self._cell_size

    @property
    def block_size(self):
        return self._block_size

    @property
    def cell_norm(self):
        return self._cell_norm

    @property
    def norm(self):
        return self._norm

    def _encode(self, img):
        raise NotImplementedError

    def _cell_aggregate(self, fmap):
        raise NotImplementedError

    def _cell_block_map(self, size):
        assert(len(size) == 2)

        # number of cells along each dimension
        ni = size[0] // self._cell_size[0]
        nj = size[1] // self._cell_size[1]

        # offset (in pixels) of the top-left cell
        i0 = (size[0] % self._cell_size[0]) // 2
        j0 = (size[1] % self._cell_size[1]) // 2

        # NIxNJx2 array with the coordinates (in pixels) of the cells
        i = range(i0, i0 + ni * self._cell_size[0], self._cell_size[0])
        j = range(j0, j0 + nj * self._cell_size[1], self._cell_size[1])
        ii, jj = np.meshgrid(i, j, indexing='ij')
        cells = np.dstack((ii, jj))
        cells_fl = cells.reshape(ni * nj, 2, order='C')  # iterable cells

        # NIxNJ array with the indices to the corresponding cells
        cells_id = np.arange(ni * nj).reshape(ni, nj, order='C')

        # NRxNSx2 array with the coordinates (in cells) of the blocks
        nr = cells.shape[0] - self._block_size[0] + 1
        ns = cells.shape[1] - self._block_size[1] + 1
        rr, ss = np.meshgrid(range(nr), range(ns), indexing='ij')
        blocks = np.dstack((rr, ss))
        blocks_fl = blocks.reshape(nr * ns, 2, order='C')  # iterable blocks

        # array storing, for each block (=rows), the indices to the cells it
        # contains
        def cell_indx_(b):
            indx = cells_id[b[0]: b[0] + self._block_size[0],
                            b[1]: b[1] + self._block_size[1]]
            return np.array(indx).ravel()
        block_cell_indx = np.array(map(cell_indx_, blocks_fl))

        return cells_fl, block_cell_indx

    def get_params(self, deep=True):
        params = MultiscaleFeature.get_params(self, deep=deep)
        params.update({'cell_size': self.cell_size,
                       'block_size': self.block_size,
                       'cell_norm': self.cell_norm,
                       'norm': self.norm})
        return params

    def _transform_layer(self, X, layer=0):
        h = X.shape[0]
        w = X.shape[1]

        # stop if scale level too small
        if h < self._cell_size[0] * self._block_size[0] or \
           w < self._cell_size[1] * self._block_size[1]:
            return None

        # compute pixel-level features
        pixfeat = self._encode(X)
        assert(pixfeat.shape[:2] == X.shape[:2])

        # get cell-block mapping
        cells, blocks = self._cell_block_map(X.shape[:2])

        # aggregate pixel level features into cell level features
        cfeat = self._aggregate(cells, pixfeat)

        # cell normalization
        if self._cell_norm is not None:
            #cell_norm = np.linalg.norm(cfeat, ord=self._cell_norm, axis=1)
            cell_norm = np.array([np.linalg.norm(x, ord=self._cell_norm) for x in cfeat])
            cfeat = cfeat / (cell_norm[:, np.newaxis] + self._eps)

        # concatenate cells into blocks using numpy row indexing
        data = np.array(map(lambda b: cfeat[b].ravel(), blocks))

        # get top-left block coordinates (in pixels) w.r.t. the base image
        cells_loc = np.hstack((cells, np.ones((len(cells), 1))))
        loc = cells_loc[blocks[:, 0]] * (self.downscale ** float(layer))

        # block normalization
        if self._norm is not None:
            #norm = np.linalg.norm(data, ord=self._norm, axis=1)
            norm = np.array([np.linalg.norm(x, ord=self._norm) for x in data])
            data = data / (norm[:, np.newaxis] + self._eps)
            return {'data': data, 'loc': loc, 'fnorm': norm}

        return {'data': data, 'loc': loc}
