# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from numpy import bincount

from .cellgrid import Cellgrid

from ..image_reader import ImageReader


class ColorHist(Cellgrid):

    def __init__(self, color_space='rgb', n_bins=4, as_uint=False, max_layer=4,
                 downscale=1.414, cell_size=(8, 8), block_size=(4, 4),
                 cell_norm=None, norm=None, random_state=None):
        ''' cellgrid feature on multichannel color histograms

        Parameters:
        -----------
        color_space:  vrl.cv.ImageReader, optional
          color space
        n_bins:  int, optional
          number of histogram bins on each cell
        as_uint: bool, optional
          if True, features are converted to uint
        max_layer: int, optional
          largest index for a resolution pyramid of (max_layer+1) levels
        downscale: float, optional
          down-scale factor between layer (i+1) and (i)
        cell_size: tuple or int, optional
          cell size in pixels. If int, a square cell is assumed
        block_size: tuple or int, optional
          block size in cells. If int, a square block is assumed
        cell_norm: int or None, optional
          cell-level Lp-normalization
        block_norm: int or None, optional
          block-level Lp-normalization
        random_state: vrl.base.RNG, optional
          external RNG. If None, creates its own
        '''
        Cellgrid.__init__(self, max_layer=max_layer, downscale=downscale,
                          cell_size=cell_size, block_size=block_size,
                          cell_norm=cell_norm, norm=norm,
                          random_state=random_state)

        self._image_reader = ImageReader(color_space=color_space)

        self._n_bins = int(n_bins)

        self._as_uint = bool(as_uint)

        self._dtype = np.uint32 if self._as_uint else np.float64

        if self.as_uint and \
           (self._cell_norm is not None or self._norm is not None):
            raise RuntimeError

    @property
    def n_bins(self):
        return self._n_bins

    @property
    def as_uint(self):
        return self._as_uint

    def _encode(self, img):
        assert img.ndim == 3

        if issubclass(img.dtype.type, np.floating):
            return (img * float(self._n_bins) * 255.0 / 256.0).astype(np.uint32)
        return ((img / 256.) * float(self._n_bins)).astype(np.uint32)

    def _aggregate(self, cells, fmap):
        def cell_feat_(c):
            cmap = fmap[c[0]: c[0] + self._cell_size[0],
                        c[1]: c[1] + self._cell_size[1]]
            h0 = bincount(cmap[:, :, 0].ravel(), minlength=self._n_bins)
            h1 = bincount(cmap[:, :, 1].ravel(), minlength=self._n_bins)
            h2 = bincount(cmap[:, :, 2].ravel(), minlength=self._n_bins)
            return np.hstack((h0, h1, h2))
        return np.array(map(cell_feat_, cells), dtype=self._dtype)

    def get_params(self, deep=True):
        params = Cellgrid.get_params(self, deep=deep)
        params.update({'n_bins': self.n_bins,
                       'as_uint': self.as_uint})
        return params
