# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from math import sqrt

from .cellgrid import Cellgrid

from ..image_reader import ImageReader


class ColorStat(Cellgrid):

    def __init__(self, color_space='rgb', max_layer=4, downscale=1.414,
                 cell_size=(8, 8), block_size=(4, 4), cell_norm=None,
                 norm=None, random_state=None):
        ''' cellgrid features on color statistics (mean and std. dev. from each channel)

        Parameters:
        -----------
        color_space:  vrl.cv.ImageReader, optional
          color space
        max_layer: int, optional
          largest index for a resolution pyramid of (max_layer+1) levels
        downscale: float, optional
          down-scale factor between layer (i+1) and (i)
        cell_size: tuple or int, optional
          cell size in pixels. If int, a square cell is assumed
        block_size: tuple or int, optional
          block size in cells. If int, a square block is assumed
        cell_norm: int or None, optional
          cell-level Lp-normalization
        block_norm: int or None, optional
          block-level Lp-normalization
        random_state: vrl.base.RNG, optional
          external RNG. If None, creates its own
        '''
        Cellgrid.__init__(self, max_layer=max_layer, downscale=downscale,
                          cell_size=cell_size, block_size=block_size,
                          cell_norm=cell_norm, norm=norm,
                          random_state=random_state)

        self._image_reader = ImageReader(color_space=color_space)

    def _encode(self, img):
        if img.ndim != 3:
            raise ValuError('input must be a 3 channel image')
        return img

    def _aggregate(self, cells, fmap):
        def cell_feat_(c):
            cmap = fmap[c[0]: c[0] + self._cell_size[0],
                        c[1]: c[1] + self._cell_size[1]]
            c0 = cmap[:, :, 0]
            m01 = np.mean(c0)
            m02 = np.mean(c0 ** 2)
            c1 = cmap[:, :, 1]
            m11 = np.mean(c1)
            m12 = np.mean(c1 ** 2)
            c2 = cmap[:, :, 2]
            m21 = np.mean(c2)
            m22 = np.mean(c2 ** 2)
            return [m01, sqrt(abs(m02 - m01 ** 2)),
                    m11, sqrt(abs(m12 - m11 ** 2)),
                    m21, sqrt(abs(m22 - m21 ** 2))]

        return np.array(map(cell_feat_, cells))
