# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from scipy.ndimage import gaussian_filter

from .cellgrid import Cellgrid


class DCov(Cellgrid):

    def __init__(self, sigma=1.0, order=1, as_matrix=False, max_layer=4,
                 downscale=1.414, cell_size=(8, 8), block_size=(4, 4),
                 cell_norm=None, norm=None, random_state=None):
        Cellgrid.__init__(self, max_layer=max_layer, downscale=downscale,
                          cell_size=cell_size, block_size=block_size,
                          cell_norm=cell_norm, norm=norm,
                          random_state=random_state)

        self._sigma = float(sigma)

        self._order = int(order)
        if not self._order in (1, 2,):
            raise ValueError('order must be in {1, 2}')

        self._as_matrix = bool(as_matrix)

        self._n_cells = self._cell_size[0] * self._cell_size[1]

    @property
    def sigma(self):
        return self._sigma

    @property
    def order(self):
        return self._order

    @property
    def as_matrix(self):
        return self._as_matrix

    def _encode(self, img):
        Ix = gaussian_filter(img, sigma=self._sigma, order=(1, 0))
        Iy = gaussian_filter(img, sigma=self._sigma, order=(0, 1))

        if self._order == 1:
            return np.dstack((Ix, Iy))

        Ixx = gaussian_filter(img, sigma=self._sigma, order=(2, 0))
        Iyy = gaussian_filter(img, sigma=self._sigma, order=(0, 2))
        Ixy = gaussian_filter(img, sigma=self._sigma, order=(1, 1))

        return np.dstack((Ix, Iy, Ixx, Ixy, Iyy))

        # Im = np.sqrt(Ix ** 2 + Iy ** 2)
        # return np.dstack((Ix, Iy, Ixx, Ixy, Iyy, Im))

    def _aggregate(self, cells, fmap):
        def cell_feat_(c):
            cmap = fmap[c[0]: c[0] + self._cell_size[0],
                        c[1]: c[1] + self._cell_size[1]]
            cmap = cmap.reshape(self._n_cells, cmap.shape[2])
            return np.cov(cmap, rowvar=0).ravel()

        return np.array(map(cell_feat_, cells))

    def _transform_layer(self, X, layer=0):
        ret = Cellgrid._transform_layer(self, X, layer)
        if self._as_matrix:
            n_channels = {1: 2, 2: 5}
            shape = (n_channels[self._order] * self._block_size[0],
                     n_channels[self._order] * self._block_size[1])
            ret['data'] = [np.ndarray(shape=shape, buffer=mat)
                           for mat in ret['data']]
        return ret

    def get_params(self, deep=True):
        params = Cellgrid.get_params(self, deep=deep)
        params.update({'sigma': self._sigma,
                       'order': self._order,
                       'as_matrix': self._as_matrix})
        return params
