# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from numpy import bincount

from .cellgrid import Cellgrid

from ..image_reader import ImageReader

from skimage.feature import local_binary_pattern as lbp


class LBPH(Cellgrid):

    def __init__(self, P=8, R=1.0, method='default', as_uint=False,
                 max_layer=4, downscale=1.414, cell_size=(8, 8),
                 block_size=(4, 4), cell_norm=None, norm=None,
                 random_state=None):
        ''' cellgrid features on LBP histograms

        Parameters:
        -----------
        P: int, optional
          sampling points
        R: float, optional
          sampling radius
        method: string, optional
          LBP variant (see skimage.feature.local_binary_pattern)
        as_uint: bool, optional
          if True, features are converted to uint
        max_layer: int, optional
          largest index for a resolution pyramid of (max_layer+1) levels
        downscale: float, optional
          down-scale factor between layer (i+1) and (i)
        cell_size: tuple or int, optional
          cell size in pixels. If int, a square cell is assumed
        block_size: tuple or int, optional
          block size in cells. If int, a square block is assumed
        cell_norm: int or None, optional
          cell-level Lp-normalization
        block_norm: int or None, optional
          block-level Lp-normalization
        random_state: vrl.base.RNG, optional
          external RNG. If None, creates its own
        '''
        Cellgrid.__init__(self, max_layer=max_layer, downscale=downscale,
                          cell_size=cell_size, block_size=block_size,
                          cell_norm=cell_norm, norm=norm,
                          random_state=random_state)

        self._P = int(P)

        self._R = int(R)

        self._method = str(method)

        self._as_uint = bool(val)

        self._dtype = np.uint32 if self._as_uint else np.float64

        if self.as_uint and \
           (self._cell_norm is not None or self._norm is not None):
            raise RuntimeError

        self._n_bins = 2 ** self._P

        self._method = 'default'

        self._image_reader = ImageReader(color_space='gray')

    @property
    def P(self):
        return self._P

    @property
    def R(self):
        return self._R

    @property
    def method(self):
        return self._method

    @property
    def as_uint(self):
        return self._as_uint

    def _encode(self, img):
        return lbp(img, P=self._P, R=self._R,
                   method=self._method).astype(np.int64)

    def _aggregate(self, cells, fmap):
        def cell_feat_(c):
            cmap = fmap[c[0]: c[0] + self._cell_size[0],
                        c[1]: c[1] + self._cell_size[1]]
            return bincount(cmap.ravel(), minlength=self._n_bins)

        return np.array(map(cell_feat_, cells), dtype=self._dtype)

    def get_params(self, deep=True):
        params = Cellgrid.get_params(self, deep=deep)
        params.update({'P': self._P,
                       'R': self._R,
                       'method': self._method,
                       'n_bins': self._n_bins,
                       'as_uint': self._as_uint})
        return params
