# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from scipy.ndimage import gaussian_filter, correlate1d, correlate

from .image_reader import ImageReader

from .multiscale_feature import MultiscaleFeature

from .utils import sample_uniform_patches

# import colorsys
# _rgb_to_hsv = np.vectorize(colorsys.rgb_to_hsv)


class DCOV(MultiscaleFeature):
    def __init__(self, max_layer=4, downscale=1.414, step=6, patch_size=24,
                 sigma=None, random_state=None):
        ''' Multi-scale covariance descriptors

        Parameters:
        ----------
        max_layer: int, optional
          largest index for a resolution pyramid of (max_layer+1) levels
        downscale: float, optional
          down-scale factor between layer (i+1) and (i)
        step: int, optional
          sampling step along each axis
        patch_size: int, optional
          sampled patches have patch_size x patch_size pixels
        sigma:  float, optional
          Gaussian filters scale. If None, no smoothing is applied
        random_state: vrl.base.RNG, optional
          external RNG. If None, creates its own
        '''
        MultiscaleFeature.__init__(self, max_layer=max_layer,
                                   downscale=downscale,
                                   random_state=random_state)
        self._step = int(step)

        self._patch_size = int(patch_size)

        self._sigma = None if sigma is None else float(sigma)
        if self._sigma is not None and self._sigma < 0.5:
            raise ValueError('sigma values below 0.5 are prone to errors')

        self._image_reader = ImageReader(color_space='gray')
        #self._image_reader = ImageReader(color_space='rgb')

    @property
    def step(self):
        return self._step

    @property
    def patch_size(self):
        return self._patch_size

    @property
    def sigma(self):
        return self._sigma

    def get_params(self, deep=True):
        params = MultiscaleFeature.get_params(self, deep=deep)
        params.update({'step': self.step, 'patch_size': self.patch_size,
                       'sigma': self._sigma})
        return params

    def _transform_layer(self, img, layer=0):
        if img.shape[0] < self.patch_size or img.shape[1] < self.patch_size:
            return None

        # get points sampled uniformly
        loc = sample_uniform_patches(img.shape, self.patch_size, self.step)
        if loc is None:
            return None

        scale = np.sqrt(float(img.shape[0] * img.shape[1]))
        xx, yy = np.meshgrid(np.arange(img.shape[1]) / scale,
                             np.arange(img.shape[0]) / scale, indexing='xy')
        #ss = scale * np.ones(img.shape[:2])
        ss = (self.downscale ** layer) * np.ones(img.shape[:2])

        if self._sigma is not None:
            # compute Gaussian derivatives
            Ix = gaussian_filter(img, sigma=self._sigma, order=(1, 0))
            Iy = gaussian_filter(img, sigma=self._sigma, order=(0, 1))
            Ixx = gaussian_filter(img, sigma=self._sigma, order=(2, 0))
            Iyy = gaussian_filter(img, sigma=self._sigma, order=(0, 2))
            Ixy = gaussian_filter(img, sigma=self._sigma, order=(1, 1))
        else:
            # http://en.wikipedia.org/wiki/Edge_detection#Other_first-order_methods
            Ix = correlate1d(img, [-0.5, 0.0, 0.5], axis=1)
            Iy = correlate1d(img, [-0.5, 0.0, 0.5], axis=0)
            # http://en.wikipedia.org/wiki/Edge_detection#Second-order_approaches_to_edge_detection
            Ixx = correlate1d(img, [1.0, -2.0, 1.0], axis=1)
            Iyy = correlate1d(img, [1.0, -2.0, 1.0], axis=0)
            Ixy = correlate(img, [[-0.25, 0,  0.25],
                                  [ 0,    0,  0   ],
                                  [ 0.25, 0, -0.25]])

        data = np.dstack((xx, yy, ss, img, np.abs(Ix), np.abs(Iy), np.abs(Ixx), np.abs(Iyy), np.abs(Ixy)))
        #data = np.dstack((xx, yy, ss, img, np.abs(Ix), np.abs(Iy), np.abs(Ixx), np.abs(Iyy)))
        eps = np.eye(data.shape[2]) * 2**-23
        ps2 = self._patch_size ** 2
        def patch_descriptor(loc_i):
            data_i = data[loc_i[0]:loc_i[0]+self._patch_size,
                          loc_i[1]:loc_i[1]+self._patch_size]
            cov_i = np.cov(data_i.reshape((ps2, data.shape[2])), rowvar=0)
            return cov_i + eps

        data = np.array([patch_descriptor(loc_i).ravel() for loc_i in loc])

        loc = np.hstack([loc, np.ones((len(loc), 1))])
        loc *= self.downscale ** float(layer)

        return {'data': data, 'loc': loc}
