# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

from os import makedirs
from os.path import join, exists, dirname

import numpy as np

from scipy.ndimage import filters

from ..base import VRLBaseEstimator
from ..io import file_exists, save
from .utils import random_samples, local_maxima_8n
from .hog import HOG
from .bounding_box import BoundingBox


class DiscriminativeParts(VRLBaseEstimator):
    def __init__(self, max_size=int(1e5), **kwargs):
        ''' HOG-based discriminative parts
        '''
        VRLBaseEstimator.__init__(self)

        def arg(key, defval):
            return defval if not key in kwargs else kwargs[key]

        self._hog = HOG(max_layer = arg('max_layer', 6),
                        downscale = arg('downscale', 1.26),
                        patch_size = arg('patch_size', (64, 64)),
                        step = arg('step', 8),
                        cell_size = arg('cell_size', 8),
                        num_orientations = arg('num_orientations', 8),
                        norm = arg('norm', None),
                        grad_threshold = arg('grad_threshold', 0.001))
        self._image_reader = self._hog.image_reader

        if max_size is not None:
            self._hog.image_reader.max_size = int(max_size)

        self._srel = 0.5
        self._max_overlap = 0.5
        self._nsamples = int(1e5)

        # self._params_runtime = ('srel', 'max_overlap', 'nsamples')

        self.background_model_ = None

    @property
    def hog(self):
        return self._hog

    def get_params(self, deep=True):
        params = self._hog.get_params(deep)
        params.update({'max_size': self._hog.image_reader.max_size, 'srel':
                       self._srel, 'max_overlap': self._max_overlap,
                       'background_model_': self.background_model_})
        return params

    def _fit_from_filelist(self, X, y=None):

        # sample non-overlaping random patches
        params = self._hog.get_params()
        params['step'] = max(self._hog.patch_size)
        hog = HOG(**params)
        hog.image_reader.max_size = self._hog.image_reader.max_size
        hog.input_path = self.input_path
        samples = random_samples(hog, X, self._nsamples)

        return self._fit(samples['data'], y)

    def _fit(self, X, y=None):
        from sklearn.covariance import OAS

        # sample covariance
        mu = np.mean(X, axis=0)
        estimator = OAS(store_precision=True, assume_centered=True)
        estimator.fit(X - mu)

        self.background_model_ = {'mu': mu,
                                  'P': estimator.precision_,
                                  'L': np.linalg.cholesky(estimator.precision_)}

    def filter_overlaping_patches(self, parts):
        ''' filter out overlapping patches '''
        from maximal_cliques import find_cliques
        N = 0
        while N != len(parts):
            N = len(parts)

            # build connectivity (overlap) graph
            graph = dict.fromkeys(xrange(N))
            for i in xrange(N):
                pi = parts[i]['bb']
                graph[i] = [j for j in xrange(N) if j != i and pi.overlap(parts[j]['bb']) >= self._max_overlap]

            # find maximal cliques using the Bron-Kerbosch algorithm
            cliques = [tuple(c) for c in find_cliques(graph)]

            # keep the most discriminative patch from each clique
            indx = []
            for clq in cliques:
                i_max = np.argmax([parts[i]['wgh'] for i in clq])
                indx.append(clq[i_max])

            parts = [parts[i] for i in np.unique(indx)]

        return parts

    def compute_part_features(self, im, parts):
        from scipy.misc import imresize

        for i, p in enumerate(parts):
            bb = p['bb']
            i1 = max(0, int(bb.y1 + 0.5))
            j1 = max(0, int(bb.x1 + 0.5))
            i2 = min(int(bb.y2 + 0.5), im.shape[0])
            j2 = min(int(bb.x2 + 0.5), im.shape[1])

            patch = imresize(im[i1:i2, j1:j2], self._hog.patch_size, interp='bilinear')
            pfeat = self._hog.transform(patch)
            parts[i].update({'data': pfeat['data']})

        return parts

    def _transform(self, X):
        '''compute candidate parts as local peaks of an interestingness
        map. Peaks are detected at different scales independently using the
        method of Aubry et al., Painting-to-3D Model Alignment Via
        Discriminative Visual Elements. 2013.
        '''
        # compute HOG features
        feat = self._hog.transform(X)
        loc = feat['loc']
        feat = np.dot(feat['data'] - self.background_model_['mu'], self.background_model_['L'])

        # compute interestingness for each feature
        wnorm = np.sum(np.multiply(feat, feat), axis=1)

        # get local peaks at each scale
        parts = []
        srel = 0.5 * self._srel
        h, w = X.shape[0], X.shape[1]
        scales = np.unique(loc[:, 2])
        for i, sc in enumerate(scales):
            indx = np.where(loc[:, 2] == sc)[0]

            # local coordinates where saved relative to the original size
            x_ = [int(x / sc + 0.5) for x in loc[indx, 0]]
            y_ = [int(y / sc + 0.5) for y in loc[indx, 1]]

            # compute interestingness map
            h_, w_ = int(h / sc + 0.5), int(w / sc + 0.5)
            wgh = np.zeros((h_, w_))
            for j, k in enumerate(indx):
                wgh[y_[j], x_[j]] = wnorm[k]# if wnorm[k] > 0.0 else 0.0
            sigma = (srel * self._hog.patch_size[0],
                     srel * self._hog.patch_size[1])
            wgh = filters.gaussian_filter(wgh, sigma, mode='nearest')

            # find local peaks
            peaks = np.where(local_maxima_8n(wgh))

            # store parts from local peaks
            phw_ = sc * self._hog.patch_size[1] / 2.0
            phh_ = sc * self._hog.patch_size[0] / 2.0
            for j in xrange(len(peaks[0])):
                i0, j0 = peaks[0][j], peaks[1][j]
                x0, y0 = sc * j0, sc * i0  # patch center wrt base scale

                # bounding box info
                x1, y1 = x0 - phw_, y0 - phh_
                x2, y2 = x0 + phw_, y0 + phh_

                parts.append({'x': x0, 'y': y0, 'scale': sc, 'wgh': wgh[i0, j0], 'bb': BoundingBox(x1, y1, x2, y2)})

        if self._max_overlap is not None:
            parts = self.filter_overlaping_patches(parts)

        parts = self.compute_part_features(X, parts)

        loc = np.array([(p['x'], p['y'], p['scale']) for p in parts])
        data = np.array([p['data'].squeeze() for p in parts])
        wgh = np.array([p['wgh'] for p in parts])
        bb = np.array([(p['bb'].x1, p['bb'].y1, p['bb'].x2, p['bb'].y2) for p in parts])

        # WHO features
        data = np.dot(data - self.background_model_['mu'], self.background_model_['L'])

        return {'loc': loc, 'data': data, 'wgh': wgh, 'bb': bb}
