# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll, c_int, c_bool, c_void_p

from numpy.ctypeslib import ndpointer

from .image_reader import ImageReader

from .multiscale_feature import MultiscaleFeature

from ..base import load_libcv
lib = load_libcv()


class DSIFT(MultiscaleFeature):
    def __init__(self, max_layer=4, downscale=1.414, step=8, size=8, norm=2,
                 root=True, fnorm_threshold=None, random_state=None):
        ''' Dense SIFT features (vlfeat implementation, 4x4 cells)

        Parameters:
        ----------
        max_layer: int, optional
          largest index for a resolution pyramid of (max_layer+1) levels
        downscale: float, optional
          down-scale factor between layer (i+1) and (i)
        step: int, optional
          sampling step along each axis
        size: int, optional
          size of a single SIFT cell in pixels
        norm: int or None, optional
          Lp-normalization
        root: bool, optional
          apply the square-root transform dimensionwise
        fnorm_threshold: float, optional
          return only those features with a norm larger than this threshold
        random_state: vrl.base.RNG, optional
          external RNG. If None, creates its own
        '''
        MultiscaleFeature.__init__(self, max_layer=max_layer,
                                   downscale=downscale,
                                   random_state=random_state)

        self._step = int(step)

        self._size = int(size)

        self._norm = None if norm is (None or norm < 0) else int(norm)

        self._root = bool(root)

        if self._root and self._norm != 2:
            self.logger.warning('root=True implies norm=2')
            self._norm = 2

        self._fnorm_threshold = None if fnorm_threshold is None else float(fnorm_threshold)

        self._image_reader = ImageReader(color_space='gray')

        self.__declare_cfn()

        cobj_norm = -1 if self._norm is None else self._norm
        self.__cobj = [None] * (self.max_layer + 1)
        for i in xrange(len(self.__cobj)):
            self.__cobj[i] = self.__cfn_new(self.step, self.size, cobj_norm,
                                            self.root)

    @property
    def step(self):
        return self._step

    @property
    def size(self):
        return self._size

    @property
    def norm(self):
        return self._norm

    @property
    def root(self):
        return self._root

    @property
    def fnorm_threshold(self):
        return self._fnorm_threshold

    def get_params(self, deep=True):
        params = MultiscaleFeature.get_params(self, deep=deep)
        params.update({'step': self._step,
                       'size': self._size,
                       'norm': self._norm,
                       'root': self._root,
                       'fnorm_threshold': self._fnorm_threshold})
        return params

    def _transform_layer(self, X, layer=0):
        h, w = X.shape

        psize = 4 * self.size
        if h < psize or w < psize:
            return None

        cobj = self.__cobj[layer]
        # vlfeat recommends to have values in the range [0, 255]
        self.__cfn_process(cobj, 255. * X.astype(np.float32), w, h)

        nfeat = self.__cfn_nfeat(cobj)
        ndim = self.__cfn_ndim(cobj)

        feat = np.empty((nfeat, ndim), dtype=np.float32)
        self.__cfn_features(cobj, feat)
        feat = feat.astype(np.float64)

        # each row corresponds to: (x, y, norm)
        keyp = np.empty((nfeat, 3), dtype=np.float32)
        self.__cfn_keypoints(cobj, keyp)
        keyp = keyp.astype(np.float64)

        # feature spatial location and scale relative to the base image
        loc = np.array([keyp[:, 0], keyp[:, 1], np.ones(nfeat)])
        loc = np.transpose(loc) * (self._downscale ** float(layer))

        # norm before normalization
        fnorm = keyp[:, 2]

        if self._fnorm_threshold is not None:
            indx = [i for i, val in enumerate(fnorm) if val > self._fnorm_threshold]
            loc = loc[indx]
            fnorm = fnorm[indx]
            feat = feat[indx]

        return {'loc': loc, 'fnorm': fnorm, 'data': feat}

    @classmethod
    def __declare_cfn(cls):
        arr2d_caw = ndpointer(dtype=np.float32, ndim=2, flags='C,A,W')
        arr2d_ca = ndpointer(dtype=np.float32, ndim=2, flags='C,A')

        cls.__cfn_new = lib['DSIFT_new']
        cls.__cfn_new.argtypes = [c_int, c_int, c_int, c_bool]
        cls.__cfn_new.restype = c_void_p

        cls.__cfn_del = lib['DSIFT_del']
        cls.__cfn_del.argtypes = [c_void_p]
        cls.__cfn_del.restype = None

        cls.__cfn_process = lib['DSIFT_Process']
        cls.__cfn_process.argtypes = [c_void_p, arr2d_ca, c_int, c_int]
        cls.__cfn_process.restype = c_int

        cls.__cfn_nfeat = lib['DSIFT_nfeat']
        cls.__cfn_nfeat.argtypes = []
        cls.__cfn_nfeat.restype = c_int

        cls.__cfn_ndim = lib['DSIFT_ndim']
        cls.__cfn_ndim.argtypes = []
        cls.__cfn_ndim.restype = c_int

        cls.__cfn_features = lib['DSIFT_Features']
        cls.__cfn_features.argtypes = [c_void_p, arr2d_caw]
        cls.__cfn_features.restype = None

        cls.__cfn_keypoints = lib['DSIFT_Keypoints']
        cls.__cfn_keypoints.argtypes = [c_void_p, arr2d_caw]
        cls.__cfn_keypoints.restype = None

    def __del__(self):
        for i in xrange(len(self.__cobj)):
            self.__cfn_del(self.__cobj[i])
