# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll, c_int, c_bool, c_void_p

from numpy.ctypeslib import ndpointer

from .image_reader import ImageReader

from .multiscale_feature import MultiscaleFeature

from .utils import sample_uniform_patches, sample_uniform_patches_grad_threshold

from ..base import load_libcv
lib = load_libcv()


class HOG(MultiscaleFeature):
    def __init__(self, max_layer=4, downscale=1.414, patch_size=(80, 80), step=8,
                 cell_size=8, num_orientations=9, norm=None, grad_threshold=None):
        MultiscaleFeature.__init__(self, max_layer=max_layer,
                                   downscale=downscale)

        self._patch_size = tuple(patch_size)

        self._step = int(step)

        self._cell_size = int(cell_size)

        self._num_orientations = int(num_orientations)

        self._norm = None if norm is None else int(norm)

        self._grad_threshold = None if grad_threshold is None else float(grad_threshold)

        self._image_reader = ImageReader(color_space='gray')

        self.__declare_cfn()

        cobj_norm = -1 if self._norm is None else self._norm
        self.__cobj = self.__cfn_new(self.num_orientations, self.cell_size, cobj_norm)

    @property
    def patch_size(self):
        return self._patch_size

    @property
    def step(self):
        return self._step

    @property
    def cell_size(self):
        return self._cell_size

    @property
    def num_orientations(self):
        return self._num_orientations

    @property
    def norm(self):
        return self._norm

    @property
    def grad_threshold(self):
        return self._grad_threshold

    def get_params(self, deep=True):
        params = MultiscaleFeature.get_params(self, deep=deep)
        params.update({'patch_size': self._patch_size,
                       'step': self._step,
                       'cell_size': self._cell_size,
                       'num_orientations': self._num_orientations,
                       'norm': self._norm,
                       'grad_threshold': self._grad_threshold})
        return params

    def _transform_layer(self, X, layer=0):
        h, w = X.shape
        if h < self._patch_size[0] or w < self._patch_size[1]:
            return None

        if self._grad_threshold is None:
            loc = sample_uniform_patches(X.shape, self._patch_size, self._step)
        else:
            loc = sample_uniform_patches_grad_threshold(X, self._patch_size, self._step, self._grad_threshold)

        if len(loc) == 0:
            return None

        data = None
        ph, pw = self._patch_size

        for n in xrange(len(loc)):
            i, j = loc[n]
            patch_i = X[i:i+self._patch_size[0], j:j+self._patch_size[1]]
            ndim = self.__cfn_process(self.__cobj, 255.*patch_i.astype(np.float32), pw, ph)

            if data is None:
                data = np.empty((len(loc), ndim), dtype=np.float32)

            self.__cfn_extract(self.__cobj, data[n].view())

        # ref. to the center of the patch and use (x,y)- instead of (i,j)-coordinates
        loc = loc.astype(np.float)
        loc[:, 0] += 0.5 * self._patch_size[0]
        loc[:, 1] += 0.5 * self._patch_size[1]
        loc = np.fliplr(loc)

        loc = np.hstack([loc, np.ones((len(loc), 1))])
        loc *= self.downscale ** float(layer)

        return {'loc': loc, 'data': data.astype(np.float64)}

    @classmethod
    def __declare_cfn(cls):
        arr1d_caw = ndpointer(dtype=np.float32, ndim=1, flags='C,A,W')
        arr2d_ca = ndpointer(dtype=np.float32, ndim=2, flags='C,A')

        cls.__cfn_new = lib['HOG_new']
        cls.__cfn_new.argtypes = [c_int, c_int, c_int]
        cls.__cfn_new.restype = c_void_p

        cls.__cfn_del = lib['HOG_del']
        cls.__cfn_del.argtypes = [c_void_p]
        cls.__cfn_del.restype = None

        cls.__cfn_process = lib['HOG_Process']
        cls.__cfn_process.argtypes = [c_void_p, arr2d_ca, c_int, c_int]
        cls.__cfn_process.restype = c_int

        cls.__cfn_extract = lib['HOG_Extract']
        cls.__cfn_extract.argtypes = [c_void_p, arr1d_caw]
        cls.__cfn_extract.restype = None

    def __del__(self):
        self.__cfn_del(self.__cobj)
