# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from PIL import Image

import colorsys

_rgb_to_hsv = np.vectorize(colorsys.rgb_to_hsv)
_rgb_to_hls = np.vectorize(colorsys.rgb_to_hls)
_rgb_to_yiq = np.vectorize(colorsys.rgb_to_yiq)


class ImageReader(object):
    def __init__(self, color_space='gray', channels='all', max_size=None):
        '''VRL image handler (to assure images are read in a consistent way)

        Parameters
        ----------
        color_space: string, optional
          color space (\'gray\', \'rgb\' or \'hsv\')
        channels: string, optional
          select one or múltiple channels (by name) to be read. if channels=\'all\', read all channels
        max_size: int, optional
          resize images to a maximum \'max_size\' pixels if larger
        '''

        self._color_space = color_space.lower()
        if self._color_space == 'grey':
            self._color_space = 'gray'

        if not self._color_space in {'gray', 'rgb', 'hsv'}:
            raise ValueError('Invalid color space')

        self.__rgb_to = {'hsv': _rgb_to_hsv, 'hls': _rgb_to_hls, 'yiq': _rgb_to_yiq}

        if isinstance(channels, str):
            self._channels = channels.lower()
        else:
            self._channels = channels

        self.max_size = max_size

    def __call__(self, filename):
        return self.read(filename)

    @property
    def max_size(self):
        return self._max_size

    @max_size.setter
    def max_size(self, val):
        self._max_size = None if val is None else int(val)

    def get_params(self):
        return {'max_size': self._max_size, 'color_space': self._color_space}

    @staticmethod
    def is_readable(filename):
        '''returns True if the image can be read'''
        try:
            Image.open(filename)
        except IOError:
            return False
        return True

    def read(self, filename):
        # read image from disk using PIL
        img = Image.open(filename)
        if self._color_space == 'gray':
            img = img.convert(mode='L')
        else:
            img = img.convert(mode='RGB')

        # resize if necessary
        if self._max_size is not None and \
           img.size[0] * img.size[1] > self._max_size:
            scale = np.sqrt(float(self._max_size) / float(img.size[0] * img.size[1]))
            new_width = int(0.5 + scale * float(img.size[0]))
            new_height = int(0.5 + scale * float(img.size[1]))
            img = img.resize((new_width, new_height), resample=Image.BILINEAR)

        # convert to numpy array and normalize to [0, 1]
        img = np.array(img, dtype=np.float64)
        img /= 255.

        # change color space if necessary
        if not self._color_space in {'gray', 'rgb'}:
            img = self.__rgb_to[self._color_space](img[:,:,0], img[:,:,1], img[:,:,2])
            img = np.dstack(img)

        # get channels
        if not self._channels == 'all':
            img = img[:, :, self._channels].squeeze()

        return img
