# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from skimage.feature import local_binary_pattern as lbp

from ..utils import bitarray

from .image_reader import ImageReader

from .multiscale_feature import MultiscaleFeature


class LBP(MultiscaleFeature):
    def __init__(self, max_layer=4, downscale=1.414, P=8, R=1.0,
                 method='default', as_bitarray=True, random_state=None):
        ''' Local Binary Patterns (skimage.feature.local_binary_pattern)

        Parameters:
        ----------
        max_layer: int, optional
          largest index for a resolution pyramid of (max_layer+1) levels
        downscale: float, optional
          down-scale factor between layer (i+1) and (i)
        P: int, optional
          sampling points
        R: float, optional
          sampling radius
        method: string, optional
          LBP variant (see skimage.feature.local_binary_pattern)
        as_bitaarray: bool, optional
          if True, returns a bit array
        root: bool, optional
          apply the square-root transform dimensionwise
        fnorm_threshold: float, optional
          return only those features with a norm larger than this threshold
        random_state: vrl.base.RNG, optional
          external RNG. If None, creates its own
        '''
        MultiscaleFeature.__init__(self, max_layer=max_layer,
                                   downscale=downscale,
                                   random_state=random_state)

        self._P = int(P)
        if self._P <= 8:
            self._n_bits = 8
        else:
            self._n_bits = 16

        self._R = int(R)

        self._method = str(method)

        self._as_bitarray = bool(as_bitarray)

        self._image_reader = ImageReader(color_space='gray')

    @property
    def P(self):
        return self._P

    @property
    def R(self):
        return self._R

    @property
    def method(self):
        return self._method

    @property
    def as_bitarray(self):
        return self._as_bitarray

    def get_params(self, deep=True):
        params = MultiscaleFeature.get_params(self, deep=deep)
        params.update({'P': self._P, 'R': self._R, 'method': self._method,
                       'as_bitarray': self._as_bitarray})
        return params

    def _transform_layer(self, X, layer=0):
        h, w = X.shape

        psize = 2 * self._R + 1
        if h < psize or w < psize:
            return None

        # lbp features
        feat = lbp(X, P=self.P, R=self.R, method=self._method)

        if self.as_bitarray:
            feat = bitarray(feat.ravel(), self._n_bits)
        else:
            # needed by the vstack in MultiscaleFeature
            feat = feat.reshape(feat.size, 1)

        # keypoints
        xx, yy = np.meshgrid(xrange(w), xrange(h), indexing='xy')
        loc = np.dstack((xx, yy, np.ones(xx.shape))).reshape(xx.size, 3)
        loc *= self.downscale ** float(layer)

        return {'data': feat, 'loc': loc}


from numpy import bincount

from ..base import VRLBaseEstimator

from math import sqrt as msqrt

from math import pow as mpow

@np.vectorize
def s_sqrt(x):
    return -msqrt(-x) if x < 0.0 else msqrt(x)


@np.vectorize
def s_pow(x, n):
    return -mpow(-x, n) if x < 0.0 else mpow(x, n)


class LBPH(VRLBaseEstimator):
    def __init__(self, max_layer=4, downscale=1.414, P=8, R=1.0,
                 method='default', alpha=0.5, norm=2, random_state=None):
        VRLBaseEstimator.__init__(self, random_state=random_state)

        self._lbp = LBP(max_layer=max_layer, downscale=downscale, P=P, R=R,
                        method=method, as_bitarray=False,
                        random_state=random_state)

        self._image_reader = self._lbp.image_reader

        if method == 'default':
            self._n_bins = 2 ** self._lbp.P
        elif method == 'nri_uniform':
            self._n_bins = (2 + self._lbp.P * (self._lbp.P - 1))
        else:
            raise ValueError('only \'default\' and \'nri_uniform\' methods are supported')

        if alpha is None or np.isclose(alpha, 1.0):
            self._alpha = None
            self._alpha_fn = None

        elif np.isclose(alpha, 0.5):
            self._alpha = float(alpha)
            self._alpha_fn = s_sqrt

        else:
            self._alpha = float(alpha)
            self._alpha_fn = s_pow

        self._norm = None if val is None else float(val)

    @property
    def n_bins(self):
        return self._n_bins

    @property
    def alpha(self):
        return self._alpha

    @property
    def norm(self):
        return self._norm

    def get_params(self, deep=True):
        params = self._lbp.get_params(deep=deep)
        params.update({'alpha': self._alpha, 'norm': self._norm})
        return params

    def _fit(self, X, y=None):
        return

    def _transform(self, X):
        data = self._lbp.transform(X)['data'].astype(np.int)

        data = bincount(data.ravel(), minlength=self._n_bins).astype(np.float64)

        # alpha transform
        if self._alpha is not None:
            data = self._alpha_fn(data)

        # Lp-normalization
        if self._norm is not None:
            data /= np.linalg.norm(data, ord=self._norm)

        return data