# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import itertools

import numpy as np

#from PIL import Image

from skimage.transform import pyramid_gaussian

from ..base import VRLBaseEstimator

from ..random import RNG

from .utils import resolution_pyramid


class MultiscaleFeature(VRLBaseEstimator):
    def __init__(self, max_layer=4, downscale=1.414, random_state=None):
        '''Multiscale features base class. Extract features from the layers of
        a resolution pyramid of (max_layer+1) levels and a down-sampling factor
        of downscale.
        '''
        VRLBaseEstimator.__init__(self)

        self._max_layer = int(max_layer)

        self._downscale = float(downscale)
        if self._downscale <= 1.0:
            raise ValueError('downscale must be greater than 1.0')

        self._mask = None

        self._fmap = None

        self._random_state = RNG() if random_state is None else random_state

    @property
    def max_layer(self):
        return self._max_layer

    @property
    def downscale(self):
        return self._downscale

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, val):
        self._mask = None if val is None else val

    @property
    def fmap(self):
        return self._fmap

    @fmap.setter
    def fmap(self, val):
        if val is not None:
            if not isinstance(val, (list, tuple)):
                val = (val, )
            if not all([hasattr(v, 'transform') for v in val]):
                raise ValueError('all mappers must implement \'transform\'')
        self._fmap = val

    @property
    def output_path(self):
        if self._fmap is not None:
            return self._fmap[-1].output_path
        return VRLBaseEstimator.output_path.fget(self)

    def get_params(self, deep=True):
        return {'max_layer': self.max_layer, 'downscale': self.downscale}

    def _fit(self, X, y=None):
        return

    def _transform(self, X):
        # build pyramid
        pyr = list(pyramid_gaussian(X, sigma=None, mode='nearest',
                                    max_layer=self._max_layer,
                                    downscale=self._downscale))

        # pyr = resolution_pyramid(X, downscale=self._downscale, max_layer=self._max_layer)

        data = map(lambda i: self._transform_layer(pyr[i], i), range(len(pyr)))
        # data['data']:  (mandatory) NxM matrix of local descriptors
        # data['loc']:   (optional)  Nx3 matrix with feature locations (x, y, scale)
        # data['fnorm']: (optional)  N vector with the feature norms before normalization

        # filter out empty layers
        data = filter(lambda d: d is not None, data)

        # check if empty in which case return empty data
        if len(data) == 0:
            return {'data': np.empty((0, 0))}

        _data = {}
        for key in data[0].keys():
            # merge all layers from all keys together
            _data[key] = map(lambda d: d[key], data)
            if isinstance(_data[key][0], np.ndarray):
                if _data[key][0].ndim > 1:
                    _data[key] = np.vstack(_data[key])
                else:
                    _data[key] = np.hstack(_data[key])
            elif isinstance(_data[key][0], (tuple, list)):
                _data[key] = list(itertools.chain(*_data[key]))
            else:
                raise ValueError('{} must be an tuple or a matrix'.format(key))

        # remove features for which the mask is not set
        if self._mask is not None and 'loc' in _data.keys():
            if self._mask.shape != X.shape:
                raise RuntimeError('be sure the mask has the same dimensions as the input')
            indx = [i for i, loc in enumerate(_data['loc'])
                    if self._mask[int(loc[1]), int(loc[0])] > 0]
            for key in _data:
                _data[key] = _data[key][indx]

        # map data according the fmap
        if self._fmap is not None:
            for fmap in self._fmap:
                _data['data'] = fmap.transform(_data['data'])

        return _data

    def _transform_layer(self, X):
        raise NotImplementedError
