# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import sys

import os

from os.path import join, exists, dirname

import numpy as np

from scipy import ndimage

from ..base import VRLBaseEstimator

from ..io import file_exists

from ..utils import get_logger

from ..random import RNG


def sample_uniform_patches_grad_threshold(img, patch_size, sampling_step, threshold):
    ''' Sample patches for which the gradient norm within them is higher than a threshold

    Parameters:
    ----------
    patch_size: tuple or int
      size of the patch. If int, use square patches
    sampling_step: tuple or int
      patch sampling step. If int, use regular sampling
    threshold: float
      gradient thresold
    '''

    loc = sample_uniform_patches(img.shape, patch_size, sampling_step)

    # gradient magnitude
    # img_x = ndimage.sobel(img, axis=0, mode='constant')
    # img_y = ndimage.sobel(img, axis=1, mode='constant')
    # norm_grad = np.sqrt(img_x * img_x + img_y * img_y)
    norm_grad = ndimage.gaussian_gradient_magnitude(img, sigma=1.0)

    # integral image
    iimg = np.cumsum(np.cumsum(norm_grad, axis=1), axis=0, dtype=np.float)

    # filter-out low contrast regions
    threshold_ = threshold * float(patch_size[0] * patch_size[1])
    indx = []

    for n, (i, j) in enumerate(loc):
        i0 = i-1
        j0 = j-1

        A = iimg[i0, j0] if i0 >= 0 and j0 >= 0 else 0.0
        B = iimg[i0, j0 + patch_size[1]] if i0 >= 0 else 0.0
        C = iimg[i0 + patch_size[0], j0] if j0 >= 0 else 0.0
        D = iimg[i0 + patch_size[0], j0 + patch_size[1]]

        if (D - B - C + A) >= threshold_:
            indx.append(n)

    return loc[indx]

def sample_uniform_patches(img_size, patch_size, sampling_step):
    ''' Sample patches uniformly on a grid

    Parameters:
    ----------
    patch_size: tuple or int
      size of the patch. If int, use square patches
    sampling_step: tuple or int
      patch sampling step. If int, use regular sampling
    '''
    if not isinstance(patch_size, (list, tuple)):
        psize = (int(patch_size), int(patch_size))
        return sample_uniform_patches(img_size, psize, sampling_step)

    if not isinstance(sampling_step, (list, tuple)):
        sstep = (int(sampling_step), int(sampling_step))
        return sample_uniform_patches(img_size, patch_size, sstep)

    if img_size[0] < patch_size[0] or img_size[1] < patch_size[1]:
        return None

    # number of patches along each direction
    ni = 1 + (img_size[0] - patch_size[0]) // sampling_step[0]
    nj = 1 + (img_size[1] - patch_size[1]) // sampling_step[1]

    # area covered by the sampled patches
    all_patches_i = (ni - 1) * sampling_step[0] + patch_size[0]
    all_patches_j = (nj - 1) * sampling_step[1] + patch_size[1]

    # top-left corner offset
    i0 = (img_size[0] - all_patches_i) // 2
    j0 = (img_size[1] - all_patches_j) // 2

    i = range(i0, img_size[0] - patch_size[0] + 1, sampling_step[0])
    j = range(j0, img_size[1] - patch_size[1] + 1, sampling_step[1])
    ii, jj = np.meshgrid(i, j, indexing='ij')

    return np.dstack((ii, jj)).reshape(ni * nj, 2, order='C')


def random_samples(estimator, filelist, n=int(1e6), ofile=None):
    from ..io import save
    from os import makedirs

    if ofile is None:
        return _random_samples(estimator, filelist, n)

    ofile = estimator.pass_through(ofile)
    if not file_exists(ofile):
        samples = _random_samples(estimator, filelist, n)
        if not exists(dirname(ofile)):
            os.makedirs(dirname(ofile))
        save(ofile, samples)
    else:
        logger = get_logger(__name__)
        logger.info('{} already exists'.format(ofile))
    return ofile


def _random_samples(estimator, filelist, n=int(1e6)):
    '''Compute N samples (using a VRL estimator, e.g. MultiscaleFeature)
    randomly from a set of images.

    Parameters:
    ----------
    estimator: vrl.base.VRLBaseEstimator
      VRL estimator implementing the transform method
    filelist: list or tuple
      list of input files
    n: int, optional
      number of samples to sample

    '''
    logger = get_logger(__name__)

    n = int(n)
    if n < 1:
        raise ValueError('the number of samples must be a positive integer')

    if not isinstance(estimator, VRLBaseEstimator):
        raise TypeError('estimator must be a valid vrl object')

    # shuffle (a copy of) the input filelist
    filelist = filelist[:]
    rng = RNG() if not hasattr(estimator, '_random_state') else estimator._random_state
    rng.shuffle(filelist)

    # read first 10 files to get an idea of the average number of samples per file
    n_files = min(10, len(filelist))
    samples_per_file = 0
    for i, f in enumerate(filelist[:n_files]):
        imgfile = join(estimator.input_path, f)
        if not file_exists(imgfile):
            raise IOError('{} not found'.format(imgfile))
        img = estimator.image_reader.read(imgfile)
        samples_per_file += len(estimator.transform(img)['data'])
    samples_per_file = samples_per_file // n_files

    logger.debug('average number of samples per file: {}'.format(samples_per_file))

    # take a small percentage of points from each file so that the desired number of samples can be sampled
    pct = 0.1
    while n // max(1, int(pct * samples_per_file)) > len(filelist) and pct < 1.0:
        pct += 0.1
    logger.info('taking {:.1f}% of the samples from each file'.format(pct*100.0))

    # approximate number of files that will be processed
    n_files = n // int(pct * samples_per_file)

    data = None

    files = []
    sindx = []

    n_samples = 0
    n_files = 0

    n_samples_rnd = 0

    for i, f in enumerate(filelist):
        imgfile = join(estimator.input_path, f)
        if not file_exists(imgfile):
            raise IOError('{} not found'.format(imgfile))

        # read image
        img = estimator.image_reader.read(imgfile)

        # compute features
        fdata = estimator.transform(img)
        ndata = np.unique([len(val) for val in fdata.values()])
        if ndata.size > 1:
            raise IOError('all data fields must be of the same lenght')
        ndata = int(ndata[0])

        n_files += 1  # update file count
        n_samples += ndata  # update extracted samples count

        # continue if empty
        if ndata == 0:
            continue

        # indices to a subset of the samples
        indx = range(ndata)
        estimator._random_state.shuffle(indx)
        indx = indx[:min(int(pct * ndata + 1), ndata)]

        if data is None:
            # can't use dict.fromkeys(fdata, []) because the init list is
            # shared by all keys (no deep copy)
            data = dict([(key, []) for key in fdata])

        for key in data:
            data[key].append(fdata[key][indx])

        n_samples_rnd += len(indx)  # update random samples count

        files.append(f)  # append image that originated the samples
        sindx.extend(len(indx) * [i])  # set sample indices to that image

        print('\r{} features from {} files'.format(n_samples_rnd, i+1), end='')
        sys.stdout.flush()

        if n_samples_rnd >= n:
            break

    print('\r')
    logger.info('{} samples taken at random from {} files'.format(n_samples_rnd, len(files)))

    indx_prm = estimator._random_state.permutation(len(sindx))
    do_shuffle = False

    # stack all fields
    for key in data:
        if isinstance(data[key][0], np.ndarray):
            if data[key][0].ndim > 1:
                data[key] = np.squeeze(np.vstack(data[key]))
            else:
                data[key] = np.squeeze(np.hstack(data[key]))

        if len(data[key]) == len(sindx):
            data[key] = data[key][indx_prm]
            do_shuffle = True

    if do_shuffle:
        sindx = np.array(sindx)[indx_prm]

    # update fields
    data.update({'samples_per_file': float(n_samples) / n_files, 'files': files, 'sindx': sindx})
    return data


def resolution_pyramid(img, downscale=1.4142, max_layer=4):
    '''Resolution pyramid

    Parameters:
    ----------
    downscale: float, optional
      down-scale factor between layer (i+1) and (i)
    max_layer: int, optional
      largest index for a resolution pyramid of (max_layer+1) levels
    '''
    from PIL import Image

    if img.ndim > 2:
        cpyr = [resolution_pyramid(img[:, :, i], downscale=downscale, max_layer=max_layer)
                for i in xrange(img.shape[2])]
        return [np.dstack([cpyr[j][i] for j in xrange(len(cpyr))])
                for i in xrange(len(cpyr[0]))]

    # build resolution pyramid incrementally while preserving input's dtype
    pyr = [img]  # layer 0 corresponds to the input image
    for i in xrange(1, max_layer+1):
        im_prev = Image.fromarray(pyr[-1])
        new_width = int(1.0 + float(im_prev.size[0]) / downscale)
        new_height = int(1.0 + float(im_prev.size[1]) / downscale)
        im_curr = im_prev.resize((new_width, new_height), resample=Image.BILINEAR)
        if im_curr.size <= 1:
            break
        elif im_curr.mode == 'F':
            pyr.append(np.asarray(im_curr, dtype=np.float64))
        else:
            pyr.append(np.asarray(im_curr))
    return pyr


def local_maxima_8n(arr):
    ''' find local maxima in a 2d array using a 8 neighbour structure'''
    b = np.roll(arr, -1, 0)
    t = np.roll(arr,  1, 0)
    r = np.roll(arr, -1, 1)
    l = np.roll(arr,  1, 1)
    tr = np.roll(t, -1, 1)
    tl = np.roll(t,  1, 1)
    br = np.roll(b, -1, 1)
    bl = np.roll(b,  1, 1)
    return ((arr > t) & (arr > b) & (arr > l) & (arr > r) &
            (arr > tl) & (arr > tr) & (arr > bl) & (arr > br))


def local_maxima_4n(arr):
    ''' find local maxima in a 2d array using a 4 neighbour structure'''
    return ((arr > np.roll(arr, 1, 0)) &
            (arr > np.roll(arr, -1, 0)) &
            (arr > np.roll(arr, 1, 1)) &
            (arr > np.roll(arr, -1, 1)) )
