# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

from os import walk, makedirs

from os.path import exists, abspath, join, commonprefix, sep, dirname, splitext

import numpy as np

import imghdr

from PIL import Image

from sklearn import metrics

from ..random import RNG

from ..utils import get_logger


class Dataset(object):

    def __init__(self, path, random_state=None):
        ''' Dataset handler base class

        Parameters
        ----------
        path:  dataset path
        random_state:  external RNG. If None, creates its own
        '''
        self.logger = get_logger(self.__module__ + '.' + self.__class__.__name__)

        # set image list and base path
        self._set_imlist(path)

        # seed the internal RNG using the name of the dataset
        if random_state is None:
            self._random_state = RNG(seed=abs(hash(self.__class__.__name__)))
        else:
            self._random_state = random_state

        self._samples = ()

    @property
    def random_state(self):
        return self._random_state

    @property
    def imlist(self):
        '''returns the list of images in the dataset relative to its base path,
        so that the full path for the i-th image can be formed as:

        dataset.impath + '/' + dataset.imlist[i]
        '''
        return self._imlist

    @property
    def impath(self):
        '''returns the largest common path for all images in the dataset'''
        return self._impath

    def _pil_readable(self, filename):
        try:
            Image.open(filename)
        except IOError:
            return False
        return True

    def _set_imlist(self, path):
        from . import PIL_SAFE
        from . import PIL_EXTENSIONS

        path = abspath(path)
        if not exists(path):
            raise ValueError('path doesn\'t exists')

        imlist = []
        for dpath, dnames, filenames in walk(path):
            for fname in filenames:
                imfile = join(dpath, fname)

                if PIL_SAFE:
                    # check if the image can be loaded
                    if self._pil_readable(imfile):
                        imlist.append(imfile.replace(path, '')[1:])

                    # if not, check if the file was actually an image => corrupted
                    elif imghdr.what(imfile) is not None:
                        self.logger.error('{} can\'t be read'.format(imfile))
                else:
                    if splitext(imfile)[1] in PIL_EXTENSIONS:
                        imlist.append(imfile.replace(path, '')[1:])

        # common path to all files
        bpath = commonprefix(imlist)
        # if sep is not presetnt, there might be a common prefix anyway. In
        # that case, do not replace anything
        bpath = bpath[:bpath.rfind(sep)+1] if sep in bpath else ''

        self._impath = abspath(join(path, bpath))
        self._imlist = sorted([im.replace(bpath, '') for im in imlist])

        self.logger.info('{} images found'.format(len(self._imlist)))


class ClassificationDataset(Dataset):
    def __init__(self, path, random_state=None):
        ''' Image classification dataset handler base class

        Parameters
        ----------
        path:  dataset path
        random_state:  external RNG. If None, create its own
        '''
        Dataset.__init__(self, path, random_state)
        self._classes = ()
        self._samples = {}

    @property
    def classes(self):
        return self._classes

    @property
    def nclasses(self):
        return len(self._classes)

    def get_images(self, set_):
        '''return the images associated to a given set'''
        if set_ not in self._samples.keys():
            raise RuntimeError('{} is not a valid set'.format(set_))
        return [entry[0] for entry in self._samples[set_]]

    def get_labels(self, set_):
        '''return the labels from the images associated to a given set'''
        if set_ not in self._samples.keys():
            raise RuntimeError('{} is not a valid set'.format(set_))
        return [entry[1] for entry in self._samples[set_]]

    def _labels_from_path(self, sep_, list_=None):
        '''class name encoded as CLASS<sep>FILENAME. Only valid for single
        label problems'''
        if not isinstance(list_, (tuple, list)) and list_ is not None:
            raise ValueError('list_ must be a tuple, list or None')

        list_ = self._imlist if list_ is None else list_

        labels = []
        for imfile in list_:
            cls = filter(lambda c: c + sep_ in imfile, self._classes)
            if len(cls) > 1:
                cls = [max(cls, key=len)]
            labels.append((self._classes.index(cls[0]),))
        return tuple(labels)

    def _n_per_class_split(self, n_pos, n_neg, labeled_samples):
        if not isinstance(labeled_samples, (list, tuple)) or \
           not isinstance(labeled_samples[0], (list, tuple)) or \
           not len(labeled_samples[0]) > 1:
            raise ValueError

        # take n_pos samples per class as positives
        n_pos = int(n_pos)
        if n_pos <= 0:
            raise ValueError

        # if n_neg < 0, make it large in order to take the rest as negatives,
        # otherwise take just n_neg per class as negatives
        n_neg = int(n_neg) if n_neg > 0 else len(self._imlist)

        self._random_state.shuffle(labeled_samples)

        train = []
        test = []
        for c in xrange(self.nclasses):
            samples_c = filter(lambda x: c in x[1], labeled_samples)
            i1 = min(n_pos, len(samples_c))
            i2 = min(i1+n_neg, len(samples_c))
            train += samples_c[:i1]
            test += samples_c[i1:i2]

        self._random_state.shuffle(train)
        self._random_state.shuffle(test)

        return train, test

    def generate_split(self):
        '''each dataset generate train/test splits in a particular way. This
        has to be defined accordingly in the derived class

        '''
        raise NotImplementedError

    def eval(self, scr, y_true, outfile=None, **kwargs):
        '''compute accuracy. For multiclass, the \'argmax\' rule is used. If
        rule=\'score\', calssification is based on a threshold on the score
        values (default: 0.0). In the later case, you can also select the
        threshold level by setting \'threshold\'=VALUE.  If \'outfile\' is not
        None, the classificatio report is written to the given file.

        '''
        if not isinstance(scr, np.ndarray):
            raise IOError('scr must be a matrix')

        if scr.shape[0] != len(y_true):
            raise ValueError('scores and labels mismatch')

        if scr.shape[1] != self.nclasses:
            raise ValueError('wrong number of scores per sample')

        rule = kwargs['rule'] if 'rule' in kwargs else 'argmax'
        if rule == 'argmax':
            y_pred = map(lambda s: (s.argmax(),), scr)
        elif rule == 'score':
            threshold = kwargs['threshold'] if 'threshold' in kwargs else 0.0
            y_pred = map(lambda s: tuple(np.where(s >= threshold)[0]), scr)
        else:
            raise ValueError('{} is not a valid rule'.format(rule))

        rep = metrics.classification_report(y_true, y_pred, target_names=self.classes)
        acc = metrics.accuracy_score(y_true, y_pred, normalize=True)
        rep += '\n---\naccuracy: {:.2f}\n'.format(acc*100.)

        if outfile is not None:
            if not exists(dirname(outfile)):
                makedirs(abspath(dirname(outfile)))
            f = open(outfile, 'w')
            f.write(rep + '\n')
            f.close()
            self.logger.info('classification report: {}'.format(outfile))
        else:
            self.logger.info('classification report: \n\n' + rep)
