# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

from os import walk, makedirs, popen

from os.path import abspath, pardir, join, exists, splitext, curdir, dirname, \
    extsep, basename

import numpy as np

from .dataset import ClassificationDataset


class PascalVOC(ClassificationDataset):
    ''' Pascal VOC datasets'''
    def __init__(self, path, test_set='val', random_state=None):
        '''
        Parameters
        ----------
        path:  dataset path
        test_set:  test set (\'train\' or \'val\')
        random_state:  external RNG. If None, create its own
        '''
        path = abspath(path)
        if not exists(path):
            raise ValueError('path doesn\'t exists')

        # walk the directory tree until the images are found
        for dpath, dnames, filenames in walk(path):
            if 'JPEGImages' in dpath:
                break

        ClassificationDataset.__init__(self, path=dpath,
                                       random_state=random_state)

        classes = ('aeroplane', 'bicycle', 'bird', 'boat', 'bottle', 'bus',
                   'car', 'cat', 'chair', 'cow', 'diningtable', 'dog', 'horse',
                   'motorbike', 'person', 'pottedplant', 'sheep', 'sofa',
                   'train', 'tvmonitor')

        self._classes = tuple(sorted(classes))

        self.test_set = test_set

        # Pascal VOC paths
        path = abspath(join(dpath, pardir))
        self._path_Devkit = abspath(join(path, pardir))
        self._path_JPEGImages = dpath
        self._path_ImageSets = join(path, 'ImageSets')
        self._path_ImageSets_Main = join(self._path_ImageSets, 'Main')
        self._path_Annotations = join(path, 'Annotations')
        self._path_VOCcode = join(self._path_Devkit, 'VOCcode')

        def _labeled_samples(set_):
            listfile = join(self._path_ImageSets_Main, set_ + '.txt')
            images = [l.strip() + '.jpg' for l in open(listfile).readlines()]
            if not (set(images)).issubset(set(self._imlist)):
                not_found = set(images).difference(set(self._imlist))
                raise RuntimeError('Dataset corrupted. '
                                   '{} not found'.format(tuple(not_found)))

            classes = []
            for c in self._classes:
                clsfile = join(self._path_ImageSets_Main,
                               '{}_{}.txt'.format(c, set_))
                classes.append([int(l.strip().split()[1])
                               for l in open(clsfile).readlines()])

            # label matrix: 1 column per class
            classes = np.transpose(np.array(classes, dtype=int))
            # treat difficult samples as negatives
            classes[classes < 1] = 0
            # for each row (image), get the indices to the columns (classes) of
            # the classes matrix that were tagged as positives
            classes = [tuple(np.where(lbl)[0].tolist()) for lbl in classes]

            return zip(images, classes)

        # set VOC splits
        self._samples = {'train': _labeled_samples('train'),
                         'val': _labeled_samples('val'),
                         'trainval': _labeled_samples('trainval'),
                         'test': _labeled_samples('test')}
    @property
    def test_set(self):
        return self._test_set

    @test_set.setter
    def test_set(self, val):
        self._test_set = val
        if not self._test_set in ('val', 'test'):
            raise ValueError('test_set must be either \'val\' or \'test\'')

    def eval(self, scr, y_true, outfile=None, **kwargs):
        if not isinstance(scr, np.ndarray):
            raise IOError('scr must be a matrix')

        if scr.shape[0] != len(y_true):
            raise ValueError('scores and labels mismatch')

        if scr.shape[1] != self.nclasses:
            raise ValueError('wrong number of scores per sample')

        if y_true != self.get_labels(self._test_set):
            raise ValueError('the labels don\'t correspond to a valid set')

        path = abspath(curdir if outfile is None else dirname(outfile))
        if not exists(path):
            makedirs(path)

        vocfile_prototype = join(path, 'comp1_cls_test_{}.txt')

        x_test = [splitext(im)[0] for im in self.get_images(self._test_set)]

        # results file in VOC format
        for i, cls in enumerate(self._classes):
            scr_c = ['{:.4e}\n'.format(s) for s in scr[:, i]]
            fobj = open(vocfile_prototype.format(cls), 'w')
            fobj.writelines([' '.join(lin) for lin in zip(x_test, scr_c)])
            fobj.close()

        mAP_ = self._voc_evalcls(path)

        if outfile is not None:
            if not exists(dirname(outfile)):
                makedirs(abspath(outfile))
            f = open(outfile, 'w')
            f.write(mAP_)
            f.close()
        else:
            self.logger.info(mAP_)


    def _voc_evalcls(self, respath):
        OCTCMD = 'octave'
        from distutils import spawn
        if spawn.find_executable(OCTCMD) is None:
            return '\n\'{}\' command not found. Can\'t compute score\n'.format(OCTCMD)

        sc = ('addpath(\'_CODE_\');\n'
              'VOCinit;\n'
              'VOCopts.clsimgsetpath = \'_MAIN_/%s_%s.txt\';\n'
              'VOCopts.clsrespath = \'_RESULTS_/%s_cls_test_%s.txt\';\n'
              'VOCopts.testset = \'_TEST_\';\n'
              'ap = zeros(1, VOCopts.nclasses);\n'
              'for i=1:VOCopts.nclasses,\n'
              '  cls = VOCopts.classes{i};\n'
              '  [R, P, ap(i)] = VOCevalcls(VOCopts, \'comp1\', cls, false);\n'
              '  printf( \'%12s %3.1f\\n\', cls, 100 * ap(i));\n'
              'endfor\n'
              'printf(\'%12s %2.2f\\n\', \'mAP:\', 100 * mean(ap));\n')
        sc = sc.replace('_CODE_', self._path_VOCcode)
        sc = sc.replace('_MAIN_', self._path_ImageSets_Main)
        sc = sc.replace('_RESULTS_', respath)
        sc = sc.replace('_TEST_', self._test_set)

        cmd = '{} --no-window-system --no-history --silent --eval=\"{}\"'.format(OCTCMD, sc)
        # the original script prints some progress info.
        # we remove these lines to avoid clutter
        retlines = filter(lambda line: ' pr: ' not in line,
                          popen(cmd).read().splitlines())
        return '\n'.join(retlines) + '\n'

    # def annotations_mask(self, imfile):
    #     import xmltodict
    #
    #     if not exists(join(self.impath, imfile)):
    #         raise RuntimeError('{} do not correspond to a valid',
    #                            'sample'.format(join(self.impath, imfile)))
    #
    #     xmlfile = join(self._path_Annotations,
    #                    imfile.replace('jpg', 'xml'))
    #     xmldata = xmltodict.parse(open(xmlfile).read())
    #     mask = np.zeros((int(xmldata['annotation']['size']['height']),
    #                      int(xmldata['annotation']['size']['width'])))
    #
    #     obj = xmldata['annotation']['object']
    #     if not isinstance(obj, (list, tuple)):
    #         obj = [obj,]
    #
    #     for obj_ in obj:
    #         xmin = int(obj_['bndbox']['xmin'])
    #         xmax = int(obj_['bndbox']['xmax'])
    #         ymin = int(obj_['bndbox']['ymin'])
    #         ymax = int(obj_['bndbox']['ymax'])
    #         mask[ymin:ymax+1, xmin:xmax+1] = 1
    #
    #     return mask


class VOC2007(PascalVOC):
    ''' Pascal VOC 2007 dataset'''

    def __init__(self, path, test_set='val', year='2007', random_state=None):
        PascalVOC.__init__(self, path=path, test_set=test_set, random_state=random_state)

        if len(self._imlist) != 9963:
            raise RuntimeError('Dataset corrupted. Wrong number of images')
