# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

import scipy.io as sio

from types import IntType, LongType, FloatType

from importlib import import_module

from os.path import exists, isfile, getsize, getmtime

from .utils import get_logger
logger = get_logger(__name__)

VRL_EXT = 'dat'  # default VRL extension


def file_exists(filename):
    ''' check if a file exists and if it size is nonzero'''
    if exists(filename) and isfile(filename) and getsize(filename) > 0:
        return True
    return False


def newer(file1, file2):
    '''returns True if file1 is newer than file2'''
    return getmtime(file1) > getmtime(file2)


def save(filename, data):
    '''save data'''
    if isinstance(data, dict):
        return _savemat(filename, data)

    elif isinstance(data, (np.ndarray, IntType, LongType, FloatType)):
        return _savemat(filename, {'data': data})

    #elif isinstance(VRLBaseEstimator):
    elif hasattr(data, 'get_params'):
        return _saveobj(filename, data)

    else:
        raise NotImplementedError('not a valid VRL type / object')


def load(filename, stack=False, **kwargs):
    '''load data. If the input is a list of files, passing stack=True returns
    all data from all files stacked into a single matrix; otherwhise, it returns
    a list of matrices, one for each file on the input list.
    '''
    if isinstance(filename, (list, tuple)):
        return _loadlst(filename, stack=stack)

    if not file_exists(filename):
        raise IOError('{} not found'.format(filename))

    keys = [var[0] for var in _whosmat(filename)]  # names of the variables
    if 'vrl_class' in keys:
        return _loadobj(filename)

    mdict = _loadmat(filename, **kwargs)

    # remove non-object keys
    for key in mdict.keys():
        if key.startswith(('_')):
            del mdict[key]

    # if just data, return value
    if 'data' in mdict and len(mdict) == 1:
        return mdict['data']

    return mdict


def _loadlst(filelist, stack):
    '''load data from a list of files. If stack=True, all data is stacked into
    a single matrix.
    '''
    data = []
    for fname in filelist:
        data.append(load(fname))

    # if data already an array or nothing to stack, return
    if not stack:
        return data if len(data) > 1 else data[0]

    # if list of ndarrays, return stacked array
    if isinstance(data[0], np.ndarray):
        return np.vstack(data)

    # if dictionary, stack each field independently
    data_ = dict.fromkeys(data[0])
    for key in data_.keys():
        data_[key] = [d[key] for d in data]
        if isinstance(data_[key][0], np.ndarray):
            data_[key] = np.vstack(data_[key])

    return data_


def read_shape(filename, key):
    '''reads the shape of an array field w/o loading it entirely'''
    assert(isinstance(key, str))
    info = _whosmat(filename)
    for i in info:
        if key in i:
            return i[1]
    return None


def _whosmat(filename):
    '''get info'''
    return sio.whosmat(filename, appendmat=False, squeeze_me=True)


def _is_obj(x):
    '''returs True if x is a valid VRL object'''
    return hasattr(x, 'get_params') and hasattr(x, 'set_params')


def _is_ndobj(x):
    '''returns True if x is a structured array (Matlab/Octave style)'''
    return isinstance(x, np.ndarray) and x.dtype.fields is not None


def _saveobj(filename, obj):
    '''saves a VRL object'''
    if not _is_obj(obj):
        raise ValueError('object is not writteable')

    def _obj_to_dict(obj):
        '''returns a dictonary representing an object so that it can be stored
        and retrieved properly from files'''
        odict = obj.get_params(deep=True)
        odict.update({'vrl_module': obj.__module__,
                      'vrl_class': obj.__class__.__name__})
        for key in odict.keys():
            # recursive call to account for nested objects
            if _is_obj(odict[key]):
                odict[key] = _obj_to_dict(odict[key])
            # sklearn nested objects store params in a "key__subkey" format
            elif '__' in key:
                del odict[key]
        return odict

    odict = _obj_to_dict(obj)
    return _savemat(filename, odict)


def _loadobj(filename):
    '''loads a VRL object'''

    # # instantiate a sklearn object
    # def _dict_to_sklearn_obj(mdict):
    #     skl_module = import_module(mdict['skl_module'])
    #     skl_class = getattr(skl_module, mdict['skl_class'])
    #     skl_model_params = {}
    #     for key in mdict.keys():
    #         if key.startswith(('__', 'skl_')):
    #             del mdict[key]
    #         elif key.endswith('_'):
    #             skl_model_params.update({key: mdict[key]})
    #             del mdict[key]
    #         elif isinstance(mdict[key], dict) and 'skl_class' in mdict[key]:
    #             mdict[key] = _dict_to_sklearn_obj(mdict[key])

    #     obj = skl_class(**mdict)
    #     # for key, val in skl_model_params.items():
    #     #     setattr(obj, key, val)
    #     return obj

    # instantiates a vrl object
    def _dict_to_obj(mdict):
        vrl_module = import_module(mdict['vrl_module'])
        vrl_class = getattr(vrl_module, mdict['vrl_class'])
        params_ = {}
        for key, val in mdict.items():
            if key.startswith(('__', 'vrl_')):
                del mdict[key]
            elif key.endswith('_'):
                params_[key] = val
                del mdict[key]
            elif isinstance(val, dict) and 'vrl_class' in val:
                mdict[key] = _dict_to_obj(val)
            # elif isinstance(mdict[key], dict) and 'skl_class' in mdict[key]:
            #     mdict[key] = _dict_to_sklearn_obj(mdict[key])
        obj = vrl_class(**mdict)
        if len(params_) > 0:
            obj.set_params(**params_)
        return obj

    mdict = _loadmat(filename)

    # if SKLEstimator, just return mdict with parameters
    if mdict['vrl_class'] == 'SKLEstimator':
        for key, val in mdict.items():
            if key.startswith(('__', 'vrl_')):
                del mdict[key]
        return mdict

    return _dict_to_obj(mdict)


def _savemat(filename, mdict):
    '''save data (MATLAB/Octave compatible)'''

    # make writeable all values that can't be stored as regular numpy arrays
    def _encode(mdict):
        for key, val in mdict.items():
            if val is None:
                mdict[key] = 'None'
            elif isinstance(val, dict):
                mdict[key] = _encode(val)
            elif isinstance(val, (list, tuple)) and isinstance(val[0], dict):
                mdict[key] = map(_encode, val)
        return mdict

    mdict = _encode(mdict)

    return sio.savemat(filename, mdict, appendmat=False, do_compression=True,
                       oned_as='row')


def _loadmat(filename, **kwargs):
    '''load data (MATLAB/Octave compatible)'''

    # force ndarrays to be row-aligned
    def _force_alignment(x):
        if isinstance(x, np.ndarray):
            shape = x.shape
            x = x.ravel() if x.flags['ALIGNED'] else x.flatten()
            x.shape = shape
        elif isinstance(x, (list, tuple)):
            return map(_force_alignment, x)
        return x

    # returns the dict equivalent of a structured array
    def _ndobj_to_dict(obj):
        odict = {}
        for key in obj.dtype.fields.keys():
            val = obj[key].item()
            if _is_ndobj(val):
                odict[key] = _ndobj_to_dict(val)
            else:
                odict[key] = _force_alignment(val)
        return odict

    # restore values that were forced to be writeable
    def _restore(mdict):
        for key, val in mdict.items():
            if isinstance(val, basestring) and val == 'None':
                mdict[key] = None
            elif isinstance(val, np.ndarray) and np.all(val == 'None'):
                mdict[key] = None
            elif _is_ndobj(val):
                mdict[key] = _restore(_ndobj_to_dict(val))
            elif isinstance(val, dict):
                mdict[key] = _restore(val)
        return mdict

    # tuples do not have the "remove" method used in sio.load
    if 'variable_names' in kwargs and \
       isinstance(kwargs['variable_names'], tuple):
        kwargs['variable_names'] = list(kwargs['variable_names'])

    try:
        mdict = sio.loadmat(filename, appendmat=False, squeeze_me=True, **kwargs)
    except:
        logger.error('error while loading {}'.format(filename))
        raise

    mdict = _restore(mdict)

    for key in mdict.keys():
        # an ndarray representation of structured data
        if _is_ndobj(mdict[key]):
            mdict[key] = _ndobj_to_dict(mdict[key])
        else:
            mdict[key] = _force_alignment(mdict[key])

    return mdict
