# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import c_double, c_void_p

from numpy.ctypeslib import ndpointer

from .exponential_family import lib, ExponentialFamily


class Bernoulli(ExponentialFamily):

    def __init__(self, nvar, min_phit=1e-4, random_state=None):
        ''' Multivariate Bernoulli distribution

        Parameters
        ----------
        nvar:  number of input dimensions
        min_phit:  min allowed \"hit\" probability
        '''
        ExponentialFamily.__init__(self, nvar=nvar, npar=nvar,
                                   random_state=random_state, dtype=np.uint8)

        # create object
        self._cobj = self._cfn_new(self._nvar)

        # set optional parameters after model update
        self.min_phit = min_phit

    @property
    def min_phit(self):
        return self._cfn_min_phit(self._cobj)

    @min_phit.setter
    def min_phit(self, val):
        if float(val) < 0.0:
            raise ValueError('min_phit must be non-negative')
        self._cfn_set_min_phit(self._cobj, float(val))

    @property
    def phit(self):
        ''' \"hit\" probabilities '''
        phit = np.empty(self.nvar)
        self._cfn_phit(self._cobj, phit)
        return phit

    def get_params(self, deep=True):
        params = ExponentialFamily.get_params(self, deep=deep)
        params.update({'min_phit': self.min_phit})
        return params

    def _declare_cfn(self):
        ExponentialFamily._declare_cfn(self)

        arr1d_caw = ndpointer(dtype=np.float64, ndim=1, flags='C,A,W')

        self._cfn_min_phit = lib['Bernoulli_min_phit']
        self._cfn_min_phit.argtypes = [c_void_p]
        self._cfn_min_phit.restype = c_double

        self._cfn_set_min_phit = lib['Bernoulli_set_min_phit']
        self._cfn_set_min_phit.argtypes = [c_void_p, c_double]
        self._cfn_set_min_phit.restype = None

        self._cfn_phit = lib['Bernoulli_phit']
        self._cfn_phit.argtypes = [c_void_p, arr1d_caw]
        self._cfn_phit.restype = None
