# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll, c_double, c_void_p

from .bernoulli import Bernoulli

from .exponential_family_mixture import lib, ExponentialFamilyMixture


class BernoulliMixture(ExponentialFamilyMixture):

    def __init__(self, ncomp=1, nvar=1, min_pst=1e-4, min_phit=1e-6,
                 random_state=None):
        ''' Multivariate Bernoulli mixture

        Parameters
        ----------
        ncomp:  number of mixture components
        nvar:  number of input dimensions
        min_pst:  posteriors below min_pst are set to zero
        min_phit:  min allowed \"hit\" probability
        '''
        ExponentialFamilyMixture.__init__(self, distr=Bernoulli, ncomp=ncomp,
                                          random_state=random_state,
                                          dtype=np.uint8)

        # create object
        self._cobj = self._cfn_new(ncomp, nvar)

        # set optional parameters
        self.min_pst = min_pst
        self.min_phit = min_phit

    @property
    def min_phit(self):
        return self._cfn_min_phit(self._cobj)

    @min_phit.setter
    def min_phit(self, val):
        if float(val) < 0.0:
            raise ValueError('min_phit must be non-negative')
        self._cfn_set_min_phit(self._cobj, float(val))

    def get_params(self, deep=True):
        params = ExponentialFamilyMixture.get_params(self, deep=deep)
        params.update({'min_phit': self.min_phit})
        return params

    def _empty_component(self):
        return Bernoulli(nvar=self.nvar, min_phit=self.min_phit,
                         random_state=self._random_state)

    def compute_initial_estimate(self, data, **kwargs):
        ''' compute and set initial parameter estimate '''
        nsamples, nvar = data.shape
        if nvar != self.nvar:
            raise ValueError('input and model dimensionality must be the same'
                             '(model: {}, data: {})'.format(self.nvar, nvar))

        beta = 0.5 if 'beta' not in kwargs else kwargs['beta']

        # see C++ default value on src/ml/bernoulli.h
        min_phit = self.min_phit
        max_pmiss = 1.0 - min_phit

        self.logger.info('Initializing mixture components ...')

        comp_npar = len(self.etas_[0])
        weights_init = np.ones(self.ncomp) / float(self.ncomp)
        etas_init = np.zeros((self.ncomp, comp_npar))

        indx = range(nsamples)

        for k in xrange(self.ncomp):
            u = (self._random_state.rand(self.nvar)).clip(min_phit, max_pmiss)
            x = data[self._random_state.choice(indx)]
            p = beta * u + (1.0 - beta) * x
            etas_init[k] = np.log(p / (1.0 - p))

        self.weights_ = weights_init
        self.etas_ = etas_init

        self.logger.info('done')

    def _declare_cfn(self):
        ExponentialFamilyMixture._declare_cfn(self)

        self._cfn_min_phit = lib['BernoulliMixture_min_phit']
        self._cfn_min_phit.argtypes = [c_void_p]
        self._cfn_min_phit.restype = c_double

        self._cfn_set_min_phit = lib['BernoulliMixture_set_min_phit']
        self._cfn_set_min_phit.argtypes = [c_void_p, c_double]
        self._cfn_set_min_phit.restype = None
