# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll

from ctypes import c_void_p

from numpy.ctypeslib import ndpointer

from .exponential_family import lib, ExponentialFamily


class Dirichlet(ExponentialFamily):

    def __init__(self, nvar, random_state=None):
        ''' Dirichlet distribution

        Parameters
        ----------
        nvar:  number of input dimensions
        '''
        ExponentialFamily.__init__(self, nvar=nvar, npar=nvar,
                                   random_state=random_state, dtype=np.float64)

        # create object
        self._cobj = self._cfn_new(self._nvar)

    @property
    def alpha(self):
        alpha = np.empty(self.nvar)
        self._cfn_alpha(self._cobj, alpha)
        return alpha

    def _fit(self, X, y=None):
        '''parameter estimation using Minka's method. It redefines the one
        provided in the base class because the iterative procedure needs an
        initial estimate which is based on the first and second moments of
        the samples

        '''
        if self._dtype != X.dtype:
            raise TypeError

        # initial estimate using the method of moments
        E = np.mean(X, axis=0)
        E2 = np.mean(X**2, axis=0)
        if np.abs(E2[0] - E[0]**2) < 2**-23:
            alpha = np.zeros(E.shape)
        else:
            alpha = ((E[0] - E2[0]) / (E2[0] - E[0]**2)) * E
        #alpha[-1] = (1.0 - np.sum(alpha[:-1])) / E[-1]

        # initial guess
        self.eta_ = alpha - 1.0

        # fit
        ExponentialFamily._fit(self, X)

    def _declare_cfn(self):
        ExponentialFamily._declare_cfn(self)

        arr1d_caw = ndpointer(dtype=np.float64, ndim=1, flags='C,A,W')

        self._cfn_alpha = lib['Dirichlet_alpha']
        self._cfn_alpha.argtypes = [c_void_p, arr1d_caw]
        self._cfn_alpha.restype = None
