# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll, c_int, c_double, c_void_p

from numpy.ctypeslib import ndpointer

from ..base import VRLBaseEstimator

from ..base import load_libml
lib = load_libml()


class ExponentialFamily(VRLBaseEstimator):
    ''' Exponential family pdf base class'''
    def __init__(self, nvar=1, npar=1, random_state=None, dtype=np.float64):
        '''
        Parameters
        ----------
        nvar:  number of input variables (dimensions)
        npar:  number of parameters
        random_state:  RNG
        dtype:  type of the input variables
        '''
        VRLBaseEstimator.__init__(self, random_state=random_state)

        self._nvar = int(nvar)
        if self._nvar <= 0:
            raise ValueError('nvar must be a positive integer')

        self._npar = int(npar)
        if self._npar <= 0:
            raise ValueError('npar must be a positive integer')

        self._dtype = dtype

        self._declare_cfn()

        self._cobj = None

    @property
    def npar(self):
        return self._npar

    @property
    def nvar(self):
        return self._nvar

    @property
    def eta_(self):
        ''' get natural parameters '''
        eta = np.ndarray(self.npar)
        self._cfn_eta(self._cobj, eta)
        return eta

    @eta_.setter
    def eta_(self, val):
        ''' set natural parameters '''
        if not isinstance(val, np.ndarray):
            raise TypeError
        if val.size != self.npar:
            raise ValueError
        self._cfn_set_eta(self._cobj, val)

    def get_params(self, deep=True):
        return {'nvar': self._nvar, 'npar': self._npar}

    def sample(self, n=1):
        ''' sample from the distribution '''
        if n <= 0:
            raise ValueError

        X = np.empty((n, self.nvar), dtype=self._dtype)
        self._cfn_Sample(self._cobj, X, n)

        return X.squeeze()

    def log_likelihood(self, x):
        ''' computes sample log-likelihood '''
        return self._cfn_LogLikelihood(self._cobj, x)

    def suff_stat(self, x, out=None):
        ''' computes sufficient statistics from a sample '''
        if out is None:
            out = np.empty(self.npar)
        self._cfn_SuffStat(self._cobj, x, out)
        return out

    def seed(self, val):
        ''' sets internal object seed '''
        return self._cfn_Seed(self._cobj, x)

    def score(self, X, out=None):
        '''Fisher score for sample
        '''
        if out is None:
            out = np.empty(self.npar)
        if X.ndim == 1:
            self._cfn_Score(self._cobj, X.reshape(1, len(X)), 1, out)
        else:
            self._cfn_Score(self._cobj, X, len(X), out)
        return out

    def _fit(self, X, y=None):
        ''' fit parameters from sample '''
        X_ = X
        if self._dtype != X.dtype:
            if not np.can_cast(X.dtype, self._dtype, casting='safe'):
                raise TypeError
            X_ = X.astype(self._dtype)

        stat = np.zeros(self.npar)
        for x in X_:
            stat += self.suff_stat(x)
        self._cfn_FitFromSuffStat(self._cobj, stat, len(X_))

    def _transform(self, X):
        return X

    def _declare_cfn(self):
        arr1d_ca = ndpointer(dtype=np.float64, ndim=1, flags='C,A')
        arr1d_caw = ndpointer(dtype=np.float64, ndim=1, flags='C,A,W')
        arr1d_t_ca = ndpointer(dtype=self._dtype, ndim=1, flags='C,A')
        arr2d_t_ca = ndpointer(dtype=self._dtype, ndim=2, flags='C,A')
        arr2d_t_caw = ndpointer(dtype=self._dtype, ndim=2, flags='C,A,W')

        dname = self.__class__.__name__

        self._cfn_new = lib[dname + '_new']
        self._cfn_new.argtypes = [c_int]
        self._cfn_new.restype = c_void_p

        self._cfn_del = lib[dname + '_del']
        self._cfn_del.argtypes = [c_void_p]
        self._cfn_del.restype = None

        self._cfn_npar = lib[dname + '_npar']
        self._cfn_npar.argtypes = [c_void_p]
        self._cfn_npar.restype = c_int

        self._cfn_nvar = lib[dname + '_nvar']
        self._cfn_nvar.argtypes = [c_void_p]
        self._cfn_nvar.restype = c_int

        self._cfn_eta = lib[dname + '_eta']
        self._cfn_eta.argtypes = [c_void_p, arr1d_caw]
        self._cfn_eta.restype = None

        self._cfn_set_eta = lib[dname + '_set_eta']
        self._cfn_set_eta.argtypes = [c_void_p, arr1d_ca]
        self._cfn_set_eta.restype = None

        self._cfn_SuffStat = lib[dname + '_SuffStat']
        self._cfn_SuffStat.argtypes = [c_void_p, arr1d_t_ca, arr1d_caw]
        self._cfn_SuffStat.restype = None

        self._cfn_FitFromSuffStat = lib[dname + '_FitFromSuffStat']
        self._cfn_FitFromSuffStat.argtypes = [c_void_p, arr1d_ca, c_double]
        self._cfn_FitFromSuffStat.restype = None

        self._cfn_grad_psi = lib[dname + '_grad_psi']
        self._cfn_grad_psi.argtypes = [c_void_p, arr1d_caw]
        self._cfn_grad_psi.restype = None

        self._cfn_psi = lib[dname + '_psi']
        self._cfn_psi.argtypes = [c_void_p]
        self._cfn_psi.restype = c_double

        self._cfn_LogLikelihood = lib[dname + '_LogLikelihood']
        self._cfn_LogLikelihood.argtypes = [c_void_p, arr1d_t_ca]
        self._cfn_LogLikelihood.restype = c_double

        self._cfn_Score = lib[dname + '_Score']
        self._cfn_Score.argtypes = [c_void_p, arr2d_t_ca, c_int, arr1d_caw]
        self._cfn_Score.restype = None

        self._cfn_Sample = lib[dname + '_Sample']
        self._cfn_Sample.argtypes = [c_void_p, arr2d_t_caw, c_int]
        self._cfn_Sample.restype = None

        self._cfn_Seed = lib[dname + '_Seed']
        self._cfn_Seed.argtypes = [c_void_p, c_int]
        self._cfn_Seed.restype = None

    def __del__(self):
        self._cfn_del(self._cobj)
