# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import c_int, c_double, c_void_p, c_bool

from numpy.ctypeslib import ndpointer

from sklearn import cluster

from ..base import VRLBaseEstimator

from ..base import load_libml
lib = load_libml()


class ExponentialFamilyMixture(VRLBaseEstimator):
    ''' Base Exponential Family Mixture class '''
    def __init__(self, distr=None, ncomp=1, random_state=None,
                 dtype=np.float64):
        '''
        Parameters
        ----------
        distr:  base distribution
        ncom:   number of mixture components
        random_state:   RNG
        dtype:  type of the input variables
        '''
        VRLBaseEstimator.__init__(self, random_state=random_state)

        self._distr = distr

        self._ncomp = int(ncomp)
        if self._ncomp <= 0:
            raise ValueError('ncomp must be a positive integer')

        self._dtype = dtype

        self._declare_cfn()

        self._cobj = None

        self._em_params = {'tol': 1e-4, 'max_iter': 500, 'alpha': 100.0, 'beta': 0.01}

    @property
    def ncomp(self):
        return self._ncomp

    @property
    def nvar(self):
        return self._cfn_nvar(self._cobj)

    @property
    def npar(self):
        return self._cfn_npar(self._cobj)

    @property
    def min_pst(self):
        return self._cfn_min_pst(self._cobj)

    @min_pst.setter
    def min_pst(self, val):
        if val < 0.0:
            raise ValueError
        self._cfn_set_min_pst(self._cobj, float(val))

    @property
    def weights_(self):
        weights = np.empty(self._ncomp)
        self._cfn_weights(self._cobj, weights)
        return weights

    @weights_.setter
    def weights_(self, val):
        if not np.all(val > 0.0):
            raise ValueError
        self._cfn_set_weights(self._cobj, val)

    @property
    def etas_(self):
        comp_npar = (self.npar - self._ncomp) // self._ncomp
        etas = [np.ndarray(comp_npar) for k in xrange(self._ncomp)]
        for k in xrange(self._ncomp):
            self._cfn_eta(self._cobj, k, etas[k])
        return etas

    @etas_.setter
    def etas_(self, val):
        if not isinstance(val, (list, tuple, np.ndarray)):
            raise TypeError
        if len(val) != self._ncomp:
            raise ValueError
        etas = np.array(val)
        for k in xrange(self._ncomp):
            self._cfn_set_eta(self._cobj, k, etas[k])

    def get_params(self, deep=True):
        return {'ncomp': self._ncomp,
                'nvar': self.nvar,
                'etas_': self.etas_,
                'weights_': self.weights_,
                'min_pst': self.min_pst}

    def score(self, X, out=None, hard=False):
        '''Fisher score for sample. If hard=True, each sample (row) is assigned
        to only one mixture component: that with the greater posterior.

        '''
        if out is None:
            out = np.empty(self.npar)
        if X.ndim == 1:
            self._cfn_Score(self._cobj, X.reshape(1, len(X)), 1, hard)
        else:
            self._cfn_Score(self._cobj, X, len(X), out, hard)
        return out

    def posteriors(self, x, out=None):
        ''' compute the posteriors (a.k.a. responsibilities) for a sample'''
        if out is None:
            out = np.empty(self._ncomp)
        self._cfn_Posteriors(self._cobj, x, out)
        return out

    def sample(self, n=1):
        ''' sample from the distribution '''
        if n <= 0:
            raise ValueError

        X = np.empty((n, self.nvar), dtype=self._dtype)
        self._cfn_Sample(self._cobj, X, n)

        return X.squeeze()

    # def sample(self, n=1):
    #     ''' sample from the mixture distribution '''
    #     if n <= 0:
    #         raise ValueError

    #     # sample from the multinomial of mixture weights
    #     counts = self._random_state.multinomial(n, self.weights_, size=1)
    #     counts = counts.ravel()

    #     # mixture components pdfs
    #     comp = [self._component(k) for k in xrange(self.ncomp)]

    #     # sample 'counts[k]' from each components
    #     X = np.empty((n, self.nvar), dtype=comp[0]._dtype)
    #     i = 0
    #     for k in np.flatnonzero(counts):
    #         ck = counts[k]
    #         Xk = comp[k].sample(ck)
    #         X[i:i+ck, :] = Xk
    #         i += ck

    #     # shuffle rows (matrix=list of row vectors)
    #     self._random_state.shuffle(X)

    #     if n == 1:
    #         return X[0]
    #     return X

    def _empty_component(self):
        raise NotImplementedError

    def _component(self, k):
        pdf = self._empty_component()
        eta_k = np.empty(pdf.npar)
        self._cfn_eta(self._cobj, k, eta_k)
        pdf.eta_ = eta_k
        return pdf

    def _kmeans_initial_estimate(self, data, **kwargs):
        ''' compute and set initial parameter estimate '''
        nsamples, nvar = data.shape
        if nvar != self.nvar:
            raise ValueError('input-model dimensionality mismatch'
                             '(model: {}, data: {})'.format(self.nvar, nvar))

        max_iter = 10 if 'max_iter' not in kwargs else kwargs['max_iter']

        self.logger.info('initializing mixture components ...')

        kmeans = cluster.KMeans(n_clusters=self.ncomp, init='k-means++',
                                n_init=1, max_iter=max_iter,
                                random_state=self._random_state,
                                verbose=True, n_jobs=-1)

        # fit the model (using 10% of the data) and get the assignment indices
        assignments = kmeans.fit_predict(data[:int(0.33 * len(data))])

        # per-cluster sample counts
        count = np.ndarray(self.ncomp)
        for k in xrange(self.ncomp):
            indx = np.where(assignments == k)
            count[k] = len(indx[0])

        # check for empty clusters
        if np.any(count == 0):
            indx = np.where(count == 0)[0]
            self.logger.error('{} empty clusters: {}'.format(len(indx), indx))

        # initial estimates from the samples assigned to each cluster
        comp_npar = (self.npar - self.ncomp) // self.ncomp
        weights_init = np.zeros(self.ncomp)
        etas_init = np.zeros((self.ncomp, comp_npar))
        for k in xrange(self.ncomp):
            indx = np.where(assignments == k)

            weights_init[k] = float(count[k]) / float(len(assignments))

            pdf = self._empty_component()
            pdf.fit(data[indx])
            etas_init[k] = pdf.eta_

        self.weights_ = weights_init
        self.etas_ = etas_init

        self.logger.info('done')

    def compute_initial_estimate(self, data, **kwargs):
        raise NotImplementedError

    def _fit(self, X, y=None):
        if not X.shape[1] == self.nvar:
            raise ValueError('input ({}) and model dimensionality ({}) must '
                             'agree'.format(X.shape[1], self.nvar))
        X_ = X
        if self._dtype != X.dtype:
            if not np.can_cast(X.dtype, self._dtype, casting='safe'):
                raise TypeError('input ({}) can\'t be cast to model\'s dtype'
                                '({}) safely'.format(X.dtype, self._dtype))
            X_ = X.astype(self._dtype)

        self.compute_initial_estimate(X_)

        self.logger.info('running EM ...')
        self._cfn_Fit(self._cobj, X_, X_.shape[0], self._em_params['tol'],
                      self._em_params['max_iter'], self._em_params['alpha'],
                      self._em_params['beta'])
        self.logger.info('done')

    def _transform(self, X):
        return X

    def _declare_cfn(self):
        arr1d_ca = ndpointer(dtype=np.float64, ndim=1, flags='C,A')
        arr1d_caw = ndpointer(dtype=np.float64, ndim=1, flags='C,A,W')
        arr1d_t_ca = ndpointer(dtype=self._dtype, ndim=1, flags='C,A')
        arr2d_t_ca = ndpointer(dtype=self._dtype, ndim=2, flags='C,A')
        arr2d_t_caw = ndpointer(dtype=self._dtype, ndim=2, flags='C,A,W')

        self._dname = self.__class__.__name__

        self._cfn_new = lib[self._dname + '_new']
        self._cfn_new.argtypes = [c_int, c_int]
        self._cfn_new.restype = c_void_p

        self._cfn_del = lib[self._dname + '_del']
        self._cfn_del.argtypes = [c_void_p]
        self._cfn_del.restype = None

        self._cfn_weights = lib[self._dname + '_weights']
        self._cfn_weights.argtypes = [c_void_p, arr1d_caw]
        self._cfn_weights.restype = None

        self._cfn_set_weights = lib[self._dname + '_set_weights']
        self._cfn_set_weights.argtypes = [c_void_p, arr1d_ca]
        self._cfn_set_weights.restype = None

        self._cfn_eta = lib[self._dname + '_eta']
        self._cfn_eta.argtypes = [c_void_p, c_int, arr1d_caw]
        self._cfn_eta.restype = None

        self._cfn_set_eta = lib[self._dname + '_set_eta']
        self._cfn_set_eta.argtypes = [c_void_p, c_int, arr1d_ca]
        self._cfn_set_eta.restype = None

        self._cfn_npar = lib[self._dname + '_npar']
        self._cfn_npar.argtypes = [c_void_p]
        self._cfn_npar.restype = c_int

        self._cfn_nvar = lib[self._dname + '_nvar']
        self._cfn_nvar.argtypes = [c_void_p]
        self._cfn_nvar.restype = c_int

        self._cfn_min_pst = lib[self._dname + '_min_pst']
        self._cfn_min_pst.argtypes = [c_void_p]
        self._cfn_min_pst.restype = c_double

        self._cfn_set_min_pst = lib[self._dname + '_set_min_pst']
        self._cfn_set_min_pst.argtypes = [c_void_p, c_double]
        self._cfn_set_min_pst.restype = None

        self._cfn_Sample = lib[self._dname + '_Sample']
        self._cfn_Sample.argtypes = [c_void_p, arr2d_t_caw, c_int]
        self._cfn_Sample.restype = None

        self._cfn_Score = lib[self._dname + '_Score']
        self._cfn_Score.argtypes = [c_void_p, arr2d_t_ca, c_int, arr1d_caw, c_bool]
        self._cfn_Score.restype = None

        self._cfn_Fit = lib[self._dname + '_Fit']
        self._cfn_Fit.argtypes = [c_void_p, arr2d_t_ca, c_int, c_double,
                                  c_int, c_double, c_double]
        self._cfn_Fit.restype = None

        self._cfn_Posteriors = lib[self._dname + '_Posteriors']
        self._cfn_Posteriors.argtypes = [c_void_p, arr1d_t_ca, arr1d_caw]
        self._cfn_Posteriors.restype = c_double

    def __del__(self):
        self._cfn_del(self._cobj)
