# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import c_int, c_double, c_void_p, c_bool

from numpy.ctypeslib import ndpointer

from sklearn import cluster

from ..base import VRLBaseEstimator

from .exponential_family_mixture import ExponentialFamilyMixture

from ..base import load_libml
lib = load_libml()


class FisherVector(VRLBaseEstimator):
    ''' Exponential Family Fisher Vectors '''
    def __init__(self, model=None, normalizer='diag', theta=None, hard=False,
                 alpha=0.5, norm=2, nsamples=1e5, random_state=None):
        '''
        Parameters
        ----------
        model:  mixture distribution
        normalizer:  normalization model ('eye', diag', 'var', 'block', 'full')
        theta:  Poisson parameter (it can be set at runtime)
        hard:  if True, use hard-assignment during encoding
        alpha:  power transformation
        norm:  Lp-normalization
        nsamples:  number of samples used to estimate L
        random_state:   RNG
        '''
        VRLBaseEstimator.__init__(self, random_state=random_state)

        if not isinstance(model, ExponentialFamilyMixture):
            raise ValueError('Not an exponential family mixture model')
        self._model = model

        self.__valid_normalizers = {'eye': 0, 'diag': 1, 'var': 2, 'block': 3, 'full': 4}
        if not normalizer.lower() in self.__valid_normalizers.keys():
            raise ValueError('normalization model must be in {{}}'.format(self.__valid_normalizers.keys()))
        self._normalizer = self.__valid_normalizers[normalizer.lower()]

        self._theta = 0.0 if theta is None else float(theta)

        self._hard = bool(hard)

        self._alpha = 1.0 if alpha is None else float(alpha)

        self._norm = None if norm is (None or norm < 0) else int(norm)

        self._nsamples = int(nsamples)

        self._params_runtime = ('alpha', 'norm', 'theta')

        self.__declare_cfn()

        cobj_norm = -1 if self._norm is None else self._norm
        self.__cobj = self.__cfn_new(self._model._cobj, self._normalizer, self._theta, self._hard, self._alpha, cobj_norm)

    @property
    def model(self):
        return self._model

    @property
    def normalizer(self):
        return self._normalizer

    @property
    def theta(self):
        return self._theta

    @theta.setter
    def theta(self, val):
        self._theta = 0.0 if val is None else float(val)
        self.__cfn_set_theta(self.__cobj, self._theta)

    @property
    def hard(self):
        return self._hard

    @property
    def alpha(self):
        return self._alpha

    @property
    def norm(self):
        return self._norm if self._norm > 0.0 else None

    @property
    def nsamples(self):
        return self._nsamples

    @property
    def L_(self):
        nL = self.__cfn_nL(self.__cobj)
        L = np.zeros(nL)
        self.__cfn_L(self.__cobj, L)
        return L

    @L_.setter
    def L_(self, val):
        if not len(val) == self.__cfn_nL(self.__cobj):
            raise ValueError('wrong number of dimensions')
        self.__cfn_set_L(self.__cobj, val)

    def get_params(self, deep=True):
        indx = self.__valid_normalizers.values().index(self._normalizer)
        normalizer = self.__valid_normalizers.keys()[indx]
        return {'model': self._model, 'normalizer': normalizer, 'theta':
                self._theta, 'hard': self._hard, 'alpha': self._alpha, 'norm':
                self._norm, 'L_': self.L_, 'nsamples': self._nsamples}

    def _fit(self, X, y=None):
        self.__cfn_Fit(self.__cobj, self._nsamples)
        self._L = np.empty(self.__cfn_nL(self.__cobj))
        self.__cfn_L(self.__cobj, self._L)

    def _transform(self, X):
        if isinstance(X, dict) and 'data' in X:
            X = X['data']
        npar = self.model.npar
        rep = np.empty(npar + 1 if self.theta > 0.0 else npar)
        self.__cfn_Process(self.__cobj, X, len(X), rep)
        return rep

    def __declare_cfn(self):
        sample_dtype = self._model._dtype
        arr1d_ca = ndpointer(dtype=np.float64, ndim=1, flags='C,A')
        arr2d_ca = ndpointer(dtype=sample_dtype, ndim=2, flags='C,A')
        arr1d_caw = ndpointer(dtype=np.float64, ndim=1, flags='C,A,W')

        self._dname = self.model.__class__.__name__ + self.__class__.__name__

        self.__cfn_new = lib[self._dname + '_new']
        self.__cfn_new.argtypes = [c_void_p, c_int, c_double, c_bool, c_double, c_int]
        self.__cfn_new.restype = c_void_p

        self.__cfn_del = lib[self._dname + '_del']
        self.__cfn_del.argtypes = [c_void_p]
        self.__cfn_del.restype = None

        self.__cfn_L = lib[self._dname + '_L']
        self.__cfn_L.argtypes = [c_void_p, arr1d_ca]
        self.__cfn_L.restype = None

        self.__cfn_set_L = lib[self._dname + '_set_L']
        self.__cfn_set_L.argtypes = [c_void_p, arr1d_ca]
        self.__cfn_set_L.restype = None

        self.__cfn_nL = lib[self._dname + '_nL']
        self.__cfn_nL.argtypes = [c_void_p]
        self.__cfn_nL.restype = c_int

        self.__cfn_theta = lib[self._dname + '_theta']
        self.__cfn_theta.argtypes = [c_void_p]
        self.__cfn_theta.restype = c_double

        self.__cfn_set_theta = lib[self._dname + '_set_theta']
        self.__cfn_set_theta.argtypes = [c_void_p, c_double]
        self.__cfn_set_theta.restype = None

        self.__cfn_Fit = lib[self._dname + '_Fit']
        self.__cfn_Fit.argtypes = [c_void_p, c_int]
        self.__cfn_Fit.restype = None

        self.__cfn_Process = lib[self._dname + '_Process']
        self.__cfn_Process.argtypes = [c_void_p, arr2d_ca, c_int, arr1d_caw]
        self.__cfn_Process.restype = None

    def __del__(self):
        self.__cfn_del(self.__cobj)
