# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll, c_double, c_void_p

from numpy.ctypeslib import ndpointer

from .exponential_family import lib, ExponentialFamily


class Gaussian(ExponentialFamily):

    def __init__(self, nvar=1, min_var=1e-7, random_state=None):
        ''' Multivariate Gaussian with diagonal covariance matrix

        Parameters
        ----------
        nvar:  number of input dimensions
        min_var:  min allowed variance
        '''
        ExponentialFamily.__init__(self, nvar=nvar, npar=2*nvar,
                                   random_state=random_state, dtype=np.float64)

        # create object
        self._cobj = self._cfn_new(self._nvar)

        # set optional parameters after model update
        self.min_var = min_var

    @property
    def min_var(self):
        return self._cfn_min_var(self._cobj)

    @min_var.setter
    def min_var(self, val):
        if float(val) < 0.0:
            raise ValueError('min_var must be non-negative')
        self._cfn_set_min_var(self._cobj, float(val))

    def get_params(self, deep=True):
        params = ExponentialFamily.get_params(self, deep=deep)
        params.update({'min_var': self.min_var})
        return params

    @property
    def mu(self):
        ''' get mean '''
        mu = np.empty(self.nvar)
        self._cfn_mu(self._cobj, mu)
        return mu

    @property
    def sigma(self):
        ''' get std. deviation '''
        sigma = np.empty(self.nvar)
        self._cfn_sigma(self._cobj, sigma)
        return sigma

    def _declare_cfn(self):
        ExponentialFamily._declare_cfn(self)

        arr1d_caw = ndpointer(dtype=np.float64, ndim=1, flags='C,A,W')

        self._cfn_min_var = lib['Gaussian_min_var']
        self._cfn_min_var.argtypes = [c_void_p]
        self._cfn_min_var.restype = c_double

        self._cfn_set_min_var = lib['Gaussian_set_min_var']
        self._cfn_set_min_var.argtypes = [c_void_p, c_double]
        self._cfn_set_min_var.restype = None

        self._cfn_mu = lib['Gaussian_mu']
        self._cfn_mu.argtypes = [c_void_p, arr1d_caw]
        self._cfn_mu.restype = None

        self._cfn_sigma = lib['Gaussian_sigma']
        self._cfn_sigma.argtypes = [c_void_p, arr1d_caw]
        self._cfn_sigma.restype = None
