# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll, c_double, c_void_p

from .gaussian import Gaussian
from .exponential_family_mixture import lib, ExponentialFamilyMixture


class GaussianMixture(ExponentialFamilyMixture):

    def __init__(self, ncomp=1, nvar=1, min_pst=1e-4, min_var=1e-7,
                 random_state=None):
        ''' Multivariate Gaussian with diagonal covariance matrices

        Parameters
        ----------
        ncomp:  number of mixture components
        nvar:  number of input dimensions
        min_pst:  posteriors below min_pst are set to zero
        min_var:  min allowed variance
        '''
        ExponentialFamilyMixture.__init__(self, distr=Gaussian, ncomp=ncomp,
                                          random_state=random_state,
                                          dtype=np.float64)

        # create object
        self._cobj = self._cfn_new(ncomp, nvar)

        # set optional parameters
        self.min_pst = min_pst
        self.min_var = min_var

    @property
    def min_var(self):
        return self._cfn_min_var(self._cobj)

    @min_var.setter
    def min_var(self, val):
        if float(val) < 0.0:
            raise ValueError('min_var must be non-negative')
        self._cfn_set_min_var(self._cobj, float(val))

    def get_params(self, deep=True):
        params = ExponentialFamilyMixture.get_params(self, deep=deep)
        params.update({'min_var': self.min_var})
        return params

    def _empty_component(self):
        return Gaussian(nvar=self.nvar, min_var=self.min_var,
                        random_state=self._random_state)

    def compute_initial_estimate(self, data, **kwargs):
        ExponentialFamilyMixture._kmeans_initial_estimate(self, data, **kwargs)

    def _declare_cfn(self):
        ExponentialFamilyMixture._declare_cfn(self)

        self._cfn_min_var = lib['GaussianMixture_min_var']
        self._cfn_min_var.argtypes = [c_void_p]
        self._cfn_min_var.restype = c_double

        self._cfn_set_min_var = lib['GaussianMixture_set_min_var']
        self._cfn_set_min_var.argtypes = [c_void_p, c_double]
        self._cfn_set_min_var.restype = None
