# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import c_int, c_double
from numpy.ctypeslib import ndpointer

from ..base import VRLBaseEstimator

from ..base import load_libml
lib = load_libml()

arr2d_ca = ndpointer(dtype=np.float64, ndim=2, flags='C,A')
arr2d_caw = ndpointer(dtype=np.float64, ndim=2, flags='C,A,W')
arr1di_caw = ndpointer(dtype=np.int32, ndim=1, flags='C,A,W')

_cfn_kmeans_fit = lib['kmeans_fit']
_cfn_kmeans_fit.argtypes = [c_int, c_int, c_int, c_double, arr2d_ca, c_int, arr2d_caw]
_cfn_kmeans_fit.restype = c_double

_cfn_kmeans_quantize = lib['kmeans_quantize']
_cfn_kmeans_quantize.argtypes = [c_int, c_int, arr2d_ca, arr2d_caw, c_int]
_cfn_kmeans_quantize.restype = c_double

_cfn_kmeans_encode = lib['kmeans_encode']
_cfn_kmeans_encode.argtypes = [c_int, c_int, arr2d_ca, arr2d_caw, c_int, arr1di_caw]
_cfn_kmeans_encode.restype = c_double


class KMeans(VRLBaseEstimator):
    def __init__(self, nclusters, max_iter=100, epsilon=0.01, quantize=False):
        ''' k-means++
        '''
        VRLBaseEstimator.__init__(self)

        self._nclusters = int(nclusters)

        self._max_iter = int(max_iter)

        self._epsilon = float(epsilon)

        self._quantize = bool(quantize)

        self._params_runtime = ('quantize',)

        self.centroids_ = None

    @property
    def nclusters(self):
        return self._nclusters

    @property
    def max_iter(self):
        return self._max_iter

    @property
    def epsilon(self):
        return self._epsilon

    @property
    def quantize(self):
        return self._quantize

    @quantize.setter
    def quantize(self, val):
        self._quantize = bool(val)

    def get_params(self, deep=True):
        return {'nclusters': self._nclusters, 'max_iter': self._max_iter,
                'epsilon': self._epsilon, 'quantize': self._quantize,
                'centroids_': self.centroids_}

    def _fit(self, X, y=None):
        nsamples, ndim = X.shape
        self.centroids_ = np.empty((self._nclusters, ndim))
        pot = _cfn_kmeans_fit(self._nclusters, ndim, self._max_iter, self._epsilon, X, nsamples, self.centroids_)
        self.logger.info('average potential = {:.4f}'.format(pot))

    def _transform(self, X):
        if isinstance(X, np.ndarray):
            data = X
        elif isinstance(X, dict) and 'data' in X:
            data = X['data']
        else:
            raise TypeError('wrong data type')

        if data.ndim == 1:
            data = data.reshape((1, len(data)))

        nsamples, ndim = data.shape
        if self._quantize:
            _cfn_kmeans_quantize(self._nclusters, ndim, self.centroids_, data, nsamples)
        else:
            labels = np.ndarray(nsamples, dtype=np.int32)
            _cfn_kmeans_encode(self._nclusters, ndim, self.centroids_, data, nsamples, labels)
            data = labels

        if isinstance(X, dict) and 'data' in X.keys():
            _X = X.copy()
            _X['data'] = data
            return _X

        return data
