# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll, c_double, c_void_p

from numpy.ctypeslib import ndpointer

from .exponential_family import lib, ExponentialFamily


class Laplace(ExponentialFamily):

    def __init__(self, nvar, min_scale=1e-7, random_state=None):
        '''Multivariate Laplacian (double exponential) with diagonal scale
        matrix

        Parameters
        ----------
        nvar:  number of input dimensions
        min_scale:  min allowed deviation (scale)
        '''
        ExponentialFamily.__init__(self, nvar=nvar, npar=nvar,
                                   random_state=random_state, dtype=np.float64)

        # create object
        self._cobj = self._cfn_new(self._nvar)

        # set optional parameters after model update
        self.min_scale = min_scale

    @property
    def min_scale(self):
        return self._cfn_min_scale(self._cobj)

    @min_scale.setter
    def min_scale(self, val):
        if float(val) < 0.0:
            raise ValueError('min_scale must be non-negative')
        self._cfn_set_min_scale(self._cobj, float(val))

    @property
    def scale(self):
        scale = np.empty(self.nvar)
        self._cfn_scale(self._cobj, scale)
        return scale

    def get_params(self, deep=True):
        params = ExponentialFamily.get_params(self, deep=deep)
        params.update({'min_scale': self.min_scale})
        return params

    def _declare_cfn(self):
        ExponentialFamily._declare_cfn(self)

        arr1d_caw = ndpointer(dtype=np.float64, ndim=1, flags='C,A,W')

        self._cfn_min_scale = lib['Laplace_min_scale']
        self._cfn_min_scale.argtypes = [c_void_p]
        self._cfn_min_scale.restype = c_double

        self._cfn_set_min_scale = lib['Laplace_set_min_scale']
        self._cfn_set_min_scale.argtypes = [c_void_p, c_double]
        self._cfn_set_min_scale.restype = None

        self._cfn_scale = lib['Laplace_scale']
        self._cfn_scale.argtypes = [c_void_p, arr1d_caw]
        self._cfn_scale.restype = None
