# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import sys

import numpy as np

from ctypes import cdll, c_double, c_void_p

from .laplace import Laplace
from .exponential_family_mixture import lib, ExponentialFamilyMixture


class LaplaceMixture(ExponentialFamilyMixture):

    def __init__(self, ncomp, nvar, min_pst=1e-4, min_scale=1e-7,
                 random_state=None):
        ''' Multivariate Laplace mixture

        Parameters
        ----------
        ncomp:  number of mixture components
        nvar:  number of input dimensions
        min_pst:  posteriors below min_pst are set to zero
        min_scale:  min allowed deviation (scale)
        '''
        ExponentialFamilyMixture.__init__(self, distr=Laplace, ncomp=ncomp,
                                          random_state=random_state,
                                          dtype=np.float64)

        # create object
        self._cobj = self._cfn_new(ncomp, nvar)

        # set optional parameters
        self.min_pst = min_pst
        self.min_scale = min_scale

    @property
    def min_scale(self):
        return self._cfn_min_scale(self._cobj)

    @min_scale.setter
    def min_scale(self, val):
        if float(val) < 0.0:
            raise ValueError('min_scale must be non-negative')
        self._cfn_set_min_scale(self._cobj, float(val))

    def get_params(self, deep=True):
        params = ExponentialFamilyMixture.get_params(self, deep=deep)
        params.update({'min_scale': self.min_scale})
        return params

    def _empty_component(self):
        return Laplace(nvar=self.nvar, min_scale=self.min_scale,
                       random_state=self._random_state)

    def compute_initial_estimate(self, data, **kwargs):
        ''' compute and set initial parameter estimate '''
        nsamples, nvar = data.shape
        if nvar != self.nvar:
            raise ValueError('input and model dimensionality must be the same'
                             '(model: {}, data: {})'.format(self.nvar, nvar))

        self.logger.info('Initializing mixture components ...')

        # initial weight
        weights_init = np.ones(self.ncomp) / float(self.ncomp)

        # random noise factor
        eps = 0.1 if 'eps' not in kwargs else kwargs['eps']

        # compute sample scale
        sample_scale = np.mean(np.abs(data), axis=0)

        # initial scale / eta
        etas_init = np.zeros((self.ncomp, nvar))
        for k in xrange(self.ncomp):
            scale_k = sample_scale * (1.0 + eps * self._random_state.rand())
            etas_init[k] = -1.0 * np.reciprocal(scale_k)

        self.weights_ = weights_init
        self.etas_ = etas_init

        self.logger.info('done')

    def _declare_cfn(self):
        ExponentialFamilyMixture._declare_cfn(self)

        self._cfn_min_scale = lib['LaplaceMixture_min_scale']
        self._cfn_min_scale.argtypes = [c_void_p]
        self._cfn_min_scale.restype = c_double

        self._cfn_set_min_scale = lib['LaplaceMixture_set_min_scale']
        self._cfn_set_min_scale.argtypes = [c_void_p, c_double]
        self._cfn_set_min_scale.restype = None
