# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from scipy.optimize import minimize

import ctypes as ct

import numpy.ctypeslib as npct

from ..random import RNG

from ..base import load_libml
lib = load_libml()

array1d_caw = npct.ndpointer(dtype=np.float64, ndim=1, flags='C,A,W')
array1d_cao = npct.ndpointer(dtype=np.float64, ndim=1, flags='C,A,O')
array2d_cao = npct.ndpointer(dtype=np.float64, ndim=2, flags='C,A,O')

lrpdd_opt_fn = lib['lrpdd_opt']
lrpdd_opt_fn.argtypes = [ct.c_int, ct.c_int, ct.c_int, array2d_cao,
                         array1d_cao, array1d_cao, array1d_caw]
lrpdd_opt_fn.restype = ct.c_double


class LRPDD(object):

    def __init__(self, k=1, gtol=1e-3, random_state=None):
        '''Low-rank covariance matrix approximation. See: Magsdom-Ismail,
        M. and Purnell, J. T., Approximating the Covariance Matrix of GMMs with
        Low-rank Perturbations, IJDMB (2009)
        '''
        self.__cobj = None

        assert (int(k) >= 1)
        self._k = k

        assert (gtol > 0.0)
        self._gtol = gtol

        self._D = None
        self._a = None

        self._random_state = RNG() if random_state is None else random_state

    @property
    def k(self):
        return self._k

    @property
    def D(self):
        return self._D

    @property
    def a(self):
        return self._a

    @property
    def covariance(self):
        # covariance computed using the Sherman-Morrison formula
        if self._D is None or self._a is None:
            raise RuntimeError('model not set')
        a = self._a.reshape((len(self._a), 1))
        D2 = np.square(self._D)
        invD2 = np.diag(np.reciprocal(D2))
        invD2a = invD2 * a
        return invD2 - np.dot(invD2a, invD2a.T) / (1.0 + np.dot(a.T, invD2a))

    @property
    def precision(self):
        # inverse convariance matrix
        return np.diagonal(self._D.ravel()) + np.dot(self._a, self._a.T)

    def __call__(self, X):
        return self.fit(X)

    def _f(self, p):
        D = p[:self._nvar].reshape((self._nvar, 1))
        a = p[self._nvar:].reshape((self._nvar, self._k), order='F')

        invD2 = np.reciprocal(np.square(D))
        invD2a = a * invD2

        Q = np.eye(self._k) + np.dot(a.T, invD2a)
        L = np.linalg.cholesky(np.linalg.inv(Q))

        Xta = np.dot(self._X, a)
        XtD = self._X * D.T
        log_detQ = - 2.0 * np.sum(np.log(np.abs(L.diagonal())))

        # objective
        err = np.mean(np.sum(XtD * XtD, axis=1)) \
            + np.mean(np.sum(Xta * Xta, axis=1)) \
            - 2.0 * np.sum(np.log(np.abs(D))) - log_detQ

        return err

    def _fprime(self, p):
        D = p[:self._nvar].reshape((self._nvar, 1))
        a = p[self._nvar:].reshape((self._nvar, self._k), order='F')

        invD2 = np.reciprocal(np.square(D))
        invD2a = a * invD2

        Q = np.eye(self._k) + np.dot(a.T, invD2a)
        L = np.linalg.cholesky(np.linalg.inv(Q))

        invD2aL = np.dot(invD2a, L)

        # gradient wrt D
        diagSigma = self._var.reshape(D.shape)
        diagP = invD2 - np.sum(invD2aL * invD2aL, axis=1, keepdims=True)
        grad_D = 2.0 * np.multiply(diagSigma - diagP, D)

        # gradient wrt a
        beta = np.dot(self._X, a)
        Sigmaa = np.zeros((self._nvar, self._k))
        for k in xrange(self._k):
            beta_k = beta[:, k].reshape((self._nsamples, 1))
            Sigmaa[:, k] = np.mean(self._X * beta_k, axis=0)

        invD2aLta = np.dot(invD2aL.T, a)
        grad_a = 2.0 * (Sigmaa - invD2a + np.dot(invD2aL, invD2aLta))

        grad = np.empty((1 + self._k) * self._nvar)
        grad[:self._nvar] = grad_D.ravel()
        grad[self._nvar:] = grad_a.ravel(order='F')

        return grad

    def fit(self, X):
        assert isinstance(X, np.ndarray)

        self._X = X
        self._nsamples = X.shape[0]
        self._nvar = X.shape[1]

        self._var = np.var(X, axis=0)

        npar = (1 + self._k) * self._nvar

        # parameter initialization
        p0 = np.empty(npar)
        p0[:self._nvar] = np.reciprocal(self._var + 2**-23)
        p0[self._nvar:] = self._random_state.rand(self._nvar * self._k)

        def cfn(p):
            grad = np.empty(p.shape)
            err = lrpdd_opt_fn(self._k, self._nsamples, self._nvar, self._X,
                               self._var, p, grad)
            return err, grad

        res = minimize(fun=cfn, jac=True, x0=p0, method='CG',
                       options={'disp': True, 'gtol': self._gtol})

        # res = minimize(fun=self._f, jac=self._fprime, x0=p0,
        #                method='BFGS', options={'disp': True})

        self._D = res.x[:self._nvar].reshape((self._nvar, 1))
        self._a = res.x[self._nvar:].reshape((self._nvar, self._k), order='F')
