# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll, c_uint32, c_double, c_void_p

from numpy.ctypeslib import ndpointer

from .exponential_family import lib, ExponentialFamily


class Multinomial(ExponentialFamily):

    def __init__(self, nvar, ntrials, min_p=1e-7, random_state=None):
        ''' Multinomial distribution '''
        ExponentialFamily.__init__(self, nvar=nvar, npar=nvar,
                                   random_state=random_state, dtype=np.uint32)

        # create object
        self._cobj = self._cfn_new(self._nvar)

        # set optional parameters after model update
        self.ntrials = ntrials
        self.min_p = min_p

    @property
    def ntrials(self):
        return self._cfn_ntrials(self._cobj)

    @ntrials.setter
    def ntrials(self, val):
        if np.uint32(val) < 1:
            raise ValueError('ntrials must be positive')
        self._cfn_set_ntrials(self._cobj, np.uint32(val))

    @property
    def min_p(self):
        return self._cfn_min_p(self._cobj)

    @min_p.setter
    def min_p(self, val):
        if float(val) < 0.0:
            raise ValueError('min_p must be non-negative')
        self._cfn_set_min_p(self._cobj, float(val))

    @property
    def p_(self):
        ''' get probabilities '''
        p = np.empty(self.nvar)
        self._cfn_p(self._cobj, p)
        return p

    def get_params(self, deep=True):
        params = ExponentialFamily.get_params(self, deep=deep)
        params.update({'ntrials': self.ntrials, 'min_lambda': self.min_lambda})
        return params

    def _declare_cfn(self):
        ExponentialFamily._declare_cfn(self)

        arr1d_caw = ndpointer(dtype=np.float64, ndim=1, flags='C,A,W')

        self._cfn_ntrials = lib['Multinomial_ntrials']
        self._cfn_ntrials.argtypes = [c_void_p]
        self._cfn_ntrials.restype = c_uint32

        self._cfn_set_ntrials = lib['Multinomial_set_ntrials']
        self._cfn_set_ntrials.argtypes = [c_void_p, c_uint32]
        self._cfn_set_ntrials.restype = None

        self._cfn_min_p = lib['Multinomial_min_p']
        self._cfn_min_p.argtypes = [c_void_p]
        self._cfn_min_p.restype = c_double

        self._cfn_set_min_p = lib['Multinomial_set_min_p']
        self._cfn_set_min_p.argtypes = [c_void_p, c_double]
        self._cfn_set_min_p.restype = None

        self._cfn_p = lib['Multinomial_p']
        self._cfn_p.argtypes = [c_void_p, arr1d_caw]
        self._cfn_p.restype = None
