# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll, c_uint32, c_double, c_void_p

from .multinomial import Multinomial
from .exponential_family_mixture import lib, ExponentialFamilyMixture


class MultinomialMixture(ExponentialFamilyMixture):

    def __init__(self, ncomp, nvar, ntrials, min_pst=1e-4, min_p=1e-7,
                 random_state=None):
        ''' Multinomial Mixture '''
        ExponentialFamilyMixture.__init__(self, distr=Multinomial, ncomp=ncomp,
                                          random_state=random_state,
                                          dtype=np.uint32)

        # create object
        self._cobj = self._cfn_new(ncomp, nvar)

        # set optional parameters
        self.min_pst = min_pst
        self.ntrials = ntrials
        self.min_p = min_p

    @property
    def ntrials(self):
        return self._cfn_ntrials(self._cobj)

    @ntrials.setter
    def ntrials(self, val):
        if np.uint32(val) < 1:
            raise ValueError('ntrials must be positive')
        self._cfn_set_ntrials(self._cobj, int(val))

    @property
    def min_p(self):
        return self._cfn_min_p(self._cobj)

    @min_p.setter
    def min_p(self, val):
        if float(val) < 0.0:
            raise ValueError('min_p must be non-negative')
        self._cfn_set_min_p(self._cobj, float(val))

    def get_params(self, deep=True):
        params = ExponentialFamilyMixture.get_params(self, deep=deep)
        params.update({'ntrials': self.ntrials, 'min_p': self.min_p})
        return params

    def _empty_component(self):
        return Multinomial(nvar=self.nvar, ntrials=self.ntrials,
                           min_p=self.min_p, random_state=self._random_state)

    def compute_initial_estimate(self, data, **kwargs):
        ExponentialFamilyMixture._kmeans_initial_estimate(self, data, **kwargs)

    def _declare_cfn(self):
        ExponentialFamilyMixture._declare_cfn(self)

        self._cfn_ntrials = lib['MultinomialMixture_ntrials']
        self._cfn_ntrials.argtypes = [c_void_p]
        self._cfn_ntrials.restype = c_uint32

        self._cfn_set_ntrials = lib['MultinomialMixture_set_ntrials']
        self._cfn_set_ntrials.argtypes = [c_void_p, c_uint32]
        self._cfn_set_ntrials.restype = None

        self._cfn_min_p = lib['MultinomialMixture_min_p']
        self._cfn_min_p.argtypes = [c_void_p]
        self._cfn_min_p.restype = c_double

        self._cfn_set_min_p = lib['MultinomialMixture_set_min_p']
        self._cfn_set_min_p.argtypes = [c_void_p, c_double]
        self._cfn_set_min_p.restype = None
