# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from os.path import join

from sys import getsizeof

from ..base import SKLEstimator

from ..io import load, read_shape

from sklearn.preprocessing import MultiLabelBinarizer
from sklearn.base import clone

import joblib


# def _partial_fit_binary(estimator, X, y):
#     estimator.partial_fit(X, y, classes=[0, 1])
#     return estimator


class OnlineOneVsRest(SKLEstimator):
    def __init__(self, estimator=None, to_show=None, to_hide=None, random_state=None):
        ''' One-vs-Rest classifier w/ online learning '''
        SKLEstimator.__init__(self, estimator, to_show, to_hide, random_state)

        if not hasattr(estimator, 'partial_fit'):
            raise ValueError('the estimator does not support online learning')

        self.max_mem = 1024**3   # 1GB max by default
        # self.n_jobs = -1         # use all CPUs
        self.n_iter = estimator.n_iter if hasattr(estimator, 'n_iter') else 10

        self.estimators_ = None

    def get_params(self, deep=True):
        params = SKLEstimator.get_params(self, deep=deep)
        params.update({'estimator': self.estimator, 'estimators_': self.estimators_})
        return params

    def _fit(self, X, y=None):
        raise RuntimeError('not supported')

    def _fit_from_list(self, X, y=None):
        raise RuntimeError('not supported')

    def _fit_from_file(self, X, y=None):
        raise RuntimeError('not supported')

    def _fit_from_filelist(self, X, y):
        X_ = load(join(self._input_path, X[0]))
        if isinstance(X_, dict) and 'data' in X_:
            X_ = X_['data']

        if not isinstance(X_, np.ndarray):
            raise TypeError('wrong sample type')

        if X_.ndim != 1:
            raise ValueError('there should be only one sample per input file')

        mlb = MultiLabelBinarizer()
        Y = mlb.fit_transform(y)

        n_samples, n_classes = Y.shape

        chunk_size = min(n_samples, (self.max_mem - Y.nbytes - X_.nbytes * n_classes) // X_.nbytes)
        n_chunks = int((n_samples + chunk_size - 1) // chunk_size)
        self.logger.info('learning with {} chunks of at most {} samples'.format(n_chunks, chunk_size))

        self.estimators_ = [clone(self.estimator) for i in range(n_classes)]

        for i in range(self.n_iter):

            if n_chunks > 1:
                for n in xrange(n_chunks):
                    self.logger.info('iter {}, chunk {}/{}'.format(i+1, n+1, n_chunks))

                    n1 = n * chunk_size
                    n2 = min(n1 + chunk_size, n_samples)

                    X_ = load([join(self.input_path, f) for f in X[n1:n2]])
                    Y_ = Y[n1:n2, :]

                    for j in range(n_classes):
                        self.logger.info('classifier {}/{}'.format(j+1, n_classes))
                        self.estimators_[j].partial_fit(X_, Y_[:, j], classes=[0, 1])

                    # self.estimators_ = joblib.Parallel(n_jobs=self.n_jobs, verbose=5)(
                    #     joblib.delayed(_partial_fit_binary)(self.estimators_[j], X_, Y_[:, j])
                    #     for j in range(n_classes))

            else:  # all data fits in memory: avoid multiple calls to load
                self.logger.info('iter {}'.format(i+1))

                X_ = load([join(self.input_path, f) for f in X])
                Y_ = Y

                for j in range(n_classes):
                    self.logger.info('learning classifier {}/{}'.format(j+1, n_classes))
                    self.estimators_[j].partial_fit(X_, Y_[:, j], classes=[0, 1])

                # self.estimators_ = joblib.Parallel(n_jobs=self.n_jobs, verbose=5)(
                #     joblib.delayed(_partial_fit_binary)(self.estimators_[j], X_, Y_[:, j])
                #     for j in range(n_classes))

    def decision_function(self, X):
        if not hasattr(self.estimators_[0], "decision_function"):
            raise AttributeError(
                "Base estimator doesn't have a decision_function attribute.")
        return np.array([est.decision_function(X).ravel() for est in self.estimators_]).T

    def _load_model(self):
        self.estimators_ = joblib.load(self.model_file)
        self.estimator = clone(self.estimators_[0])

    def _save_model(self):
        joblib.dump(value=self.estimators_, filename=self.model_file, compress=1)
