# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ..base import VRLBaseEstimator


class PCA(VRLBaseEstimator):
    def __init__(self, dim=-1, whiten=False):
        ''' PCA projections '''
        VRLBaseEstimator.__init__(self)

        self._dim = int(dim)

        self._whiten = bool(whiten)

        self._params_runtime = ('dim', 'whiten')

        self.mean_ = None
        self.sval_ = None
        self.svec_ = None

    @property
    def dim(self):
        return self._dim

    @dim.setter
    def dim(self, val):
        self._dim = int(val)
        if self._dim < 1:
            raise ValueError('dim must be > 0')

    @property
    def whiten(self):
        return self._whiten

    def get_params(self, deep=True):
        return {'dim': self._dim, 'whiten': self._whiten, 'mean_': self.mean_,
                'sval_': self.sval_, 'svec_': self.svec_}

    def _fit(self, X, y=None):
        if X.shape[0] <= X.shape[1]:
            raise IOError('the number of samples must be greater than the '
                          'number of input dimensions')

        self.mean_ = np.mean(X, axis=0)

        # sample covariance
        cov_data = np.cov(X, rowvar=0)

        # svec[:, i] corresponds to sval[i]
        self.sval_, self.svec_ = np.linalg.eigh(cov_data)

        # sort by decreasing eigenvalue
        indx = np.flipud(np.argsort(self.sval_))
        self.sval_ = self.sval_[indx]
        self.sval_ = np.sqrt(np.abs(self.sval_))

        # sort rows and transpose it back
        self.svec_ = np.transpose((self.svec_.T)[indx])

    def _transform(self, X):
        if isinstance(X, np.ndarray):
            data = X
        elif isinstance(X, dict) and 'data' in X:
            data = X['data']
        else:
            raise TypeError('wrong data type')

        if self.svec_.shape[1] != data.shape[1]:
            raise ValueError('wrong data dimensionality')

        if self._dim < 0:
            self.dim = X.shape[1]

        proj = np.asmatrix(self.svec_[:, :self._dim].view())
        if self._whiten:
            pca_data = np.asmatrix((data - self.mean_) / self.sval_) * proj
        else:
            pca_data = np.asmatrix(data - self.mean_) * proj
        pca_data = np.asarray(pca_data)

        # if dict, keep all other fields unchanged
        if isinstance(X, dict):
            _X = X.copy()
            _X['data'] = pca_data
            return _X

        return pca_data
