# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll, c_double, c_void_p

from numpy.ctypeslib import ndpointer

from .exponential_family import lib, ExponentialFamily


class Poisson(ExponentialFamily):

    def __init__(self, nvar, min_lambda=1e-7, random_state=None):
        ''' Multivariate Poisson distribution

        Parameters
        ----------
        nvar:  number of input dimensions
        min_lambda:  min allowed variance
        '''
        ExponentialFamily.__init__(self, nvar=nvar, npar=nvar,
                                   random_state=random_state, dtype=np.uint32)

        # create object
        self._cobj = self._cfn_new(self._nvar)

        # set optional parameters after model update
        self.min_lambda = min_lambda

    @property
    def min_lambda(self):
        return self._cfn_min_lambda(self._cobj)

    @min_lambda.setter
    def min_lambda(self, val):
        if float(val) < 0.0:
            raise ValueError('min_lambda must be non-negative')
        self._cfn_set_min_lambda(self._cobj, float(val))

    @property
    def lambda_(self):
        ''' get mean '''
        lambda_ = np.empty(self.nvar)
        self._cfn_lambda(self._cobj, lambda_)
        return lambda_

    def get_params(self, deep=True):
        params = ExponentialFamily.get_params(self, deep=deep)
        params.update({'min_lambda': self.min_lambda})
        return params

    def _declare_cfn(self):
        ExponentialFamily._declare_cfn(self)

        arr1d_caw = ndpointer(dtype=np.float64, ndim=1, flags='C,A,W')

        self._cfn_min_lambda = lib['Poisson_min_lambda']
        self._cfn_min_lambda.argtypes = [c_void_p]
        self._cfn_min_lambda.restype = c_double

        self._cfn_set_min_lambda = lib['Poisson_set_min_lambda']
        self._cfn_set_min_lambda.argtypes = [c_void_p, c_double]
        self._cfn_set_min_lambda.restype = None

        self._cfn_lambda = lib['Poisson_lambda']
        self._cfn_lambda.argtypes = [c_void_p, arr1d_caw]
        self._cfn_lambda.restype = None
