# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import c_double, c_void_p

from .poisson import Poisson
from .exponential_family_mixture import lib, ExponentialFamilyMixture


class PoissonMixture(ExponentialFamilyMixture):

    def __init__(self, ncomp, nvar, min_pst=1e-4, min_lambda=1e-7,
                 random_state=None):
        ''' Multivariate Poisson mixture

        Parameters
        ----------
        ncomp:  number of mixture components
        nvar:  number of input dimensions
        min_pst:  posteriors below min_pst are set to zero
        min_lambda:  min allowed variance
        '''
        ExponentialFamilyMixture.__init__(self, distr=Poisson, ncomp=ncomp,
                                          random_state=random_state,
                                          dtype=np.uint32)

        # create object
        self._cobj = self._cfn_new(ncomp, nvar)

        # set optional parameters
        self.min_pst = min_pst
        self.min_lambda = min_lambda

    @property
    def min_lambda(self):
        return self._cfn_min_lambda(self._cobj)

    @min_lambda.setter
    def min_lambda(self, val):
        if float(val) < 0.0:
            raise ValueError('min_lambda must be non-negative')
        self._cfn_set_min_lambda(self._cobj, float(val))

    def get_params(self, deep=True):
        params = ExponentialFamilyMixture.get_params(self, deep=deep)
        params.update({'min_lambda': self.min_lambda})
        return params

    def _empty_component(self):
        return Poisson(nvar=self.nvar, min_lambda=self.min_lambda,
                       random_state=self._random_state)

    def compute_initial_estimate(self, data, **kwargs):
        ExponentialFamilyMixture._kmeans_initial_estimate(self, data, **kwargs)

    def _declare_cfn(self):
        ExponentialFamilyMixture._declare_cfn(self)

        self._cfn_min_lambda = lib['PoissonMixture_min_lambda']
        self._cfn_min_lambda.argtypes = [c_void_p]
        self._cfn_min_lambda.restype = c_double

        self._cfn_set_min_lambda = lib['PoissonMixture_set_min_lambda']
        self._cfn_set_min_lambda.argtypes = [c_void_p, c_double]
        self._cfn_set_min_lambda.restype = None
