# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

from os.path import join, abspath

import numpy as np

from ..base import VRLBaseEstimator

from .kmeans import _cfn_kmeans_fit, _cfn_kmeans_quantize, _cfn_kmeans_encode


class ProductQuantizer(VRLBaseEstimator):
    def __init__(self, nclusters, nsplits, max_iter=100, epsilon=0.01, quantize=False,
                 random_state=None):
        VRLBaseEstimator.__init__(self, random_state=random_state)

        self._nsplits = int(nsplits)

        self._nclusters = int(nclusters)

        self._max_iter = int(max_iter)

        self._epsilon = float(epsilon)

        self._quantize = bool(quantize)

        self._params_runtime = ('quantize',)

        self.centroids_ = None

    @property
    def nsplits(self):
        return self._nsplits

    @property
    def max_iter(self):
        return self._max_iter

    @property
    def epsilon(self):
        return self._epsilon

    @property
    def quantize(self):
        return self._quantize

    @quantize.setter
    def quantize(self, val):
        self._quantize = bool(val)

    def get_params(self, deep=True):
        return {'nsplits': self._nsplits, 'nclusters': self._nclusters,
                'max_iter': self._max_iter, 'epsilon': self._epsilon,
                'quantize': self._quantize, 'centroids_': self.centroids_}

    def _fit(self, X, y=None):
        nsamples, ndim = X.shape
        if not (ndim % self._nsplits) == 0:
            raise ValueError('the number of input dimensions must be a multiple of the number of splits')
        vq_ndim = int(ndim / self._nsplits)

        self.centroids_ = [np.ndarray((self._nclusters, vq_ndim)) for i in xrange(self._nsplits)]
        pot = 0.0
        for i in xrange(self._nsplits):
            self.logger.info('quantizer {}/{}'.format(i, self._nsplits))
            Xi = X[:, i * vq_ndim: (i + 1) * vq_ndim]
            Xi = np.reshape(Xi.ravel(), (nsamples, vq_ndim))
            pot += _cfn_kmeans_fit(self._nclusters, vq_ndim, self._max_iter, self._epsilon, Xi, nsamples, self.centroids_[i])
        self.logger.info('average potential = {:.4f}'.format(pot / float(self._nsplits)))

    def _transform(self, X):
        if isinstance(X, np.ndarray):
            data = X
        elif isinstance(X, dict) and 'data' in X:
            data = X['data']
        else:
            raise TypeError('wrong data type')

        if data.ndim == 1:
            data = data.reshape((1, len(data)))

        nsamples, ndim = data.shape
        vq_ndim = int(ndim / self._nsplits)
        if self._quantize:
            for i in xrange(self._nsplits):
                data_i = data[:, i * vq_ndim: (i + 1) * vq_ndim]
                data_i = np.reshape(data_i.ravel(), (nsamples, vq_ndim))
                _cfn_kmeans_quantize(self._nclusters, vq_ndim, self.centroids_[i], data_i, nsamples)
                data[:, i * vq_ndim: (i + 1) * vq_ndim] = data_i
        else:
            labels = [np.ndarray(nsamples, dtype=np.int32) for i in xrange(self._nsplits)]
            for i in xrange(self._nsplits):
                data_i = data[:, i * vq_ndim: (i + 1) * vq_ndim]
                data_i = np.reshape(data_i.ravel(), (nsamples, vq_ndim))
                _cfn_kmeans_encode(self._nclusters, vq_ndim, self.centroids_[i], data_i, nsamples, labels[i])
            labels = np.vstack(labels).T
            data = labels

        if isinstance(X, dict) and 'data' in X.keys():
            _X = X.copy()
            _X['data'] = data
            return _X

        return data
