# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll, c_double, c_void_p

from numpy.ctypeslib import ndpointer

from .exponential_family import lib, ExponentialFamily

from .utils import matrix_is_positive_definite, matrix_is_symmetric


class Wishart(ExponentialFamily):

    def __init__(self, nvar, dof=-1, random_state=None):
        ''' Wishart distribution with given DOFs

        Parameters
        ----------
        nvar:  number of input dimensions
        dof:  number of degrees of freedom. If dof<0, set it to sqrt(nvar)
        '''
        ExponentialFamily.__init__(self, nvar=nvar, npar=nvar,
                                   random_state=random_state, dtype=np.float64)

        # create object
        self._cobj = self._cfn_new(self._nvar)

        # set optional parameters after model update
        self.dof = np.sqrt(nvar) if dof < 0 else dof

    @property
    def dof(self):
        return self._cfn_dof(self._cobj)

    @dof.setter
    def dof(self, val):
        if float(val) < 0.0:
            raise ValueError('dof must be non-negative')
        self._cfn_set_dof(self._cobj, float(val))

    @property
    def V(self):
        ''' get scale matrix '''
        ndim = int(self.nvar**0.5)
        V = np.empty((ndim, ndim))
        self._cfn_V(self._cobj, V)
        return V

    @property
    def L(self):
        ndim = int(self.nvar**0.5)
        L = np.empty((ndim, ndim))
        self._cfn_L(self._cobj, L)
        return L

    def get_params(self, deep=True):
        params = ExponentialFamily.get_params(self, deep=deep)
        params.update({'dof': self.dof})
        return params

    def _declare_cfn(self):
        ExponentialFamily._declare_cfn(self)

        arr2d_caw = ndpointer(dtype=np.float64, ndim=2, flags='C,A,W')

        self._cfn_dof = lib['Wishart_dof']
        self._cfn_dof.argtypes = [c_void_p]
        self._cfn_dof.restype = c_double

        self._cfn_set_dof = lib['Wishart_set_dof']
        self._cfn_set_dof.argtypes = [c_void_p, c_double]
        self._cfn_set_dof.restype = None

        self._cfn_V = lib['Wishart_V']
        self._cfn_V.argtypes = [c_void_p, arr2d_caw]
        self._cfn_V.restype = None

        self._cfn_L = lib['Wishart_L']
        self._cfn_L.argtypes = [c_void_p, arr2d_caw]
        self._cfn_L.restype = None
