# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

import numpy as np

from ctypes import cdll, c_double, c_void_p

from .wishart import lib, Wishart
from .exponential_family_mixture import ExponentialFamilyMixture


class WishartMixture(ExponentialFamilyMixture):

    def __init__(self, ncomp, nvar, dof=-1, min_pst=1e-4, random_state=None):
        ''' Wishart mixture with given DOFs

        Parameters
        ----------
        ncomp:  number of mixture components
        nvar:  number of input dimensions
        dof:  number of degrees of freedom. If dof<0, set it to sqrt(nvar)
        min_pst:  posteriors below min_pst are set to zero
        '''
        ExponentialFamilyMixture.__init__(self, distr=Wishart, ncomp=ncomp,
                                          random_state=random_state,
                                          dtype=np.float64)

        # create object
        self._cobj = self._cfn_new(ncomp, nvar)

        # set optional parameters
        self.min_pst = min_pst
        self.dof = np.sqrt(nvar) if dof < 0 else dof

    @property
    def dof(self):
        return self._cfn_dof(self._cobj)

    @dof.setter
    def dof(self, val):
        if float(val) < 0.0:
            raise ValueError('dof must be non-negative')
        self._cfn_set_dof(self._cobj, float(val))

    def get_params(self, deep=True):
        params = ExponentialFamilyMixture.get_params(self, deep=deep)
        params.update({'dof': self.dof})
        return params

    def _empty_component(self):
        return Wishart(nvar=self.nvar, dof=self.dof)

    def compute_initial_estimate(self, data, **kwargs):
        ExponentialFamilyMixture._kmeans_initial_estimate(self, data, **kwargs)

    def _declare_cfn(self):
        ExponentialFamilyMixture._declare_cfn(self)

        self._cfn_dof = lib['WishartMixture_dof']
        self._cfn_dof.argtypes = [c_void_p]
        self._cfn_dof.restype = c_double

        self._cfn_set_dof = lib['WishartMixture_set_dof']
        self._cfn_set_dof.argtypes = [c_void_p, c_double]
        self._cfn_set_dof.restype = None
