# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import division

from os.path import dirname, commonprefix, abspath, sep, split, join, curdir, exists

import numpy as np

from struct import calcsize

import logging


LOGFILE = None
LOGLEVEL = logging.DEBUG


def set_logfile(filename, append=True):
    global LOGFILE
    LOGFILE = None if filename is None else abspath(filename)
    if filename is not None and not append:
        open(filename, 'w').write('')


def set_loglevel(level):
    global LOGLEVEL
    LOGLEVEL = level
    for name in logging.Logger.manager.loggerDict:
        logger = logging.getLogger(name)
        for i in range(len(logger.handlers)):
            logger.handlers[i].setLevel(LOGLEVEL)
        logger.setLevel(LOGLEVEL)


def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(LOGLEVEL)
    if len(logger.handlers) > 0:
        return logger

    #fmt = logging.Formatter(fmt='%(asctime)s.%(msecs)d - %(levelname)s:%(name)s.%(funcName)s: %(message)s', datefmt='%m-%d %H:%M:%S')
    fmt = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(name)s.%(funcName)s - %(message)s')

    if LOGFILE is None:
        # add console handler
        ch = logging.StreamHandler()
        ch.setLevel(LOGLEVEL)
        ch.setFormatter(fmt)
        logger.addHandler(ch)
    else:
        # add file handler
        fh = logging.FileHandler(LOGFILE, mode='a')
        fh.setLevel(LOGLEVEL)
        fh.setFormatter(fmt)
        logger.addHandler(fh)

    return logger


def is_64bit():
    ''' returns true if running on a 64 bit system '''
    return calcsize('P') * 8 == 64


def basepath(filelist):
    ''' get largest common base directory from a list of files '''
    if not isinstance(filelist, (list, tuple)) or \
       not isinstance(filelist[0], basestring):
        raise ValueError('filelist must be a list of strings')

    if len(filelist) == 1:
        bdir = dirname(filelist[0])
    else:
        bdir = commonprefix(filelist)

    # if it does not ends with '/' is because it has some common prefix in the
    # filenames, eg. ['/a/b/c.d', '/a/b/cc.d'] -> '/a/b/c'
    if not bdir[-1] == sep:
        bdir = split(bdir)[0]

    return abspath(bdir)


def longest_common_substring(s1, s2):
    ''' taken from https://en.wikibooks.org/wiki/Algorithm_Implementation/
    Strings/Longest_common_substring#Python'''
    m = [[0] * (1 + len(s2)) for i in xrange(1 + len(s1))]
    longest, x_longest = 0, 0
    for x in xrange(1, 1 + len(s1)):
        for y in xrange(1, 1 + len(s2)):
            if s1[x - 1] == s2[y - 1]:
                m[x][y] = m[x - 1][y - 1] + 1
                if m[x][y] > longest:
                    longest = m[x][y]
                    x_longest = x
            else:
                m[x][y] = 0

    return s1[x_longest - longest: x_longest]


def bitarray(x, n_bits=8):
    ''' converts a 1D array of N ints to a binary matrix of NxN_BITS '''
    x = np.array(x).squeeze()
    if not x.ndim == 1:
        raise ValueError('x must be a 1D array')

    # def _bitarray(val):
    #     return np.fromstring(np.binary_repr(val, n_bits), dtype=np.uint8)
    # barr = np.array(map(_bitarray, x.astype(np.int))) - ord('0')

    xx = x.astype(np.int)[:, np.newaxis] // 2 ** np.arange(n_bits-1, -1, -1)
    barr = np.bitwise_and(int(1), xx)

    return barr.astype(np.uint8)
